 /**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Andreas Pfeiffer, CERN/IT                       *
  *                    <andreas.pfeiffer@cern.ch>                      *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/


#include <iostream>

#include <sstream>

#ifndef INCLUDED_AIDAANNOTATION_H
#include "AIDA_Annotation/AIDAAnnotation.h"
#endif

// for further decoupling the standalone version (mainly of Qplotter)
#ifndef STANDALONE
#include "HepUtilities/SimpleTokenizer.h"
#else
#include "SimpleTokenizerCopy.h"
using Anaphe::SimpleTokenizer;
#endif

using Anaphe::IAnnotation;
using Anaphe::AIDAAnnotation;

bool   AIDAAnnotation::add(std::string key, std::string value, bool _visibility) 
{ 
  if (annotations.find(key) == annotations.end() ) { // not yet in, add key to order list
    keyOrder.push_back(key); 
  }
  annotations[key] = value; 
  visibility [key] = _visibility; 

  return true;
}

std::string AIDAAnnotation::find(std::string key) const
{ 
  if (annotations.find(key) == annotations.end() ) { // key not in map, return empty string
    return std::string("");
  }

  return annotations.find(key)->second;
}

int AIDAAnnotation::size() const 
{ 
  return annotations.size(); 
}

void AIDAAnnotation::reset() 
{ 
  annotations.clear(); 
  keyOrder.clear();
}


std::string AIDAAnnotation::key   (int i) const
{
  size_t index = size_t(i);
  if ( index < 0 || index > keyOrder.size() ) return std::string("");

  std::string key = keyOrder[index];
  return key;

}

std::string AIDAAnnotation::value (int i) const
{
  size_t index = size_t(i);
  if ( index < 0 || index > keyOrder.size() ) return std::string("");

  std::string key = keyOrder[index];
  return annotations.find(key)->second;
}

bool    AIDAAnnotation::visibilityFlag (int i) const
{
  size_t index = size_t(i);
  if ( index < 0 || index > keyOrder.size() ) return false;

  std::string key = keyOrder[index];
  return visibility.find(key)->second;
}

bool    AIDAAnnotation::visibilityFlag (std::string key) const
{
  return visibility.find(key)->second;
}

void    AIDAAnnotation::setVisibility (std::string key, bool value)
{
  visibility.find(key)->second = value;
}


std::ostream & AIDAAnnotation::toAscii(std::ostream &os) const
{
  os << "# <AIDAAnnotation size=\"" << annotations.size() << "\" >" << std::endl;

  std::vector<std::string>::const_iterator itr = keyOrder.begin();
  while (itr != keyOrder.end()) {
    os << "#   <AnnotationAttribute "
       << "key=\"" << *itr 
       << "\" val=\"" << annotations.find(*itr)->second
       << "\" vis=\"" << ( visibility.find(*itr)->second ? "true" : "false" )
       << "\" />" << std::endl;
    itr++;
  }

  os << "# </AIDAAnnotation>" << std::endl;
  return os;
}

std::istream & AIDAAnnotation::fromAscii(std::istream &is) 
{
  SimpleTokenizer tkn("\"");

  std::string buffer;
  std::vector<std::string> tokens;

  getline(is, buffer);

  // we now should find: <AIDAAnnotation size="nn">
  if ( buffer.find("<AIDAAnnotation") == std::string::npos ) {
    std::cerr << "AIDAAnnotation::fromAscii> expected '<AIDAAnnotation', but found: '" << buffer << "'" << std::endl;
    std::cerr << "                           returning to caller ... " << std::endl;
    return is;
  }

  tokens = tkn.tokenize(buffer);
  int size = 0;
  if (tokens.size() > 1) size = atoi( tokens[1].c_str() );
  // cerr << "found tokens: " << endl;
  // for (size_t i=0; i<tokens.size(); i++) { cerr << "\t" << tokens[i] << endl; }

  for (int i=0; i<size; i++) {	// read "size" pairs of structure: # <AnnotationAttribute key="key" value="val" />

    getline(is, buffer);

    tokens = tkn.tokenize(buffer);
    //
    // would probably be a good idea to check on the keys 
    // instead of relying on order only ...
    //
    if (tokens.size() == 5) {
      add(tokens[1], tokens[3]);
    } else if (tokens.size() == 7) {
      if (tokens[5] == "true") { // ... could do with a little more checking here, I'd guess 
	add(tokens[1], tokens[3], true);
      } else {
	add(tokens[1], tokens[3], false);
      }
    } else {
      std::cerr << "AIDAAnnotation::fromAscii> expected 5 or 7 tokens, found " << tokens.size() << " instead. Skipping." << std::endl;
      std::cerr << "\ttokens found: " << std::endl;
      for (size_t i=0; i<tokens.size(); i++) { std::cerr << "\t" << tokens[i] << std::endl; }
    }
  }

  getline(is, buffer);

  // we now should find: </AIDAAnnotation>
  if ( buffer.find("</AIDAAnnotation>") == std::string::npos ) { 
    std::cerr << "buffer ='" << buffer << "'" << std::endl;
    tokens = tkn.tokenize(buffer);
    std::cerr << "found tokens: " << std::endl;
    for (size_t i=0; i<tokens.size(); i++) { std::cerr << "\t" << tokens[i] << std::endl; }
  }

  return is;
}

/// define I/O operators also for base class here
std::ostream & operator << (std::ostream &os, const IAnnotation& ann)    { return ann.toAscii(os); }
std::istream & operator >> (std::istream &is,       IAnnotation& ann)    { return ann.fromAscii(is); }

std::ostream & operator << (std::ostream &os, const AIDAAnnotation& ann) { return ann.toAscii(os); }
std::istream & operator >> (std::istream &is,       AIDAAnnotation& ann) { return ann.fromAscii(is); }

