 /**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Dino Ferrero Merlino, CERN/IT                   *
  *                    <bernardino.ferrero.merlino@cern.ch>            *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/

#include "AIDA_ScatterPlot/AIDAScatter.h"

using Anaphe::AIDAScatter2D;
using Anaphe::STLScatter2DAdapter;

AIDAScatter2D::AIDAScatter2D(double xmin ,double xmax ,double ymin ,
			     double ymax, Resolution res) :
  minx(xmin) , maxx(xmax) , miny (ymin) , maxy(ymax) , current(0)
{
  maxSize = res*res/wordSize;
  bitMap = new unsigned char[maxSize];
  stepX = (maxx-minx)/res;
  stepY = (maxy-miny)/res;
  scaledResolution = res/wordSize;
  int i;
  masks[0] = 1;
  for (i=1;i<wordSize;i++)
    masks[i] = masks[i-1]*2;
}

AIDAScatter2D::~AIDAScatter2D() {
  delete [] bitMap;
}

double AIDAScatter2D::minX() const {
  return minx;
}

double AIDAScatter2D::maxX() const {
  return maxx;
}

double AIDAScatter2D::maxY() const {
  return maxy;
}

double AIDAScatter2D::minY() const {
  return miny;
}

bool AIDAScatter2D::begin() {
  current = 0;
  curVals = 0;
  return (maxSize > 0);
}

bool AIDAScatter2D::nextPoint(double &x, double &y) {
  // Decode in one go all X coordinates in a byte
  if (curVals == 0) {
    // Look for another set of values
    while (bitMap[current] == 0x00) {
      current++;
      if (current >= maxSize)
	return false;
    }
    int i=0;
    xindex = (current % scaledResolution)*wordSize;
    while (i < wordSize) {
      if (bitMap[current] & masks[i]) {
	xVals[curVals] =  minx + stepX*(xindex+i);
	curVals++;
      }
      i++; 
    }
  }
  // Consume one value from the last byte
  x = xVals[curVals-1];
  y = miny + stepY*(current / scaledResolution);
  curVals--;
  if (curVals == 0)
    current++;
  return (current < maxSize);
}

void AIDAScatter2D::putPoint(double x, double y) {
  if (( x >= minx) && (x < maxx) && (y >=miny) && (y < maxy) ) {
    xindex = static_cast<int>((x - minx) /  stepX);
    unsigned int mask = masks[xindex % wordSize];
    yindex = static_cast<int>((y - miny) /  stepY);
    bitMap[(xindex/wordSize)+yindex*(scaledResolution)] |= mask;
  }
}

STLScatter2DAdapter::STLScatter2DAdapter( std::vector<double> *xV,  
					  std::vector<double> *yV) {
  ownVectors = false;
  xValues = xV;
  yValues = yV;
  current = 0;
  maxSize = 0 ;
}

STLScatter2DAdapter::STLScatter2DAdapter() {
  ownVectors = true;
  xValues = new std::vector<double>;
  yValues = new std::vector<double>;
  current = 0;
  maxSize = 0 ;
}

STLScatter2DAdapter::~STLScatter2DAdapter() {
  if (ownVectors) {
    delete xValues;
    delete yValues;
  }
}

double STLScatter2DAdapter::minX() const {
  std::vector <double>::const_iterator minx;
  minx = std::min_element(xValues->begin(),xValues->end());
  return (*minx);
}

double STLScatter2DAdapter::maxX() const {
  std::vector <double>::const_iterator maxx;
  maxx = std::max_element(xValues->begin(),xValues->end());
  return (*maxx);
}

double STLScatter2DAdapter::maxY() const {
  std::vector <double>::const_iterator maxy;
  maxy = std::max_element(yValues->begin(),yValues->end());
  return (*maxy);
}

double STLScatter2DAdapter::minY() const {
  std::vector <double>::const_iterator miny;
  miny = std::min_element(yValues->begin(),yValues->end());
  return (*miny);
}

bool STLScatter2DAdapter::begin() {
  // The maximum size is the minimum between the two sizes
  // (we actually expect they're equal)
  if (xValues->size() > yValues->size())
    maxSize = yValues->size();
  else
    maxSize = xValues->size();
  current = 0;
  return (maxSize > 0);
}

bool STLScatter2DAdapter::nextPoint(double &x, double &y) {
  x = (*xValues)[current];
  y = (*yValues)[current];
  current++;
  return (current < maxSize);
}

void STLScatter2DAdapter::putPoint(double x, double y) {
  xValues->push_back(x);
  yValues->push_back(y);
  maxSize++;
}
