/**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Dino Ferrero Merlino, CERN/IT                   *
  *                    <bernardino.ferrero.merlino@cern.ch>            *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/

#ifndef AIDASCATTERPLOT_H
#define AIDASCATTERPLOT_H

/** Class to provide a bitmap based implementation of scatterplot.
  The idea is that we sample a very thick grid corresponding more
  or less to the screen/printer resolution (e.g. 640X640, 1024X1024).
  The grid it's very much like a 2D histogram with very fine binning,
  except that the value is not a sum-of-weights (with error) but a
  boolean value: on/off (pixel).
*/

#include "Interfaces/IScatter2D.h"
#include <vector>
#include <algorithm>

namespace Anaphe {

  const int wordSize=8;

  class AIDAScatter2D : virtual public IScatter2D {
  public:
    /// Constructor. Paramenters are min/max on both axis and resolution
    AIDAScatter2D(double xmin ,double xmax ,double ymin ,double ymax, 
		  Resolution res) ;
    double minX() const ;
    double maxX() const ;
    double minY() const ;
    double maxY() const ;
    bool nextPoint(double &x, double &y)  ;
    void putPoint(double x, double y)  ;
    bool begin() ;
    ~AIDAScatter2D();
  private:
    unsigned char *bitMap;
    double minx;
    double maxx;
    double miny;
    double maxy;
    int current;
    int maxSize;
    double stepX;
    double stepY;
    int scaledResolution;
    int xindex;
    int yindex;
    unsigned char masks[wordSize];
    double xVals[wordSize];
    int curVals;
  };

  /** Class to adapt STL vectors to Qplotter scatterplot classes.
      Be aware the class keeps pointers to the original vector, so better
      not to delete them as long as they're used.
  */

  class STLScatter2DAdapter : virtual public IScatter2D {
  public:
    /// Constructor taking two vector<double> pointers.
    STLScatter2DAdapter(std::vector<double> *xV, 
			std::vector<double> *yV)  ;
    /// Empty constructor (creates two vector<double> instances)
    STLScatter2DAdapter()  ;
    double minX() const ;
    double maxX() const ;
    double minY() const ;
    double maxY() const ;
    bool nextPoint(double &x, double &y)  ;
    void putPoint(double x, double y)  ;
    bool begin() ;
    ~STLScatter2DAdapter();
  private:
    bool ownVectors;
    size_t current;
    size_t maxSize;
    std::vector<double> *xValues;
    std::vector<double> *yValues;
  };

} //end namespace Anaphe

#endif
