// 
// Copyright (C) CERN, Geneva 
// 
// Header file for classes IFitter, IFitParameter 
// Created  by: Andreas Pfeiffer  at Fri Sep  3 19:11:35 1999
// Modified by: Jakub T. Moscicki
//
// Last update: Time-stamp: <2002-01-21 14:22:31 moscicki>
// 
//      ap, 29-jan-01 : default arguments back, as otherwise the interface
//                      can't be used without them ....
//      ap, 16-jan-01 : removed default arguments
// 

#ifndef INTERFACE_IFITTER_H
#define INTERFACE_IFITTER_H 1

#include <iostream>
#include <string>

namespace Anaphe
{

class IVector;
class IVectorManager;

/***************************************************************/
/**
 * IFitParameter interface is used for the following purposes:
 *  - setting initial fit parameters
 *  - modification of fit parameters between fits
 *  - retrieval of the results of the fit
 *
 * Name identifier nameId() is a unique name of parameter.
 * IFitParameter is used in IFitter interface.
 */

class IFitParameter  { 

  // read-only attributes of fit parameter
public: 		
  virtual double value() const = 0;
  virtual double error() const = 0;

#ifdef AIDA_DONT_USE_STD
  virtual      string nameId() const = 0;
#else
  virtual std::string nameId() const = 0;
#endif

  virtual double start() const = 0;
  virtual double step()  const = 0;
  virtual double upperBound() const = 0;
  virtual double lowerBound() const = 0;
  virtual bool isBound() const = 0;
  virtual bool isFixed() const = 0;

  // modifiable attributes of fit parameter
public: 		
  virtual void setStart(double start) = 0;
  virtual void setStep(double step) = 0;
  virtual void setBounds(double lo, double up) = 0;
  virtual void release() = 0;

  /**
   * Return the value corresponding to +INF.
   * To set lower-only bound use: setBounds(xlow, noUpperBound())
   */
  virtual double noUpperBound() const = 0;

  /**
   * Return the value corresponding to -INF.
   * To set upper-only bound use: setBounds(noLowerBound(),xhigh)
   */
  virtual double noLowerBound() const = 0;

  // PENDING: removed for the time being
  //virtual void fix(double x)     = 0;
  //virtual void free()    = 0;

public: 
  virtual ~IFitParameter() { /* nop */; }
};

/**********************************************************************/
/**
 * IFitter interface is used to perform fits. 
 * Parameters are accessed via IFitParameter interface. 
 * Model function setModel() determines number and names of parameters.
 * Unless overriden, default values of parameters are applied (possibly
 * taking into account data vector).
 *
 * PENDING: names of parameters to be discussed (for combined models).
 */

class IFitter  { 

public:

  /** 
   * Set model function as a combination of letters:
   * G (Gaussian), E (Exponential), Pn (Polynomial of order n).
   * The number of models and their naming depends of underlying
   * implementation. However some standard naming rules apply.
   * PENDING: exact syntax for combinations to be defined
   * PENDING: at present only basic models supported
   */
  virtual void setModel(const char *)   = 0;

  /**
   * Set data vector. 
   */
  virtual void setData(const IVector *)  = 0;

  /**
   * Perform chi-square fit of previosly defined model function
   * and data vector.
   * False indicates that error occured and results are not available.
   * This may happen either if model or data are not set, or 
   * if fitting failed (minimization not converged).
   */
  virtual bool chiSquareFit()   = 0;

  /** 
   * Set starting point for given parameter.
   * Other atrributes are set to their default values.
   * Parameters are identified by case-insensitive nameId.
   * If nameId does not exist return false.
   */
  virtual bool setParameter(const char *nameId, double start, double step=1.0) = 0;

  /**
   * Retrieve fit parameter (identified by nameId).
   * Return null pointer if nameId does not exist. 
   */
  virtual IFitParameter* fitParameter(const char *nameId) = 0;

  /**
   * Include single point. 
   * Return false if index out of bounds (see IVector).
   */
  virtual bool includePoint(int i) = 0;

  /**
   * Exclude single point. 
   * Return false if index out of bounds (see IVector).
   */
  virtual bool excludePoint(int i) = 0;

  /**
   * Include all points from iMin to iMax inclusively. 
   * Return false if range invalid (see also indexing in IVector).
   */
  virtual bool includeRange(int iMin, int iMax) = 0;

  /**
   * Exclude all points from iMin to iMax inclusively.
   * Return false if range invalid (see also indexing in IVector).
   */
  virtual bool excludeRange(int iMin, int iMax) = 0;

  /**
   * Include all points (remove all previous exclusions).
   */
  virtual void includeAll() = 0;

  /**
   * Exclude all points. 
   * Should be followed by some point inclusion otherwise no data for fitting.
   */
  virtual void excludeAll() = 0;

#ifdef AIDA_DONT_USE_STD
  virtual void printParameters(     ostream &os =      cout) = 0;
  virtual void printResult    (     ostream &os =      cout) = 0;
#else
  virtual void printParameters(std::ostream &os = std::cout) = 0;
  virtual void printResult    (std::ostream &os = std::cout) = 0;
#endif

public:
  /**
   * Vector representation of the fitted curve. 
   * PENDING: not fully discussed (ownership)
   */
  virtual IVector * fittedVector(IVectorManager*) = 0;

public: 
  virtual ~IFitter() { /* nop */; }
}; 


extern "C" {
  IFitter * createIFitter();
}

} // end namespace Anaphe

#endif /* INTERFACE_FITTER_H */ 
