// 
// Header file for IFunction interface
// Created by: Jakub T. Moscicki, Oct 2000
//
// Time-stamp: <2002-01-21 14:23:03 moscicki>
// 

# ifndef INTERFACE_IFUNCTION_H__included__
# define INTERFACE_IFUNCTION_H__included__ 1

# include <vector>
# include <iostream>

#ifdef AIDA_DONT_USE_STD
typedef ostream std_ostream;
#else
typedef std::ostream std_ostream;
#endif

namespace Anaphe
{

/*************************************************************/
/**
 * Interface of scalar function \f$R^n->R\f$  
 * with optional gradient computation (\f$R^n\f$).
 */

class IFunction
{
 public:

  /**
   * Scalar value of the function.
   */
#ifdef AIDA_DONT_USE_STD
  virtual double value(const      vector<double>& x) const = 0;
#else
  virtual double value(const std::vector<double>& x) const = 0;
#endif

  /**
   * Gradient of the function.
   * The result is valid only if hasGradient()==true. 
   * Otherwise the result is some default (empty) data-point.
   */
#ifdef AIDA_DONT_USE_STD
  virtual const      vector<double>& grad(const      vector<double>& x)  const = 0;
#else
  virtual const std::vector<double>& grad(const std::vector<double>& x)  const = 0;
#endif

  /**
   * Determine whether function is able to compute gradient.
   */
  virtual bool hasGradient() const = 0;

  /**
   * Dimensionality of the domain space.
   */
  virtual int dimension() const = 0;

  /**
   * Dump textual representation on output stream.
   */
  virtual void toAscii(std_ostream &os) const = 0;

  /**
   * Compare if functions are the same.  
   * PENDING: define exactly what it means
   */
  virtual bool isEqual(const IFunction * f) const = 0;

  virtual ~IFunction() {}
};

} // end namespace Anaphe

# endif
