
#ifndef INTERFACES_IHISTOGRAM2D_H
#define INTERFACES_IHISTOGRAM2D_H 1


// Include files
#include "Interfaces/IHistogram.h"

namespace Anaphe {

  // Forward declarations
  class IHistogram1D;


  //------------------------------------------------------------------------------
  //
  // ClassName:    IHistogram2D
  //  
  // Description:  Definition of the IHistogram2D interface class
  //
  //               All methods that accept a bin number as an argument will
  //               also accept the UNDERFLOW or OVERFLOW as the argument,
  //               and as a result will give the contents of the resulting
  //               UNDERFLOW or OVERFLOW bin.
  //               The in-range bin indeces are expected to be from 0 to N-1.
  //               See also <a href="http://wwwinfo.cern.ch/asd/lhc++/AIDA/">AIDA</a>
  //
  // Authors:      Pavel Binko, Dino Ferrero Merlino, Andreas Pfeiffer
  // Date:         04/02/2000
  //
  //------------------------------------------------------------------------------


  class IHistogram2D : virtual public IHistogram {

  public:
    virtual ~IHistogram2D() { /* nop */ }

  public:


    /// ==========================================================================
    /// Histogram characteristics
    /// ==========================================================================

    /// --------------------------------------------------------------------------
    /// Fill histogram
    virtual void fill( double x, double y, double weight = 1 )                = 0;


    /// ==========================================================================
    /// Entries
    ///   The indexX accepts UNDERFLOW, values from 0 to Nx-1, or OVERFLOW
    ///   The indexY accepts UNDERFLOW, values from 0 to Ny-1, or OVERFLOW
    /// ==========================================================================

    /// Number of entries in a specific in-range bin (number of fills)
    virtual int binEntries( int indexX, int indexY ) const                    = 0;

    /// Total number of entries or bin heights in the projected slice
    /// Equivalent to projectionX().binEntries( indexX );
    virtual int binEntriesX( int indexX ) const                               = 0;
    /// Equivalent to projectionY().binEntries( indexY );
    virtual int binEntriesY( int indexY ) const                               = 0;

    /// Height of a specific in-range bin (height respects weights of entries)
    virtual double binHeight( int indexX, int indexY ) const                  = 0;

    /// Total number of bin heights in the projected slice
    /// Equivalent to projectionX().binHeight( indexX );
    virtual double binHeightX( int indexX ) const                             = 0;
    /// Equivalent to projectionY().binHeight( indexY );
    virtual double binHeightY( int indexY ) const                             = 0;

    /// Bin contents error
    virtual double binError( int indexX, int indexY ) const                   = 0;


    /// ==========================================================================
    /// Statistics
    /// ==========================================================================

    /// meanX() returns the mean of the whole histogram
    ///   as calculated on filling-time projected on the axis X and Y
    virtual double meanX() const                                              = 0;
    virtual double meanY() const                                              = 0;

    /// rmsX() returns the rms of the whole histogram
    ///   as calculated on filling-time projected on the axis X and Y
    virtual double rmsX() const                                               = 0;
    virtual double rmsY() const                                               = 0;


    /// ==========================================================================
    /// Minimum and maximum heights and bins
    /// ==========================================================================

    /// Min height of in-range bins in the whole histogram
    virtual double minBinHeight() const                                       = 0;

    /// IndexX of the in-range bin containing the maxBinHeight()
    virtual int minBinX() const                                               = 0;
    /// IndexY of the in-range bin containing the maxBinHeight()
    virtual int minBinY() const                                               = 0;

    /// Max height of in-range bins in the whole histogram
    virtual double maxBinHeight() const                                       = 0;

    /// IndexX of the in-range bin containing the maxBinHeight()
    virtual int maxBinX() const                                               = 0;
    /// IndexY of the in-range bin containing the maxBinHeight()
    virtual int maxBinY() const                                               = 0;


    /// ==========================================================================
    /// Axis
    /// ==========================================================================

    /// Get the X axis
    virtual IAxis* xAxis() const                                              = 0;

    /// Get the Y axis
    virtual IAxis* yAxis() const                                              = 0;


    /// ==========================================================================
    /// Conversions between coordinates and bin indices
    /// ==========================================================================

    /// Equivalent to xAxis().coordToIndex( coordX );
    virtual int coordToIndexX( double coordX ) const                          = 0;

    /// equivalent to yAxis().coordToIndex( coordY );
    virtual int coordToIndexY( double coordY ) const                          = 0;


    /// ==========================================================================
    /// Projections
    /// ==========================================================================

    /// --------------------------------------------------------------------------
    /// Projection on axis X
    /// Equivalent to sliceX( IHistogram::UNDERFLOW, IHistogram::OVERFLOW );
    virtual IHistogram1D* projectionX() const                                 = 0;

    /// --------------------------------------------------------------------------
    /// Projection on axis Y
    /// Equivalent to sliceY( IHistogram::UNDERFLOW, IHistogram::OVERFLOW );
    virtual IHistogram1D* projectionY() const                                 = 0;


    /// ==========================================================================
    /// Slices
    /// ==========================================================================

    /// --------------------------------------------------------------------------
    /// Slice parallel with the axis X, identified by bin indexY
    /// Equivalent to sliceX( indexY, indexY );
    virtual IHistogram1D* sliceX( int indexY ) const                          = 0;

    /// --------------------------------------------------------------------------
    /// Slice parallel with the axis Y, identified by bin indexX
    /// Equivalent to sliceY( indexX, indexX );
    virtual IHistogram1D* sliceY( int indexX ) const                          = 0;

    /// --------------------------------------------------------------------------
    /// Slice parallel with the axis X, between indexY1 and indexY2
    virtual IHistogram1D* sliceX( int indexY1, int indexY2 ) const            = 0;

    /// --------------------------------------------------------------------------
    /// Slice parallel with the axis Y, between indexX1 and indexX2
    virtual IHistogram1D* sliceY( int indexX1, int indexX2 ) const            = 0;

  };


} //end namespace Anaphe

#endif    // INTERFACES_IHISTOGRAM2D_H

