// 
// Copyright (C) CERN, Geneva 
// 
// header file for class IHistoFactory 
// Created by: Andreas Pfeiffer  at Thu Jan 13 16:17:26 2000
// 
// Last update: 
//              
// 
#ifndef INTERFACE_IHISTOFACTORY_H
#define INTERFACE_IHISTOFACTORY_H 1

#include <vector>

#include "Interfaces/IHistogramFactory.h"

namespace Anaphe {

  class HistoParameters1D;
  class HistoParameters2D;

  class IHistogram1D;
  class IHistogram2D;
  class IHistogram3D;

  class IProfileHistogram;

  class IHistoManager;

  class IHistoFactory : virtual public IHistogramFactory { 

  public: 
    virtual ~IHistoFactory() { /* nop */; }

    // --------------------------------------------------------------------------------

  public: // need to re-declare those, as they are overloaded ...
    virtual IHistogram1D * create1D( std::string title
				     , int nBinsx, double xmin, double xmax
				     , const int ID = 0) = 0;

    virtual IHistogram2D * create2D(std::string title
				    , int nBinsx, double xmin, double xmax
				    , int nBinsy, double ymin, double ymax
				    , const int ID = 0) = 0;

    // --------------------------------------------------------------------------------

  public:
    // creation of histograms from parameter object
    virtual IHistogram1D * create1D( const HistoParameters1D & par) = 0;
    virtual IHistogram2D * create2D( const HistoParameters2D & par) = 0;

    // --------------------------------------------------------------------------------

  public:
    // creation of histograms with known boundaries 
    virtual IHistogram1D * create1D(const char *label, const char *title, 
				    int nBinsx, double xmin, double xmax
				    , const char *options = 0 ) = 0;

    virtual IHistogram2D * create2D(const char *label, const char *title
				    , int nBinsx, double xmin, double xmax
				    , int nBinsy, double ymin, double ymax
				    , const char *options = 0 ) = 0;

    virtual IHistogram3D * create3D(const char *label, const char *title
				    , int nBinsx, double xmin, double xmax
				    , int nBinsy, double ymin, double ymax
				    , int nBinsz, double zmin, double zmax
				    , const char *options = 0 ) = 0;

    /// profile histograms
    virtual IProfileHistogram * createProfile(const char *label, const char *title
					      , int nBinsx, double xmin, double xmax
					      , const char *options = 0 ) = 0;

    /// dynamic histograms (adjusting min/max automatically)
    virtual IHistogram1D * dynamic1D(const char * label, const char * title, const int nBinsX) = 0;

    /// histograms with variable binning
#ifdef AIDA_DONT_USE_STD
    virtual IHistogram1D * create1DVar(const char *label, const char *title, 
				       vector<float> binEdge,
				       const char * options = 0 ) = 0;

    virtual IHistogram2D * create2DVar(const char *label, const char *title, 
				       vector<float> binEdgeX,
				       vector<float> binEdgeY,
				       const char * options = 0 ) = 0;
#else /* def AIDA_DONT_USE_STD */
    virtual IHistogram1D * create1DVar(const char *label, const char *title, 
				       std::vector<float> binEdge,
				       const char * options = 0 ) = 0;

    virtual IHistogram2D * create2DVar(const char *label, const char *title, 
				       std::vector<float> binEdgeX,
				       std::vector<float> binEdgeY,
				       const char * options = 0 ) = 0;
#endif /* def AIDA_DONT_USE_STD */
    // --------------------------------------------------------------------------------

    // select persistent store by name (e.g., Objy-DB, HBOOK-file)
    virtual void selectStore(const char * store, const char * topDir=0) = 0;

    // loading from persistent store via histo-label:
    virtual IHistogram1D * load1D(const char * label) = 0;
    virtual IHistogram2D * load2D(const char * label) = 0;
    virtual IHistogram3D * load3D(const char * label) = 0;
    virtual IProfileHistogram * loadProf(const char * label) = 0;

    // removes persistent histo by ID
    virtual void scratchHisto(const char * label) = 0;

    // copies histo into persistent store
    virtual bool store1D(IHistogram1D * h) = 0;
    virtual bool store2D(IHistogram2D * h) = 0;
    virtual bool store3D(IHistogram3D * h) = 0;
    virtual bool storeProf(IProfileHistogram * h) = 0;

    // for navigation in persistent store:
    virtual void mkdir(const char * label) = 0;
    virtual void rmdir(const char * label) = 0;
    virtual void cd   (const char * label) = 0;
    virtual void ls   ()      const  = 0;
    virtual void pwd  ()      const  = 0;

    // for call-backs:
    virtual void setManager(IHistoManager * hm) = 0;

  }; 

} //end namespace Anaphe

extern "C" {
  Anaphe::IHistoFactory * createIHistoFactory();
}

#endif /* INTERFACE_IHISTOFACTORY_H */ 
