#ifndef INTERFACES_IHISTOGRAM3D_H
#define INTERFACES_IHISTOGRAM3D_H 1


// Include files
#include "Interfaces/IHistogram.h"

namespace Anaphe {

  // Forward declarations
  class IHistogram1D;
  class IHistogram2D;


  //------------------------------------------------------------------------------
  //
  // ClassName:   IHistogram3D
  //  
  // Description: Definition of the IHistogram3D interface class
  //              See also <a href="http://wwwinfo.cern.ch/asd/lhc++/AIDA/">AIDA</a>
  //
  //              All methods that accept a bin number as an argument
  //              shall accept in-range bin indices from 0 to N-1,
  //              and also the UNDERFLOW_BIN or OVERFLOW_BIN.
  // 
  //              Number of bins N can be also obtained from
  //              the corresponding axis e.g. by xAxis()->bins()
  //
  //              Other index values are invalid,
  //              and their use shall result in an error
  //
  // Authors:     Pavel Binko, Dino Ferrero Merlino, Andreas Pfeiffer
  // Date:        16/01/2001
  //
  //------------------------------------------------------------------------------


  class IHistogram3D : virtual public IHistogram {

  public:
    virtual ~IHistogram3D() { /* nop */ }


    /// ==========================================================================
    /// Histogram characteristics
    /// ==========================================================================

    /// Fill histogram
    virtual void fill( double x, double y, double z, double weight = 1. )     = 0;


    /// ==========================================================================
    /// Entries
    /// ==========================================================================

    /// Number of entries in a specific in-range bin (number of fills)
    virtual int binEntries( int indexX, int indexY, int indexZ ) const        = 0;

    /// Total number of bin entries in projected slices can be calculated e.g. by
    ///   sliceX( 0, Ny-1, 0, Nz-1 )->binEntries( indexX )
    ///   sliceYZ( 0, Nx-1 )->binEntries( indexY, indexZ )

    /// --------------------------------------------------------------------------

    /// Height of a specific in-range bin (height respects weights of entries)
    virtual double binHeight( int indexX, int indexY, int indexZ ) const      = 0;

    /// --------------------------------------------------------------------------

    /// Bin contents error
    virtual double binError( int indexX, int indexY, int indexZ ) const       = 0;


    /// ==========================================================================
    /// Statistics
    /// ==========================================================================

    /// mean[X|Y|Z]() returns the mean of the whole histogram
    ///   as calculated on filling-time projected on the axis X, Y and Z
    virtual double meanX() const                                              = 0;
    virtual double meanY() const                                              = 0;
    virtual double meanZ() const                                              = 0;

    /// --------------------------------------------------------------------------

    /// rms[X|Y|Z]() returns the rms of the whole histogram
    ///   as calculated on filling-time projected on the axis X, Y and Z
    virtual double rmsX() const                                               = 0;
    virtual double rmsY() const                                               = 0;
    virtual double rmsZ() const                                               = 0;


    /// ==========================================================================
    /// Minimum and maximum heights and bins
    /// ==========================================================================

    /// Min height of in-range bins in the whole histogram
    virtual double minBinHeight() const                                       = 0;

    /// IndexX, indexY or indexZ of the in-range bin containing the minBinHeight()
    virtual int minBinX() const                                               = 0;
    virtual int minBinY() const                                               = 0;
    virtual int minBinZ() const                                               = 0;

    /// --------------------------------------------------------------------------

    /// Max height of in-range bins in the whole histogram
    virtual double maxBinHeight() const                                       = 0;

    /// IndexX, indexY or indexZ of the in-range bin containing the maxBinHeight()
    virtual int maxBinX() const                                               = 0;
    virtual int maxBinY() const                                               = 0;
    virtual int maxBinZ() const                                               = 0;


    /// ==========================================================================
    /// Axes
    /// ==========================================================================

    /// Get the pointer to X, Y or Z axis
    virtual IAxis* xAxis() const                                              = 0;
    virtual IAxis* yAxis() const                                              = 0;
    virtual IAxis* zAxis() const                                              = 0;


    /// ==========================================================================
    /// Conversions between coordinates and bin indices
    /// ==========================================================================

    virtual int coordToIndexX( double coordX ) const                          = 0;
    virtual int coordToIndexY( double coordY ) const                          = 0;
    virtual int coordToIndexZ( double coordZ ) const                          = 0;


    /// ==========================================================================
    /// Slices - to create the histograms returned, the slice methods should
    ///          use the same factory, which created the parent 3D histogram.
    ///          The users are fully responsible for the created histograms.
    /// ==========================================================================

    /// --------------------------------------------------------------------------

    /**
     * Create a projection parallel to the X axis.
     * Equivalent to <tt>sliceXY(UNDERFLOW_BIN,OVERFLOW_BIN)</tt>.
     */
    virtual IHistogram2D* projectionXY()const  = 0;
    /**
     * Create a projection parallel to the Z axis.
     * Equivalent to <tt>sliceXZ(UNDERFLOW_BIN,OVERFLOW_BIN)</tt>.
     */
    virtual IHistogram2D* projectionXZ() const = 0;
    /**
     * Create a projection parallel to the Y axis.
     * Equivalent to <tt>sliceYZ(UNDERFLOW_BIN,OVERFLOW_BIN)</tt>.
     */
    virtual IHistogram2D* projectionYZ() const = 0;

    /// --------------------------------------------------------------------------

    /// Slice parallel to the XY plane, range from bin indexZ1 to indexZ2 (inclusive)
    /// The returned IHistogram2D represents an instantaneous snapshot of the
    /// histogram at the time the slice was created.
    /// The X axis of the returned histogram corresponds to the X axis of this histogram.
    /// The Y axis of the returned histogram corresponds to the Y axis of this histogram.
    virtual IHistogram2D* sliceXY( int indexZ1, int indexZ2  ) const          = 0;

    /// Slice parallel to the XZ plane, range from bin indexY1 to indexY2 (inclusive)
    /// The X axis of the returned histogram corresponds to the X axis of this histogram.
    /// The Y axis of the returned histogram corresponds to the Z axis of this histogram.
    virtual IHistogram2D* sliceXZ( int indexY1, int indexY2  ) const          = 0;

    /// Slice parallel to the YZ plane, range from bin indexX1 to indexX2 (inclusive)
    /// The X axis of the returned histogram corresponds to the Y axis of this histogram.
    /// The Y axis of the returned histogram corresponds to the Z axis of this histogram.
    virtual IHistogram2D* sliceYZ( int indexX1, int indexX2 ) const           = 0;

    /// --------------------------------------------------------------------------

  };


} //end namespace Anaphe

#endif    // INTERFACES_IHISTOGRAM3D_H

