 /**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Andreas Pfeiffer, CERN/IT                       *
  *                    <andreas.pfeiffer@cern.ch>                      *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/

// 
// header file for class IVector 
// Created by: Andreas Pfeiffer  at Wed Sep  1 11:12:01 1999
// 
// Last update: 
//              
//   ap, 16-jan-01 : cleaned up std:: stuff 
// 
#ifndef INTERFACES_IVECTOR_H
#define INTERFACES_IVECTOR_H 1

#include <iostream>
#include <vector>

#ifdef AIDA_DONT_USE_STD
typedef istream std_istream;
#else
typedef std::istream std_istream;
#endif

namespace Anaphe {

  class IAnnotation;

//-ap  class IHistogram1D;
//-ap  class IHistogram2D;
//-ap  class IProfileHistogram;

  class IScalar  { 

  public: 
    virtual ~IScalar() { /* nop */; }

  public:
    // for assignment and copyConstructor rely on compiler's version

  public: 		// public methods
    virtual void   scale(double s) = 0;
    virtual void   shift(double s) = 0;

  public: 		// public methods
    virtual double value()  const = 0;
    virtual double ePlus()  const = 0;
    virtual double eMinus() const = 0;

  public: 		// public methods
    virtual void setValue (double val) = 0;
    virtual void setEPlus (double val) = 0;
    virtual void setEMinus(double val) = 0;

  public:
    virtual void fromAscii(std_istream &is) = 0;

  public:
    //
    // Arithmetic operations with other IScalars.
    // In order to avoid creating new objects (and problems with ownership)
    // the operations modify "this" like: "this" [+-*/]= "other" ...
    // The implementation has to correctly propagate the errors (assuming independence).
    // (this should be done using an external ErrorPropagationStrategy which can be
    // re-defined by the user).
    // Division by 0 sets this.value to 0, 
    // other.error == 0  sets  this.error = 0
    virtual void addScalar(const IScalar * other) = 0;
    virtual void subScalar(const IScalar * other) = 0;
    virtual void mulScalar(const IScalar * other) = 0;
    virtual void divScalar(const IScalar * other) = 0;

    // ... and with doubles (implemented using scale/shift)
    virtual void add(double val) = 0;
    virtual void sub(double val) = 0;
    virtual void mul(double val) = 0;
    virtual void div(double val) = 0;

    // merging with another scalar. This sets the value to the
    // weighted average of the two and the error to the combined error
    virtual void merge(const IScalar * other) = 0;

  }; 

  class IPoint  { 

  public: 
    virtual ~IPoint() { /* nop */; }

  public:
    // for assignment and copyConstructor rely on compiler's version

  public: 		// public methods
    virtual int  dimension() const = 0;

    // set dimensionality (== # of axes). This call is ignored, if the Point has already 
    // been initialized, i.e., the coordinates have been set.
    virtual void setDimension(int dim) = 0; 

  public: 		// public methods
    virtual double value()     const  = 0;
    virtual double vPlus()     const  = 0;
    virtual double vMinus()    const  = 0;

  public:
    virtual void scaleValue(double scale)  = 0;
    virtual void shiftValue(double scale)  = 0;

  public:
    virtual double coordinate(int index) const = 0;
    virtual double coordPlus (int index) const = 0;
    virtual double coordMinus(int index) const = 0;

    virtual void  scaleCoordinate(int index, double scale) = 0;
    virtual void  shiftCoordinate(int index, double scale) = 0;

  public:
    // reading from ASCII file. This will overwrite (== destroy) any information already stored.
    virtual void  fromAscii(std_istream &is) = 0;

    // modifiers ... (first, for SWIGification :-))
    virtual IScalar * coordScalar(int index) = 0;
    virtual IScalar * vScalar()  = 0;

    virtual const IScalar * coordScalar(int index) const = 0;
    virtual const IScalar * vScalar() const  = 0;

  public:
    //
    // Arithmetic operations with other IPoints, acting on the "value" only,
    // the coordinates are not updated.
    // In order to avoid creating new objects (and problems with ownership)
    // the operations modify "this" like: "this" [+-*/]= "other" ...
    // The implementation has to correctly propagate the errors (assuming independence).
    // These operations are only defined if the dimension (number of coordinates) is
    // equal. 
    // 
    virtual void addPoint(const IPoint * other) = 0;
    virtual void subPoint(const IPoint * other) = 0;
    virtual void mulPoint(const IPoint * other) = 0;
    virtual void divPoint(const IPoint * other) = 0;

    // ... and with doubles (implemented using scale/shift) acting on the "value"
    virtual void add(double val) = 0;
    virtual void sub(double val) = 0;
    virtual void mul(double val) = 0;
    virtual void div(double val) = 0;
  };


  class IVector  { 

  public: 
    virtual ~IVector()  { /* nop */; }

  public:
    // for assignment and copyConstructor rely on compiler's version

    virtual IVector * deepClone() = 0;          // return an independent copy of self

  public:
    virtual       IAnnotation * annotation()       = 0; 
    virtual const IAnnotation * annotation() const = 0;

  public:
    virtual int  dimension()          const = 0; // dimension = number of coordinate axes

    // set # of points along each of axes. This call is ignored, if the Vector has already
    // been "set", i.e., the points are already stored.
#ifdef AIDA_DONT_USE_STD
    virtual void setPointsPerDim(     vector<int> vIn) = 0; 
#else
    virtual void setPointsPerDim(std::vector<int> vIn) = 0; 
#endif

    virtual int nPoints()             const = 0; // total number of Points in vector 
    virtual int nPoints(int index)    const = 0; // number of Points along dimension "index"

  public:
    // remove point specified by index in linear access (even for higher dimensions)
    virtual void removePoint(int index) = 0; 

    // add new point 
    virtual void addPoint(IPoint * newPoint) = 0; 

    // modifiers ... (first, for SWIGification :-))
    virtual IPoint * point(int index) = 0;                          // index by linear access (even for higher dimensions)
    virtual IPoint * point(int indexX, int indexY) = 0;             // index by x/y    access (for user conveniency)
    virtual IPoint * point(int indexX, int indexY, int indexZ) = 0; // index by x/y/z  access (for user conveniency)

    virtual const IPoint * point(int index) const = 0;                          // index by linear access (even for higher dimensions)
    virtual const IPoint * point(int indexX, int indexY) const = 0;             // index by x/y    access (for user conveniency)
    virtual const IPoint * point(int indexX, int indexY, int indexZ) const = 0; // index by x/y/z  access (for user conveniency)
    //-ap #ifdef AIDA_DONT_USE_STD
    //-ap 	virtual IPoint * point(     vector<int> ijk) const = 0; // index by "array" access (for higher dims, validating dim)
    //-ap #else
    //-ap 	virtual IPoint * point(std::vector<int> ijk) const = 0; // index by "array" access (for higher dims, validating dim)
    //-ap #endif

  public:
    virtual IPoint * maxValue() const = 0;
    virtual IPoint * minValue() const = 0;

  public:
    virtual void scaleV(double scale) = 0;
    virtual void shiftV(double shift) = 0;

    virtual void scaleCoordinate(int index, double scale) = 0;
    virtual void shiftCoordinate(int index, double shift) = 0;

  public:

    // writing the vector to an ASCII file. Format fixed at present, later XML.
    virtual void toAscii(const char * file)  = 0;

    // reading vector back from ASCII file (format fixed, later XML). This will 
    // overwrite (== destroy) any information already stored in the vector.
    virtual void fromAscii(const char *file) = 0;

//-ap	 virtual void fromHisto1D(IHistogram1D * hist) = 0;
//-ap	 virtual void fromHisto2D(IHistogram2D * hist) = 0;
//-ap	 virtual void fromProfile(IProfileHistogram * hist, const char * opt = 0) = 0;

    virtual void replaceFromFunction(int i, double val) = 0;

  public:
    //
    // Arithmetic operations with other IVectors, acting on the "value" only,
    // the coordinates are not updated.
    // In order to avoid creating new objects (and problems with ownership)
    // the operations modify "this" like: "this" [+-*/]= "other" ...
    // The implementation has to correctly propagate the errors (assuming independence).
    // These operations are only defined if the dimension (number of coordinates) is
    // equal. 
    // 
    virtual void addVector(const IVector * other) = 0;
    virtual void subVector(const IVector * other) = 0;
    virtual void mulVector(const IVector * other) = 0;
    virtual void divVector(const IVector * other) = 0;

    // ... and with doubles (implemented using scale/shift) acting on the "value"
    virtual void add(double val) = 0;
    virtual void sub(double val) = 0;
    virtual void mul(double val) = 0;
    virtual void div(double val) = 0;
  }; 

} //end namespace Anaphe


#endif /* INTERFACES_IVECTOR_H */ 
