
#ifndef INTERFACES_IANNOTATION_H
#define INTERFACES_IANNOTATION_H 1

// Include files
#include <list>
#include <string>

#include <utility>		// that's where we find the "pair"s ... 

//------------------------------------------------------------------------------
//
// ClassName:    IAnnotation
//  
// Description:  Definition of the IAnnotation interface class
//
//               IAnnotation is a store for information correlated with
//               histograms. It provides a (subset of a) map-like interface
//               for storing and retrieving key/value (string/string) pairs.
//
// Design principles :
//
//  1) Only interfaces, basic types, and types from foundation libraries
//     are allowed (STL is currently the only one foundation library)
//     - Uses only int and double as basic types
//     - Internals (classes, enums, or others) of concrete packages
//       do not appear in the interface
//  2) Accessors - XXX() (no getXXX() functions)
//  3) In member function names
//     - in-range bins are not specified explicitly (e.g. sumBinHeights())
//     - ...All... mean all (in-range and extra) bins (e.g. sumAllBinHeights())
//     - ...Extra... mean extra bins (e.g. sumExtraBinHeights())
//
//
// Authors:      Andreas Pfeiffer
// Date:         26/07/2000
//
// last modification: 
//      ap, 03-nov-00 : removed access to full list, introduced access by index instead.
//                      added visibility flag feature ...
//      ap, 16-dec-00 : added std:: stuff
//      ap, 16-jan-01 : removed default arguments
//------------------------------------------------------------------------------

namespace Anaphe {

  class IAnnotation {

  public:
    virtual ~IAnnotation() { /* nop */ }

  public:
    /// Get the number of items in the Annotation
    virtual int size() const                                      = 0;

#ifdef AIDA_DONT_USE_STD

    /// add a key/value pair with a given visibility 
    virtual bool add(string key, string value, bool visibility = true) = 0;

    /// Retrieve the value for a given key
    virtual string find(string key) const                 = 0;

    /// Retrieve the visibility flag for a given key
    virtual bool visibilityFlag(string key) const         = 0;

    virtual void setVisibility(string key, bool value)    = 0;

    /// For "iterative" access to the objects ... 
    /// in future releases we might have an iterator ...

    /// Idividual access to the Annotation-items
    /// This probably should be replaced by a call which 
    /// returns a AnnotationItem object ...
    virtual string key   	       (int index) const        = 0;
    virtual string value 	       (int index) const        = 0;
    virtual bool visibilityFlag  (int index) const        = 0;

    /// Write to and read from Ascii stream
    virtual ostream& toAscii (ostream &os) const          = 0;
    virtual istream& fromAscii (istream &is)              = 0;

#else 

    /// add a key/value pair with a given visibility 
    virtual bool add(std::string key, std::string value, bool visibility = true) = 0;

    /// Retrieve the value for a given key
    virtual std::string find(std::string key) const               = 0;

    /// Retrieve the visibility flag for a given key
    virtual bool visibilityFlag(std::string key) const            = 0;

    virtual void setVisibility(std::string key, bool value)       = 0;

    /// For "iterative" access to the objects ... 
    /// in future releases we might have an iterator ...

    /// Idividual access to the Annotation-items
    /// This probably should be replaced by a call which 
    /// returns a AnnotationItem object ...
    virtual std::string key   	       (int index) const        = 0;
    virtual std::string value 	       (int index) const        = 0;
    virtual bool        visibilityFlag   (int index) const        = 0;

    /// Write to and read from Ascii stream
    virtual std::ostream& toAscii   (std::ostream &os) const      = 0;
    virtual std::istream& fromAscii (std::istream &is)            = 0;

#endif // dont use std
    /// Reset contents
    virtual void reset()                                          = 0;

  };

#ifndef SWIG

#ifdef AIDA_DONT_USE_STD
  ostream & operator << (ostream &os, const IAnnotation& ann);
  istream & operator >> (istream &is,       IAnnotation& ann);
#else 
  std::ostream & operator << (std::ostream &os, const IAnnotation& ann);
  std::istream & operator >> (std::istream &is,       IAnnotation& ann);
#endif // dont use std

#endif // ndef swig

} //end namespace Anaphe

#endif    // INTERFACES_IANNOTATION_H
