// 
// Copyright (C) CERN, Geneva 
// 
// header file for class IHistoManager 
// Created by: Andreas Pfeiffer  at Thu Sep  2 14:29:35 1999
// 
// Last update: 
//              
// 
#ifndef INTERFACE_IHISTOMANAGER_H
#define INTERFACE_IHISTOMANAGER_H 1

#include <iostream>

#include <vector>

namespace Anaphe {

  class IHistogram;
  class IHistogram1D;
  class IHistogram2D;
  class IHistogram3D;
  class IProfileHistogram;

  class IHistoFactory;

  class IHistoManager { 

  public: 
    virtual ~IHistoManager() { /* nop */; }

  public:
    virtual bool unregister(IHistogram * h) = 0;

  public: 		// public methods
    virtual bool register1D(IHistogram1D * h) = 0;
    virtual bool register2D(IHistogram2D * h) = 0;
    virtual bool register3D(IHistogram3D * h) = 0;
    virtual bool registerProf(IProfileHistogram * h) = 0;

    virtual IHistogram1D * retrieveHisto1D(const char * label) = 0;
    virtual IHistogram2D * retrieveHisto2D(const char * label) = 0;
    virtual IHistogram3D * retrieveHisto3D(const char * label) = 0;
    virtual IProfileHistogram * retrieveProf(const char * label) = 0;
  
    virtual void deleteHisto(const char * label) = 0;

#ifdef AIDA_DONT_USE_STD
    virtual void list(     ostream& os =      cout) = 0;
#else
    virtual void list(std::ostream& os = std::cout) = 0;
#endif

    virtual void disableOverwrite() = 0;
    virtual void enableOverwrite () = 0;
    virtual void disableWarnOverwrite() = 0;
    virtual void enableWarnOverwrite () = 0;

  public: 		// public methods
    virtual void setFactory(IHistoFactory *f) = 0;

    // --------------------------------------------------------------------------------

    virtual IHistogram1D * create1D(const char * label, const char *title
				    , int nBinsx, double xmin, double xmax
				    , const char * options = 0) = 0;
    virtual IHistogram2D * create2D(const char * label, const char *title
				    , int nBinsx, double xmin, double xmax
				    , int nBinsy, double ymin, double ymax
				    , const char * options = 0) = 0;

    virtual IHistogram3D * create3D(const char * label, const char *title
				    , int nBinsx, double xmin, double xmax
				    , int nBinsy, double ymin, double ymax
				    , int nBinsz, double zmin, double zmax
				    , const char * options = 0) = 0;

    /// profile histograms
    virtual IProfileHistogram * createProfile(const char * label, const char *title
					      , int nBinsx=10, double xmin=0, double xmax=10
					      , const char *options = 0) = 0;

    /// dynamic histograms (adjusting min/max automatically)
    virtual IHistogram1D * createDynamic1D(const char *label, const char *title, int nBinsx) = 0;

    /// histograms with variable binning
#ifdef AIDA_DONT_USE_STD
    virtual IHistogram1D * create1DVar(const char *label, const char *title, 
				       vector<float> binEdge,
				       const char * options = 0 ) = 0;

    virtual IHistogram2D * create2DVar(const char *label, const char *title, 
				       vector<float> binEdgeX,
				       vector<float> binEdgeY,
				       const char * options = 0 ) = 0;
#else /* def AIDA_DONT_USE_STD */
    virtual IHistogram1D * create1DVar(const char *label, const char *title, 
				       std::vector<float> binEdge,
				       const char * options = 0 ) = 0;

    virtual IHistogram2D * create2DVar(const char *label, const char *title, 
				       std::vector<float> binEdgeX,
				       std::vector<float> binEdgeY,
				       const char * options = 0 ) = 0;
#endif /* def AIDA_DONT_USE_STD */

    // --------------------------------------------------------------------------------

    // loading from persistent store via histo-ID:
    virtual IHistogram1D * load1D(const char * label) = 0;
    virtual IHistogram2D * load2D(const char * label) = 0;
    virtual IHistogram3D * load3D(const char * label) = 0;
    virtual IProfileHistogram * loadProf(const char * label) = 0;

    virtual void scratchHisto(const char * label) = 0;

    virtual void store(const char * label)  = 0;

    // for navigation in persistent store:
    virtual void selectStore(const char *name, const char *topDir=0) = 0;
    virtual void mkdir(const char *name) = 0;
    virtual void rmdir(const char *name) = 0;
    virtual void cd   (const char *name) = 0;
    virtual void ls   ()          const  = 0;
    virtual void pwd  ()          const  = 0;

  }; 

} //end namespace Anaphe

extern "C" {
  Anaphe::IHistoManager * createIHistoManager();
}
#endif /* INTERFACE_IHISTOMANAGER_H */ 
