
#ifndef INTERFACES_IPROFILEHISTOGRAM_H
#define INTERFACES_IPROFILEHISTOGRAM_H 1


// Include files
#include "Interfaces/IHistogram.h"

namespace Anaphe {

  // Forward declarations


  //------------------------------------------------------------------------------
  //
  // ClassName:    IProfileHistogram
  //  
  // Description:  Definition of the IProfileHistogram interface class
  //
  //               All methods that accept a bin number as an argument will
  //               also accept the UNDERFLOW or OVERFLOW as the argument,
  //               and as a result will give the contents of the resulting
  //               UNDERFLOW or OVERFLOW bin.
  //               The in-range bin indeces are expected to be from 0 to N-1.
  //               See also <a href="http://wwwinfo.cern.ch/asd/lhc++/AIDA/">AIDA</a>
  //
  // Authors:      Pavel Binko, Dino Ferrero Merlino, Andreas Pfeiffer
  // Date:         22/03/2000
  //
  //------------------------------------------------------------------------------


  class IProfileHistogram : virtual public IHistogram {

  public:
    virtual ~IProfileHistogram() { /* nop */ }

  public:


    /// ==========================================================================
    /// Histogram characteristics
    /// ==========================================================================

    /// --------------------------------------------------------------------------
    /// Fill histogram
    virtual void fill( double x, double y, double weight = 1 )                = 0;


    /// ==========================================================================
    /// Entries
    ///   The index accepts UNDERFLOW, values from 0 to N-1, or OVERFLOW
    /// ==========================================================================

    /// Number of entries in a specific in-range bin (number of fills)
    virtual int binEntries( int index ) const                                 = 0;

    /// Height of a specific in-range bin (height respects weights of entries)
    virtual double binHeight( int index ) const                               = 0;

    /// Bin contents error
    virtual double binError( int index ) const                                = 0;

    /// Bin contents spread
    virtual double binSpread( int index ) const                               = 0;


    /// ==========================================================================
    /// Statistics
    /// ==========================================================================

    /// mean() returns the mean of the whole histogram
    ///   as calculated on filling-time
    virtual double mean() const                                               = 0;

    /// rms() returns the rms of the whole histogram
    ///   as calculated on filling-time
    virtual double rms() const                                                = 0;


    /// ==========================================================================
    /// Minimum and maximum heights and bins
    /// ==========================================================================

    /// Min height of in-range bins in the whole histogram
    virtual double minBinHeight() const                                       = 0;
    /// Index of the in-range bin containing the minBinHeight()
    virtual int minBin() const                                                = 0;

    /// Max height of in-range bins in the whole histogram
    virtual double maxBinHeight() const                                       = 0;
    /// Index of the in-range bin containing the maxBinHeight()
    virtual int maxBin() const                                                = 0;


    /// ==========================================================================
    /// Axis
    /// ==========================================================================

    /// Get the X axis
    virtual IAxis* xAxis() const                                              = 0;


    /// ==========================================================================
    /// Conversions between coordinates and bin indices
    /// ==========================================================================

    /// Equivalent to xAxis().coordToIndex( coord );
    virtual int coordToIndex( double coord ) const                            = 0;

  };


} //end namespace Anaphe

#endif    // INTERFACES_IPROFILEHISTOGRAM_H
