#ifndef GBLIB_BASICVECTOR2D_HH
#define GBLIB_BASICVECTOR2D_HH

namespace gblib {

  /**
   * Base class for Point2D, Vector2D and Normal2D.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  template<class T> class BasicVector2D {
  protected:
    T v_[2];
    
  public:
    // Default constructor.
    BasicVector2D() { v_[0] = 0; v_[1] = 0; }

    // Constructor from three numbers.
    BasicVector2D(T x, T y) { v_[0] = x; v_[1] = y; }

    // Copy constructor.
    // Small trick: BasicVector2D<double> hass constructors
    // from BasicVector2D<double> (provided by compiler) and
    // form BasicVector2D<float> (defined below);
    // BasicVector2D<float> has only the last one.
    BasicVector2D(const BasicVector2D<float> & v) {
      v_[0] = v.x(); v_[1] = v.y();
    }

    // Destructor.
    ~BasicVector2D() {}

    // Assignment/modify.
    BasicVector2D<T> & operator= (const BasicVector2D<T> & v) {
      v_[0] = v.v_[0]; v_[1] = v.v_[1]; return *this;
    }
    BasicVector2D<T> & operator+=(const BasicVector2D<T> & v) {
      v_[0] += v.v_[0]; v_[1] += v.v_[1]; return *this;
    }
    BasicVector2D<T> & operator-=(const BasicVector2D<T> & v) {
      v_[0] -= v.v_[0]; v_[1] -= v.v_[1]; return *this;
    }
    BasicVector2D<T> & operator*=(T a) {
      v_[0] *= a; v_[1] *= a; return *this;
    }
    BasicVector2D<T> & operator/=(T a) {
      v_[0] /= a; v_[1] /= a; return *this;
    }
    
    // Get components by index.
    T operator()(int i) const { return v_[i]; }
    T operator[](int i) const { return v_[i]; }
    
    // Set components by index.
    T & operator()(int i) { return v_[i]; }
    T & operator[](int i) { return v_[i]; }
    
    // Get components in cartesian coordinate system.
    T x()    const { return v_[0]; }
    T y()    const { return v_[1]; }
    T getX() const { return v_[0]; }
    T getY() const { return v_[1]; }

    // Set components in cartesian coordinate system.
    void setX(T a) { v_[0] = a; }
    void setY(T a) { v_[1] = a; }
    void set(T x, T y) { v_[0] = x; v_[1] = y; }

    // Magnitude of the vector.
    T mag2() const { return x()*x()+y()*y(); }
    T mag()  const { return sqrt(mag2()); }

    // Scalar product.
    T dot(const BasicVector2D<T> & v) const {
      return x()*v.x()+y()*v.y();
    }

    // Returns unit vector parallel to this.
    BasicVector2D<T> unit() const {
      T len = mag();
      return (len > 0.) ?
	BasicVector2D<T>(x()/len, y()/len) : BasicVector2D<T>();
    }
  };

  /*************************************************************************
   *                                                                       *  
   * Non-member functions for BasicVector2D<float>                         *
   *                                                                       *  
   *************************************************************************/

  // Stream operators
  std::ostream &
  operator<<(std::ostream &, const BasicVector2D<float> &);

  std::istream &
  operator>>(std::istream &, BasicVector2D<float> &);

  // Arithmetic
  inline BasicVector2D<float>
  operator+(const BasicVector2D<float> & v) { return v; }

  inline BasicVector2D<float>
  operator+(const BasicVector2D<float> & a, const BasicVector2D<float> & b) {
    return BasicVector2D<float>(a.x()+b.x(), a.y()+b.y());
  }
  inline BasicVector2D<float>
  operator-(const BasicVector2D<float> & v) {
    return BasicVector2D<float>(-v.x(), -v.y());
  }
  inline BasicVector2D<float>
  operator-(const BasicVector2D<float> & a, const BasicVector2D<float> & b) {
    return BasicVector2D<float>(a.x()-b.x(), a.y()-b.y());
  }
  inline BasicVector2D<float>
  operator*(const BasicVector2D<float> & v, float a) {
    return BasicVector2D<float>(v.x()*a, v.y()*a);
  }
  inline float
  operator*(const BasicVector2D<float> & a, const BasicVector2D<float> & b) {
    return a.dot(b);
  }
  inline BasicVector2D<float>
  operator*(float a, const BasicVector2D<float> & v) {
    return BasicVector2D<float>(a*v.x(), a*v.y());
  }
  inline BasicVector2D<float>
  operator/(const BasicVector2D<float> & v, float a) {
    return BasicVector2D<float>(v.x()/a, v.y()/a);
  }
  
  // Comparisons
  inline bool
  operator==(const BasicVector2D<float> & a, const BasicVector2D<float> & b) {
    return (a.x()==b.x() && a.y()==b.y());
  }
  inline bool
  operator!=(const BasicVector2D<float> & a, const BasicVector2D<float> & b) {
    return (a.x()!=b.x() || a.y()!=b.y());
  }

  /*************************************************************************
   *                                                                       *  
   * Non-member functions for BasicVector2D<double>                        *
   *                                                                       *  
   *************************************************************************/

  // Stream operators
  std::ostream &
  operator<<(std::ostream &, const BasicVector2D<double> &);

  std::istream &
  operator>>(std::istream &, BasicVector2D<double> &);

  // Arithmetic
  inline BasicVector2D<double>
  operator+(const BasicVector2D<double> & v) { return v; }

  inline BasicVector2D<double>
  operator+(const BasicVector2D<double> & a,const BasicVector2D<double> & b) {
    return BasicVector2D<double>(a.x()+b.x(), a.y()+b.y());
  }
  inline BasicVector2D<double>
  operator-(const BasicVector2D<double> & v) {
    return BasicVector2D<double>(-v.x(), -v.y());
  }
  inline BasicVector2D<double>
  operator-(const BasicVector2D<double> & a,const BasicVector2D<double> & b) {
    return BasicVector2D<double>(a.x()-b.x(), a.y()-b.y());
  }
  inline BasicVector2D<double>
  operator*(const BasicVector2D<double> & v, double a) {
    return BasicVector2D<double>(v.x()*a, v.y()*a);
  }
  inline double
  operator*(const BasicVector2D<double> & a,const BasicVector2D<double> & b) {
    return a.dot(b);
  }
  inline BasicVector2D<double>
  operator*(double a, const BasicVector2D<double> & v) {
    return BasicVector2D<double>(a*v.x(), a*v.y());
  }
  inline BasicVector2D<double>
  operator/(const BasicVector2D<double> & v, double a) {
    return BasicVector2D<double>(v.x()/a, v.y()/a);
  }
  
  // Comparisons
  inline bool
  operator==(const BasicVector2D<double> & a, const BasicVector2D<double> & b)
  {
    return (a.x()==b.x() && a.y()==b.y());
  }
  inline bool
  operator!=(const BasicVector2D<double> & a, const BasicVector2D<double> & b)
  {
    return (a.x()!=b.x() || a.y()!=b.y());
  }
}

#endif /* GBLIB_BASICVECTOR2D_HH */
