#ifndef GBLIB_BASICVECTOR3D_HH
#define GBLIB_BASICVECTOR3D_HH

namespace gblib {

  /**
   * Base class for Point3D, Vector3D and Normal3D.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  template<class T> class BasicVector3D {
  protected:
    T v_[3];
    
  public:
    // Default constructor.
    BasicVector3D() { v_[0] = 0; v_[1] = 0; v_[2] = 0; }

    // Constructor from three numbers.
    BasicVector3D(T x, T y, T z) { v_[0] = x; v_[1] = y; v_[2] = z; }

    // Copy constructor.
    // Small trick: BasicVector3D<double> hass constructors
    // from BasicVector3D<double> (provided by compiler) and
    // form BasicVector3D<float> (defined below);
    // BasicVector3D<float> has only the last one.
    BasicVector3D(const BasicVector3D<float> & v) {
      v_[0] = v.x(); v_[1] = v.y(); v_[2] = v.z();
    }

    // Destructor.
    ~BasicVector3D() {}

    // Assignment/modify.
    BasicVector3D<T> & operator= (const BasicVector3D<T> & v) {
      v_[0] = v.v_[0]; v_[1] = v.v_[1]; v_[2] = v.v_[2]; return *this;
    }
    BasicVector3D<T> & operator+=(const BasicVector3D<T> & v) {
      v_[0] += v.v_[0]; v_[1] += v.v_[1]; v_[2] += v.v_[2]; return *this;
    }
    BasicVector3D<T> & operator-=(const BasicVector3D<T> & v) {
      v_[0] -= v.v_[0]; v_[1] -= v.v_[1]; v_[2] -= v.v_[2]; return *this;
    }
    BasicVector3D<T> & operator*=(T a) {
      v_[0] *= a; v_[1] *= a; v_[2] *= a; return *this;
    }
    BasicVector3D<T> & operator/=(T a) {
      v_[0] /= a; v_[1] /= a; v_[2] /= a; return *this;
    }
    
    // Get components by index.
    T operator()(int i) const { return v_[i]; }
    T operator[](int i) const { return v_[i]; }
    
    // Set components by index.
    T & operator()(int i) { return v_[i]; }
    T & operator[](int i) { return v_[i]; }
    
    // Get components in cartesian coordinate system.
    T x()    const { return v_[0]; }
    T y()    const { return v_[1]; }
    T z()    const { return v_[2]; }
    T getX() const { return v_[0]; }
    T getY() const { return v_[1]; }
    T getZ() const { return v_[2]; }

    // Set components in cartesian coordinate system.
    void setX(T a) { v_[0] = a; }
    void setY(T a) { v_[1] = a; }
    void setZ(T a) { v_[2] = a; }
    void set(T x, T y, T z) { v_[0] = x; v_[1] = y; v_[2] = z; }

    // Magnitude of the vector.
    T mag2() const { return x()*x()+y()*y()+z()*z(); }
    T mag()  const { return sqrt(mag2()); }

    // Scalar product.
    T dot(const BasicVector3D<T> & v) const {
      return x()*v.x()+y()*v.y()+z()*v.z();
    }

    // Vector product.
    BasicVector3D<T> cross(const BasicVector3D<T> & v) const {
      return BasicVector3D<T>(y()*v.z()-v.y()*z(),
			      z()*v.x()-v.z()*x(),
			      x()*v.y()-v.x()*y());
    }

    // Returns unit vector parallel to this.
    BasicVector3D<T> unit() const {
      T len = mag();
      return (len > 0.) ?
	BasicVector3D<T>(x()/len, y()/len, z()/len) : BasicVector3D<T>();
    }
  };

  /*************************************************************************
   *                                                                       *  
   * Non-member functions for BasicVector3D<float>                         *
   *                                                                       *  
   *************************************************************************/

  // Stream operators
  std::ostream &
  operator<<(std::ostream &, const BasicVector3D<float> &);

  std::istream &
  operator>>(std::istream &, BasicVector3D<float> &);

  // Arithmetic
  inline BasicVector3D<float>
  operator+(const BasicVector3D<float> & v) { return v; }

  inline BasicVector3D<float>
  operator+(const BasicVector3D<float> & a, const BasicVector3D<float> & b) {
    return BasicVector3D<float>(a.x()+b.x(), a.y()+b.y(), a.z()+b.z());
  }
  inline BasicVector3D<float>
  operator-(const BasicVector3D<float> & v) {
    return BasicVector3D<float>(-v.x(), -v.y(), -v.z());
  }
  inline BasicVector3D<float>
  operator-(const BasicVector3D<float> & a, const BasicVector3D<float> & b) {
    return BasicVector3D<float>(a.x()-b.x(), a.y()-b.y(), a.z()-b.z());
  }
  inline BasicVector3D<float>
  operator*(const BasicVector3D<float> & v, float a) {
    return BasicVector3D<float>(v.x()*a, v.y()*a, v.z()*a);
  }
  inline float
  operator*(const BasicVector3D<float> & a, const BasicVector3D<float> & b) {
    return a.dot(b);
  }
  inline BasicVector3D<float>
  operator*(float a, const BasicVector3D<float> & v) {
    return BasicVector3D<float>(a*v.x(), a*v.y(), a*v.z());
  }
  inline BasicVector3D<float>
  operator/(const BasicVector3D<float> & v, float a) {
    return BasicVector3D<float>(v.x()/a, v.y()/a, v.z()/a);
  }
  
  // Comparisons
  inline bool
  operator==(const BasicVector3D<float> & a, const BasicVector3D<float> & b) {
    return (a.x()==b.x() && a.y()==b.y() && a.z()==b.z());
  }
  inline bool
  operator!=(const BasicVector3D<float> & a, const BasicVector3D<float> & b) {
    return (a.x()!=b.x() || a.y()!=b.y() || a.z()!=b.z());
  }

  /*************************************************************************
   *                                                                       *  
   * Non-member functions for BasicVector3D<double>                        *
   *                                                                       *  
   *************************************************************************/

  // Stream operators
  std::ostream &
  operator<<(std::ostream &, const BasicVector3D<double> &);

  std::istream &
  operator>>(std::istream &, BasicVector3D<double> &);

  // Arithmetic
  inline BasicVector3D<double>
  operator+(const BasicVector3D<double> & v) { return v; }

  inline BasicVector3D<double>
  operator+(const BasicVector3D<double> & a,const BasicVector3D<double> & b) {
    return BasicVector3D<double>(a.x()+b.x(), a.y()+b.y(), a.z()+b.z());
  }
  inline BasicVector3D<double>
  operator-(const BasicVector3D<double> & v) {
    return BasicVector3D<double>(-v.x(), -v.y(), -v.z());
  }
  inline BasicVector3D<double>
  operator-(const BasicVector3D<double> & a,const BasicVector3D<double> & b) {
    return BasicVector3D<double>(a.x()-b.x(), a.y()-b.y(), a.z()-b.z());
  }
  inline BasicVector3D<double>
  operator*(const BasicVector3D<double> & v, double a) {
    return BasicVector3D<double>(v.x()*a, v.y()*a, v.z()*a);
  }
  inline double
  operator*(const BasicVector3D<double> & a,const BasicVector3D<double> & b) {
    return a.dot(b);
  }
  inline BasicVector3D<double>
  operator*(double a, const BasicVector3D<double> & v) {
    return BasicVector3D<double>(a*v.x(), a*v.y(), a*v.z());
  }
  inline BasicVector3D<double>
  operator/(const BasicVector3D<double> & v, double a) {
    return BasicVector3D<double>(v.x()/a, v.y()/a, v.z()/a);
  }
  
  // Comparisons
  inline bool
  operator==(const BasicVector3D<double> & a, const BasicVector3D<double> & b)
  {
    return (a.x()==b.x() && a.y()==b.y() && a.z()==b.z());
  }
  inline bool
  operator!=(const BasicVector3D<double> & a, const BasicVector3D<double> & b)
  {
    return (a.x()!=b.x() || a.y()!=b.y() || a.z()!=b.z());
  }
}

#endif /* GBLIB_BASICVECTOR3D_HH */
