#ifndef GBLIB_PLANE3D_H
#define GBLIB_PLANE3D_H

#include <math.h>
#include <iosfwd>
#include "Point3D.hh"
#include "Normal3D.hh"
#include "Transformation3D.hh"

namespace gblib {

  /**
   * Geometrical 3D Plane.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  template<class T>
  class Plane3D {
  private:
    T a_, b_, c_, d_;
  };

//----------------------------------------------------------------------------
  template<>
  class Plane3D<float> {
  protected:
    float a_, b_, c_, d_;
 
  public:
    // Default constructor.
    Plane3D() : a_(1.), b_(0.), c_(0.), d_(0.) {}

    // Constructor from four numbers.
    Plane3D(float a, float b, float c, float d)
      : a_(a), b_(b), c_(c), d_(d) {}

    // Constructor from normal and point.
    Plane3D(const Normal3D<float> & n, const Point3D<float> & p)
      : a_(n.x()), b_(n.y()), c_(n.z()), d_(-n*p) {}

    // Constructor from three points.
    Plane3D(const Point3D<float> & p1,
	    const Point3D<float> & p2,
	    const Point3D<float> & p3) {
      Normal3D<float> n = (p2-p1).cross(p3-p1);
      a_ = n.x(); b_ = n.y(); c_ = n.z(); d_ = -n*p1;
    }

    // Copy constructor.
    Plane3D<float>(const Plane3D<float> & p)
      : a_(p.a_), b_(p.b_), c_(p.c_), d_(p.d_) {}

    // Destructor.
    ~Plane3D() {};

    // Assignment
    Plane3D<float> & operator=(const Plane3D<float> & p) {
      a_ = p.a_; b_ = p.b_; c_ = p.c_; d_ = p.d_; return *this;
    }

    // Access functions
    float a() const { return a_; }
    float b() const { return b_; }
    float c() const { return c_; }
    float d() const { return d_; }

    // Test for equality
    bool operator==(const Plane3D<float> & p) const {
      return a() == p.a() && b() == p.b() && c() == p.c() && d() == p.d();
    }

    // Test for inequality
    bool operator != (const Plane3D<float> & p) const {
      return a() != p.a() || b() != p.b() || c() != p.c() || d() != p.d();
    }

    // Normalization.
    Plane3D<float> & normalize() {
      double ll = sqrt(a()*a() + b()*b() + c()*c());
      if (ll > 0.) { a_ /= ll; b_ /= ll; c_ /= ll, d_ /= ll; }
      return *this;
    }

    // Returns normal.
    Normal3D<float> normal() const { return Normal3D<float>(a(),b(),c()); }

    // Returns distance to the point.
    float distance(const Point3D<float> & p) const {
      return a()*p.x() + b()*p.y() + c()*p.z() + d();
    }

    // Projection of the point to the plane
    Point3D<float> point(const Point3D<float> & p) const {
      float k = distance(p)/(a()*a()+b()*b()+c()*c());
      return Point3D<float>(p.x()-a()*k, p.y()-b()*k, p.z()-c()*k);
    }

    // Projection of the origin to the plane
    Point3D<float> point() const {
      float k = -d()/(a()*a()+b()*b()+c()*c());
      return Point3D<float>(a()*k, b()*k, c()*k);
    }

    // Transformation
    Plane3D<float>& transform(const Transformation3D & m) {
      Normal3D<float> n(a(),b(),c());
      n.transform(m);
      d_ = -n*point().transform(m); a_ = n.x(); b_ = n.y(); c_ = n.z();
      return *this;
    }
  };

  std::ostream & operator<<(std::ostream &, const Plane3D<float> &);

//----------------------------------------------------------------------------
  template<>
  class Plane3D<double> {
  protected:
    double a_, b_, c_, d_;
 
  public:
    // Default constructor.
    Plane3D() : a_(1.), b_(0.), c_(0.), d_(0.) {}

    // Constructor from four numbers.
    Plane3D(double a, double b, double c, double d)
      : a_(a), b_(b), c_(c), d_(d) {}

    // Constructor from normal and point.
    Plane3D(const Normal3D<double> & n, const Point3D<double> & p)
      : a_(n.x()), b_(n.y()), c_(n.z()), d_(-n*p) {}

    // Constructor from three points.
    Plane3D(const Point3D<double> & p1,
	    const Point3D<double> & p2,
	    const Point3D<double> & p3) {
      Normal3D<double> n = (p2-p1).cross(p3-p1);
      a_ = n.x(); b_ = n.y(); c_ = n.z(); d_ = -n*p1;
    }

    // Constructor from Plane3D<float>.
    Plane3D<double>(const Plane3D<float> & p)
      : a_(p.a()), b_(p.b()), c_(p.c()), d_(p.d()) {}

    // Copy constructor.
    Plane3D<double>(const Plane3D<double> & p)
      : a_(p.a_), b_(p.b_), c_(p.c_), d_(p.d_) {}

    // Destructor.
    ~Plane3D() {};

    // Assignment from Plane3D<float>.
    Plane3D<double> & operator=(const Plane3D<float> & p) {
      a_ = p.a(); b_ = p.b(); c_ = p.c(); d_ = p.d(); return *this;
    }

    // Assignment.
    Plane3D<double> & operator=(const Plane3D<double> & p) {
      a_ = p.a_; b_ = p.b_; c_ = p.c_; d_ = p.d_; return *this;
    }

    // Access functions
    double a() const { return a_; }
    double b() const { return b_; }
    double c() const { return c_; }
    double d() const { return d_; }

    // Test for equality
    bool operator==(const Plane3D<double> & p) const {
      return a() == p.a() && b() == p.b() && c() == p.c() && d() == p.d();
    }

    // Test for inequality
    bool operator != (const Plane3D<double> & p) const {
      return a() != p.a() || b() != p.b() || c() != p.c() || d() != p.d();
    }

    // Normalization.
    Plane3D<double> & normalize() {
      double ll = sqrt(a()*a() + b()*b() + c()*c());
      if (ll > 0.) { a_ /= ll; b_ /= ll; c_ /= ll, d_ /= ll; }
      return *this;
    }

    // Returns normal.
    Normal3D<double> normal() const { return Normal3D<double>(a(),b(),c()); }

    // Returns distance to the point.
    double distance(const Point3D<double> & p) const {
      return a()*p.x() + b()*p.y() + c()*p.z() + d();
    }

    // Projection of the point to the plane
    Point3D<double> point(const Point3D<double> & p) const {
      double k = distance(p)/(a()*a()+b()*b()+c()*c());
      return Point3D<double>(p.x()-a()*k, p.y()-b()*k, p.z()-c()*k);
    }

    // Projection of the origin to the plane
    Point3D<double> point() const {
      double k = -d()/(a()*a()+b()*b()+c()*c());
      return Point3D<double>(a()*k, b()*k, c()*k);
    }

    // Transformation
    Plane3D<double>& transform(const Transformation3D &m) {
      Normal3D<double> n(a(),b(),c());
      n.transform(m);
      d_ = -n*point().transform(m); a_ = n.x(); b_ = n.y(); c_ = n.z();
      return *this;
    }
  };

  std::ostream & operator<<(std::ostream &, const Plane3D<double> &);

} /* namespace gblib */

#endif /* GBLIB_PLANE3D_H */
