#include <math.h>
#include <iostream>
#include "Projection3D.hh"

namespace gblib {
//----------------------------------------------------------------------------
  Projection3D Projection3D::inverse() const
  {
    double XX=xx(), XY=xy(), XZ=xz(), XW=xw();
    double YX=yx(), YY=yy(), YZ=yz(), YW=yw();
    double ZX=zx(), ZY=zy(), ZZ=zz(), ZW=zw();
    double WX=wx(), WY=wy(), WZ=wz(), WW=ww();

    double zz_ww_wz_zw = (ZZ*WW-WZ*ZW);
    double zy_ww_wy_zw = (ZY*WW-WY*ZW);
    double zy_wz_wy_zz = (ZY*WZ-WY*ZZ);
    double zx_ww_wx_zw = (ZX*WW-WX*ZW); 
    double zx_wz_wx_zz = (ZX*WZ-WX*ZZ);
    double zx_wy_wx_zy = (ZX*WY-WX*ZY);

    double detxx = YY*(zz_ww_wz_zw) - YZ*(zy_ww_wy_zw) + YW*(zy_wz_wy_zz);
    double detxy = YX*(zz_ww_wz_zw) - YZ*(zx_ww_wx_zw) + YW*(zx_wz_wx_zz);
    double detxz = YX*(zy_ww_wy_zw) - YY*(zx_ww_wx_zw) + YW*(zx_wy_wx_zy);
    double detxw = YX*(zy_wz_wy_zz) - YY*(zx_wz_wx_zz) + YZ*(zx_wy_wx_zy);

    double detyx = XY*(zz_ww_wz_zw) - XZ*(zy_ww_wy_zw) + XW*(zy_wz_wy_zz);
    double detyy = XX*(zz_ww_wz_zw) - XZ*(zx_ww_wx_zw) + XW*(zx_wz_wx_zz);
    double detyz = XX*(zy_ww_wy_zw) - XY*(zx_ww_wx_zw) + XW*(zx_wy_wx_zy);
    double detyw = XX*(zy_wz_wy_zz) - XY*(zx_wz_wx_zz) + XZ*(zx_wy_wx_zy);

    double det   = XX*detxx - XY*detxy + XZ*detxz - XW*detxw;
    if (det == 0.) {
      std::cerr
	<< "Projection3D::inverse() : zero determinant"
	<< std::endl;
      return Projection3D();
    }

    double yy_xx_xy_yx = (YY*XX-XY*YX);
    double yz_xx_xz_yx = (YZ*XX-XZ*YX);
    double yz_xy_xz_yy = (YZ*XY-XZ*YY);
    double yw_xx_xw_yx = (YW*XX-XW*YX); 
    double yw_xy_xw_yy = (YW*XY-XW*YY);
    double yw_xz_xw_yz = (YW*XZ-XW*YZ);

    double detzx = WW*(yz_xy_xz_yy) - WZ*(yw_xy_xw_yy) + WY*(yw_xz_xw_yz);
    double detzy = WW*(yz_xx_xz_yx) - WZ*(yw_xx_xw_yx) + WX*(yw_xz_xw_yz);
    double detzz = WW*(yy_xx_xy_yx) - WY*(yw_xx_xw_yx) + WX*(yw_xy_xw_yy);
    double detzw = WZ*(yy_xx_xy_yx) - WY*(yz_xx_xz_yx) + WX*(yz_xy_xz_yy);

    double detwx = ZW*(yz_xy_xz_yy) - ZZ*(yw_xy_xw_yy) + ZY*(yw_xz_xw_yz);
    double detwy = ZW*(yz_xx_xz_yx) - ZZ*(yw_xx_xw_yx) + ZX*(yw_xz_xw_yz);
    double detwz = ZW*(yy_xx_xy_yx) - ZY*(yw_xx_xw_yx) + ZX*(yw_xy_xw_yy);
    double detww = ZZ*(yy_xx_xy_yx) - ZY*(yz_xx_xz_yx) + ZX*(yz_xy_xz_yy);

    return Projection3D(detxx/det, -detyx/det,  detzx/det, -detwx/det,
                       -detxy/det,  detyy/det, -detzy/det,  detwy/det,
                        detxz/det, -detyz/det,  detzz/det, -detwz/det,
                       -detxw/det,  detyw/det, -detzw/det,  detww/det);
  }

//----------------------------------------------------------------------------
  Projection3D & Projection3D::transform(const Translation3D & m)
  {
    return set
      (xx()+wx()*m.dx(), xy()+wy()*m.dx(), xz()+wz()*m.dx(), xw()+ww()*m.dx(),
       yx()+wx()*m.dy(), yy()+wy()*m.dy(), yz()+wz()*m.dy(), yw()+ww()*m.dy(),
       zx()+wx()*m.dz(), zy()+wy()*m.dz(), zz()+wz()*m.dz(), zw()+ww()*m.dz(),
       wx(),             wy(),             wz(),             ww()           );
  }

//----------------------------------------------------------------------------
  Projection3D & Projection3D::transform(const Rotation3D & m)
  {
    return set(xx()*m.xx()+yx()*m.xy()+zx()*m.xz(),
	       xy()*m.xx()+yy()*m.xy()+zy()*m.xz(),
	       xz()*m.xx()+yz()*m.xy()+zz()*m.xz(),
	       xw()*m.xx()+yw()*m.xy()+zw()*m.xz(),
	       xx()*m.yx()+yx()*m.yy()+zx()*m.yz(),
	       xy()*m.yx()+yy()*m.yy()+zy()*m.yz(),
	       xz()*m.yx()+yz()*m.yy()+zz()*m.yz(),
	       xw()*m.yx()+yw()*m.yy()+zw()*m.yz(),
	       xx()*m.zx()+yx()*m.zy()+zx()*m.zz(),
	       xy()*m.zx()+yy()*m.zy()+zy()*m.zz(),
	       xz()*m.zx()+yz()*m.zy()+zz()*m.zz(),
	       xw()*m.zx()+yw()*m.zy()+zw()*m.zz(),
	       wx(), wy(), wz(), ww());
  }

//----------------------------------------------------------------------------
  Projection3D & Projection3D::transform(const Transformation3D & m)
  {
    return set(xx()*m.xx()+yx()*m.xy()+zx()*m.xz()+wx()*m.dx(),
	       xy()*m.xx()+yy()*m.xy()+zy()*m.xz()+wy()*m.dx(),
	       xz()*m.xx()+yz()*m.xy()+zz()*m.xz()+wz()*m.dx(),
	       xw()*m.xx()+yw()*m.xy()+zw()*m.xz()+ww()*m.dx(),
	       xx()*m.yx()+yx()*m.yy()+zx()*m.yz()+wx()*m.dy(),
	       xy()*m.yx()+yy()*m.yy()+zy()*m.yz()+wy()*m.dy(),
	       xz()*m.yx()+yz()*m.yy()+zz()*m.yz()+wz()*m.dy(),
	       xw()*m.yx()+yw()*m.yy()+zw()*m.yz()+ww()*m.dy(),
	       xx()*m.zx()+yx()*m.zy()+zx()*m.zz()+wx()*m.dz(),
	       xy()*m.zx()+yy()*m.zy()+zy()*m.zz()+wy()*m.dz(),
	       xz()*m.zx()+yz()*m.zy()+zz()*m.zz()+wz()*m.dz(),
	       xw()*m.zx()+yw()*m.zy()+zw()*m.zz()+ww()*m.dz(),
	       wx(), wy(), wz(), ww());
  }

//----------------------------------------------------------------------------
  Projection3D & Projection3D::transform(const Projection3D & m)
  {
    return set(xx()*m.xx()+yx()*m.xy()+zx()*m.xz()+wx()*m.xw(),
	       xy()*m.xx()+yy()*m.xy()+zy()*m.xz()+wy()*m.xw(),
	       xz()*m.xx()+yz()*m.xy()+zz()*m.xz()+wz()*m.xw(),
	       xw()*m.xx()+yw()*m.xy()+zw()*m.xz()+ww()*m.xw(),
	       xx()*m.yx()+yx()*m.yy()+zx()*m.yz()+wx()*m.yw(),
	       xy()*m.yx()+yy()*m.yy()+zy()*m.yz()+wy()*m.yw(),
	       xz()*m.yx()+yz()*m.yy()+zz()*m.yz()+wz()*m.yw(),
	       xw()*m.yx()+yw()*m.yy()+zw()*m.yz()+ww()*m.yw(),
	       xx()*m.zx()+yx()*m.zy()+zx()*m.zz()+wx()*m.zw(),
	       xy()*m.zx()+yy()*m.zy()+zy()*m.zz()+wy()*m.zw(),
	       xz()*m.zx()+yz()*m.zy()+zz()*m.zz()+wz()*m.zw(),
	       xw()*m.zx()+yw()*m.zy()+zw()*m.zz()+ww()*m.zw(),
	       xx()*m.wx()+yx()*m.wy()+zx()*m.wz()+wx()*m.ww(),
	       xy()*m.wx()+yy()*m.wy()+zy()*m.wz()+wy()*m.ww(),
	       xz()*m.wx()+yz()*m.wy()+zz()*m.wz()+wz()*m.ww(),
	       xw()*m.wx()+yw()*m.wy()+zw()*m.wz()+ww()*m.ww());
  }
} /* namespace gblib */
