#ifndef GBLIB_PROJECTION3D_HH
#define GBLIB_PROJECTION3D_HH

#include "Transformation3D.hh"
#include "Point3D.hh"

namespace gblib {

  /**
   * Class for projection transformation of 3D objects.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  class Projection3D : public Transformation3D {
  protected:
    double w_[4];

  protected:
    // Protected constructor.
    Projection3D(double xx, double xy, double xz, double dx,
		 double yx, double yy, double yz, double dy,
		 double zx, double zy, double zz, double dz,
		 double wx, double wy, double wz, double ww)
      : Transformation3D(xx,xy,xz,dx, yx,yy,yz,dy, zx,zy,zz,dz) {
      w_[0] = wx; w_[1] = wy; w_[2] = wz; w_[3] = ww;
    }

    // Sets matrix elements.
    Projection3D & set(double xx, double xy, double xz, double dx,
		       double yx, double yy, double yz, double dy,
		       double zx, double zy, double zz, double dz,
		       double wx, double wy, double wz, double ww) {
      Transformation3D::set(xx,xy,xz,dx, yx,yy,yz,dy, zx,zy,zz,dz);
      w_[0] = wx; w_[1] = wy; w_[2] = wz; w_[3] = ww;
      return *this;
    }

    // Sets elements in the last row of the matrix.
    Projection3D & setLastRow() {
      w_[0] = 0.; w_[1] = 0.; w_[2] = 0.; w_[3] = 1.; return *this;
    }

  public:
    /**
     * Default constructor. Sets identity transformation.
     */
    Projection3D() { setLastRow(); }

    /**
     * Constructor from Translation3D.
     */
    Projection3D(const Translation3D & t)
      : Transformation3D(t) { setLastRow(); }

    /**
     * Constructor from Rotation3D.
     */
    Projection3D(const Rotation3D & r)
      : Transformation3D(r) { setLastRow(); }

    /**
     * Constructor from Rotation3D and Translation3D.
     */
    Projection3D(const Rotation3D & r, const Translation3D & t)
      : Transformation3D(r,t) { setLastRow(); }

    /**
     * Constructor from Transformation3D.
     */
    Projection3D(const Transformation3D & m)
      : Transformation3D(m) { setLastRow(); }

    /**
     * Copy constructor.
     */
    Projection3D(const Projection3D & m) : Transformation3D(m) {
      w_[0] = m.w_[0]; w_[1] = m.w_[1]; w_[2] = m.w_[2]; w_[3] = m.w_[3];
    }

    /**
     * Destructor.
     */
    ~Projection3D() {}

    /**
     * Assignment.
     */
    Projection3D & operator=(const Projection3D & m) {
      Transformation3D::operator=(m);
      w_[0] = m.w_[0]; w_[1] = m.w_[1]; w_[2] = m.w_[2]; w_[3] = m.w_[3];
      return *this;
    }
			       
    /**
     * Get elements of the transformation matrix
     */
    double xx() const { return Transformation3D::xx(); }
    double xy() const { return Transformation3D::xy(); }
    double xz() const { return Transformation3D::xz(); }
    double xw() const { return Transformation3D::dx(); }
    double yx() const { return Transformation3D::yx(); }
    double yy() const { return Transformation3D::yy(); }
    double yz() const { return Transformation3D::yz(); }
    double yw() const { return Transformation3D::dy(); }
    double zx() const { return Transformation3D::zx(); }
    double zy() const { return Transformation3D::zy(); }
    double zz() const { return Transformation3D::zz(); }
    double zw() const { return Transformation3D::dz(); }
    double wx() const { return w_[0]; }
    double wy() const { return w_[1]; }
    double wz() const { return w_[2]; }
    double ww() const { return w_[3]; }
    double dx() const { return Transformation3D::dx(); }
    double dy() const { return Transformation3D::dy(); }
    double dz() const { return Transformation3D::dz(); }
    
    /**
     * Returns inverse transformation.
     */
    Projection3D inverse() const;

    /**
     * Inverts current transformation.
     */
    Projection3D & invert() { *this = this->inverse(); return *this; }

    /**
     * Checks whether current transformation is the identity transformation
     * or not.
     */
    bool isIdentity() const {
      return (Transformation3D::isIdentity() &&
	      w_[0] == 0. && w_[1] == 0. &&  w_[2] == 0. && w_[3] == 1.);
    }

    /**
     * Sets the identity transformation.
     */
    Projection3D & setIdentity() {
      Transformation3D::setIdentity(); return setLastRow();
    } 

    /**
     * Sets translation along x-axis.
     */
    Projection3D & setTranslationX(double x) {
      Transformation3D::setTranslationX(x); return setLastRow();
    }

    /**
     * Sets translation along y-axis.
     */
    Projection3D & setTranslationY(double y) {
      Transformation3D::setTranslationY(y); return setLastRow();
    }

    /**
     * Sets translation along z-axis.
     */
    Projection3D & setTranslationZ(double z) {
      Transformation3D::setTranslationZ(z); return setLastRow();
    }

    /**
     * Sets translation on (x,y,z).
     */
    Projection3D & setTranslation(double x, double y, double z) {
      Transformation3D::setTranslation(x, y, z); return setLastRow();
    }

    /**
     * Sets rotation around x-axis.
     */
    Projection3D & setRotationX(double a) {
      Transformation3D::setRotationX(a); return setLastRow();
    } 
    /**
     * Sets rotation around y-axis.
     */
    Projection3D & setRotationY(double a) {
      Transformation3D::setRotationY(a); return setLastRow();
    } 
    /**
     * Sets rotation around z-axis.
     */
    Projection3D & setRotationZ(double a) {
      Transformation3D::setRotationZ(a); return setLastRow();
    } 
    Projection3D & setRotationXYZ(double a, double b, double c) {
      Transformation3D::setRotationXYZ(a, b, c); return setLastRow();
    } 
    Projection3D & setRotationEulerZXZ(double a, double b, double c) {
      Transformation3D::setRotationEulerZXZ(a, b, c); return setLastRow();
    } 
    Projection3D & setRotationEulerZYZ(double a, double b, double c) {
      Transformation3D::setRotationEulerZYZ(a, b, c); return setLastRow();
    } 
    Projection3D &
    setRotationAngleAxis(double a, const BasicVector3D<double> & v) {
      Transformation3D::setRotationAngleAxis(a, v); return setLastRow();
    } 
    Projection3D &
    setRotationAxes(const BasicVector3D<double> & newX,
		    const BasicVector3D<double> & newY,
		    const BasicVector3D<double> & newZ) {
      Transformation3D::setRotationAxes(newX, newY, newZ);
      return setLastRow();
    }  
    Projection3D &
    setRotationPoints(const Point3D<double> & fr1,
		      const Point3D<double> & fr2,
		      const Point3D<double> & to1,
		      const Point3D<double> & to2) {
      Transformation3D::setRotationPoints(fr1,fr2, to1,to2);
      return setLastRow();
    }  

    // Set reflections.
    Projection3D & setReflectionX(double x=0.) {
      Transformation3D::setReflectionX(x); return setLastRow();
    }
    Projection3D & setReflectionY(double y=0.) {
      Transformation3D::setReflectionY(y); return setLastRow();
    }
    Projection3D & setReflectionZ(double z=0.) {
      Transformation3D::setReflectionZ(z); return setLastRow();
    }
    Projection3D & setReflection(double a, double b, double c, double d) {
      Transformation3D::setReflection(a,b,c,d); return setLastRow();
    }

    // Set scales
    Projection3D & setScaleX(double x) {
      Transformation3D::setScaleX(x); return setLastRow();
    }
    Projection3D & setScaleY(double y) {
      Transformation3D::setScaleY(y); return setLastRow();
    }
    Projection3D & setScaleZ(double z) {
      Transformation3D::setScaleZ(z); return setLastRow();
    }
    Projection3D & setScale(double a) {
      Transformation3D::setScale(a); return setLastRow();
    } 
    Projection3D & setScale(double x, double y, double z) {
      Transformation3D::setScale(x, y, z); return setLastRow();
    }

    // Set transformation.
    Projection3D &
    setTransformationPoints(const Point3D<double> & fr0,
			    const Point3D<double> & fr1,
			    const Point3D<double> & fr2,
			    const Point3D<double> & to0,
			    const Point3D<double> & to1,
			    const Point3D<double> & to2) {
      Transformation3D::setTransformationPoints(fr0,fr1,fr2, to0,to1,to2);
      return setLastRow();
    }

    /**
     * Sets transformation for perspective-view frustum.
     *
     * @param l left
     * @param r right
     * @param b bottom
     * @param t top
     * @param n near
     * @param f far
     */
    Projection3D & setFrustum(double l, double r,
			      double b, double t,
			      double n, double f) {
      return set(2*n/(r-l),    0,      (r+l)/(r-l),      0,
		 0,         2*n/(t-b), (t+b)/(t-b),      0,
		 0,            0,      (n+f)/(n-f), 2*n*f/(n-f),
		 0,            0,          -1,           0);
    }

    // Transformations
    Projection3D & transform(const Translation3D & m);
    Projection3D & transform(const Rotation3D & m);
    Projection3D & transform(const Transformation3D & m);
    Projection3D & transform(const Projection3D & m);
  };

  // Composite transformations
  inline Projection3D
  operator*(const Translation3D & a, const Projection3D & b) {
    return Projection3D(b).transform(a);
  }
  inline Projection3D
  operator*(const Rotation3D & a, const Projection3D & b) {
    return Projection3D(b).transform(a);
  }
  inline Projection3D
  operator*(const Transformation3D & a, const Projection3D & b) {
    return Projection3D(b).transform(a);
  }
  inline Projection3D
  operator*(const Projection3D & a, const Translation3D & b) {
    return Projection3D(b).transform(a);
  }
  inline Projection3D
  operator*(const Projection3D & a, const Rotation3D & b) {
    return Projection3D(b).transform(a);
  }
  inline Projection3D
  operator*(const Projection3D & a, const Transformation3D & b) {
    return Projection3D(b).transform(a);
  }
  inline Projection3D
  operator*(const Projection3D & a, const Projection3D & b) {
    return Projection3D(b).transform(a);
  }

  /**
   * Transforms Point3D<double>. 
   */
  inline Point3D<double>
  operator*(const Projection3D & m, const Point3D<double> & p) {
    double x = m.xx()*p.x() + m.xy()*p.y() + m.xz()*p.z() + m.dx();
    double y = m.yx()*p.x() + m.yy()*p.y() + m.yz()*p.z() + m.dy();
    double z = m.zx()*p.x() + m.zy()*p.y() + m.zz()*p.z() + m.dz();
    double w = m.wx()*p.x() + m.wy()*p.y() + m.wz()*p.z() + m.ww();
    return Point3D<double>(x/w, y/w, z/w);
  }

  /**
   * Transforms Point3D<float>. 
   */
  inline Point3D<float>
  operator*(const Projection3D & m, const Point3D<float> & p) {
    double x = m.xx()*p.x() + m.xy()*p.y() + m.xz()*p.z() + m.dx();
    double y = m.yx()*p.x() + m.yy()*p.y() + m.yz()*p.z() + m.dy();
    double z = m.zx()*p.x() + m.zy()*p.y() + m.zz()*p.z() + m.dz();
    double w = m.wx()*p.x() + m.wy()*p.y() + m.wz()*p.z() + m.ww();
    return Point3D<float>(float(x/w), float(y/w), float(z/w));
  }
} /* namespace gblib */

#endif /* GBLIB_PROJECTION3D_HH */
