#include <math.h>
#include <iostream>
#include "Point2D.hh"
#include "Rotation2D.hh"

template <class T>
inline T abs(const T & a) {
  return a < 0 ? -a : a;
}

namespace gblib {

  typedef BasicVector2D<double> vector;
  typedef Point2D<double>       point;

//----------------------------------------------------------------------------
  Rotation2D::Rotation2D(double a)
      : xrow_(cos(a),-sin(a)), yrow_(sin(a),cos(a)) {}

//----------------------------------------------------------------------------
  bool Rotation2D::isRotation(double tolerance) const
  {
    return (!(abs(xx()*yy()-yx()*xy()-1.) > tolerance ||  // det(m)    == 1
	      abs(xx()*xx()+xy()*xy()-1.) > tolerance ||  // abs(xrow) == 1
	      abs(yx()*yx()+yy()*yy()-1.) > tolerance ||  // abs(yrow) == 1
	      abs(xx()*yx()+xy()*yy())    > tolerance));  // xrow*yrow == 0
  }
  
//----------------------------------------------------------------------------
  Rotation2D & Rotation2D::setRotation(double a)
  {
    double cosa = cos(a);
    double sina = sin(a);
    return set(cosa,-sina, sina, cosa);
  }

//----------------------------------------------------------------------------
  Rotation2D & Rotation2D::setRotationAxes(const vector & newX,
					   const vector & newY)
  {
    Rotation2D m(newX.x(), newY.x(), newX.y(), newY.y());
    if (m.isRotation()) {
      return *this = m;
    }else{  
      std::cerr
	<< "Rotation2D::setRotationAxes() : bad axis vectors"
	<< std::endl;
      return *this;
    }
  }
  
//----------------------------------------------------------------------------
  Rotation2D & Rotation2D::setRotationPoints(const point & fr,
					     const point & to)
  {
    double tolerance = 0.000001;

    //   C H E C K   D I S T A N C E S   F R O M   T H E   O R I G I N

    if (abs(fr.mag2()-to.mag2()) > tolerance) {
      std::cerr
	<< "Rotation2D::setRotationPoints() : "
	<< "distances from the origin are different"
	<< std::endl;
    return setIdentity();
  }  
    if (fr.mag2() <= tolerance) {
      std::cerr
	<< "Rotation2D::setRotationPoints() : "
	<< "points are too close to the origin"
	<< std::endl;
      return setIdentity();
  }  

  //   S E T  R O T A T I O N

  vector a = fr.unit();
  vector b = to.unit();
  double cosa = a.x()*b.x()+a.y()*b.y();
  double sina = a.y()*b.x()-a.x()*b.y();
  return set(cosa, sina, -sina, cosa);
}

//----------------------------------------------------------------------------
  double Rotation2D::getAngle() const { return atan2(xx(), yx()); }

//----------------------------------------------------------------------------
  Rotation2D & Rotation2D::transform(const Rotation2D & m)
  {
    return set(m.xx()*xx() + m.xy()*yx(), m.xx()*xy() + m.xy()*yy(),
	       m.yx()*xx() + m.yy()*yx(), m.yx()*xy() + m.yy()*yy());
  }

} /* namespace gblib */
