#ifndef GBLIB_ROTATION2D_HH
#define GBLIB_ROTATION2D_HH

#include "BasicVector2D.hh"

namespace gblib {

  template<class T> class Point2D;

  /**
   * Class for rotation of 2D objects.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  class Rotation2D {
  protected:
    BasicVector2D<double> xrow_;
    BasicVector2D<double> yrow_;

    // Protected constructor.
    Rotation2D (double xx, double xy, double yx, double yy)
      : xrow_(xx, xy), yrow_(yx, yy) {}

    // Set elements of the rotation matrix.
    Rotation2D & set(double xx, double xy, double yx, double yy) {
      xrow_.set(xx, xy); yrow_.set(yx, yy); return *this;
    }

  public:
    // Default constructor.
    Rotation2D() : xrow_(1.,0.), yrow_(0.,1.) {}

    // Constructor from a double.
    explicit Rotation2D(double);

    // Copy constructor.
    Rotation2D(const Rotation2D & m) : xrow_(m.xrow_), yrow_(m.yrow_) {}

    // Destructor.
    ~Rotation2D() {}

    // Assignment.
    Rotation2D & operator=(const Rotation2D & m) {
      xrow_ = m.xrow_; yrow_ = m.yrow_; return *this;
    }

    // Get elements of the rotation matrix
    double xx() const { return xrow_.x(); }
    double xy() const { return xrow_.y(); }
    double yx() const { return yrow_.x(); }
    double yy() const { return yrow_.y(); }

    // Checkings.
    bool isIdentity() const {
      return (xx() == 1. && xy() == 0. && yx() == 0. && yy() == 1.);
    }
    bool isRotation(double tolerance = 1.e-5) const;

    // Invert.
    Rotation2D inverse() const { return Rotation2D(xx(), yx(), xy(), yy()); }
    Rotation2D & invert() { return set(xx(), yx(), xy(), yy()); }

    // Set the Identity rotation.
    Rotation2D & setIdentity() { set(1.,0., 0.,1.); return *this; } 

    // Set rotation.
    Rotation2D & setRotation(double);

    // Set rotation defined by positions of rotated axes. 
    Rotation2D &
    setRotationAxes(const BasicVector2D<double> & newX,
		    const BasicVector2D<double> & newY);

    // Set rotation defined by original and rotated positions of a point.
    Rotation2D &
    setRotationPoints(const Point2D<double> & fr, const Point2D<double> & to);

    // Get the rotation angle.
    double getAngle() const;

    // Transformation.
    Rotation2D & transform(const Rotation2D &);
  };

  // Composite rotation.
  inline Rotation2D
  operator*(const Rotation2D & a, const Rotation2D & b) {
    Rotation2D m(b);
    return m.transform(a);
  }

} /* namespace gblib */

#endif /* GBLIB_ROTATION2D_HH */
