#include <math.h>
#include <float.h>
#include <iostream>
#include "Point3D.hh"
#include "Rotation3D.hh"

template <class T>
inline T abs(const T & a) {
  return a < 0 ? -a : a;
}

#define SET_RESULT \
set(m.xx()*xx()+m.xy()*yx()+m.xz()*zx(), \
    m.xx()*xy()+m.xy()*yy()+m.xz()*zy(), \
    m.xx()*xz()+m.xy()*yz()+m.xz()*zz(), \
    m.yx()*xx()+m.yy()*yx()+m.yz()*zx(), \
    m.yx()*xy()+m.yy()*yy()+m.yz()*zy(), \
    m.yx()*xz()+m.yy()*yz()+m.yz()*zz(), \
    m.zx()*xx()+m.zy()*yx()+m.zz()*zx(), \
    m.zx()*xy()+m.zy()*yy()+m.zz()*zy(), \
    m.zx()*xz()+m.zy()*yz()+m.zz()*zz())

namespace gblib {

  typedef BasicVector3D<double> vector;
  typedef Point3D<double>       point;

//----------------------------------------------------------------------------
  bool
  Rotation3D::isRotation(double tolerance) const
  {
    vector x(xx(),xy(),xz());
    vector y(yx(),yy(),yz());
    vector z(zx(),zy(),zz());
    vector w = x.cross(y);
    
    return (!(abs(z.x()-w.x()) > tolerance ||
	      abs(z.y()-w.y()) > tolerance ||
	      abs(z.z()-w.z()) > tolerance ||
	      abs(x.mag2()-1.) > tolerance ||
	      abs(y.mag2()-1.) > tolerance || 
	      abs(z.mag2()-1.) > tolerance ||
	      abs(x.dot(y))    > tolerance ||
	      abs(y.dot(z))    > tolerance ||
	      abs(z.dot(x))    > tolerance));
  }
  
//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationX(double a)
  {
    double cosa = cos(a);
    double sina = sin(a);
    return set(1.,    0.,    0.,
	       0.,  cosa, -sina,
	       0.,  sina,  cosa);
  }
  
//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationY(double a)
  {
    double cosa = cos(a);
    double sina = sin(a);
    return set(cosa,  0., sina,
	       0.,    1.,   0.,
	       -sina, 0., cosa);
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationZ(double a)
  {
    double cosa = cos(a);
    double sina = sin(a);
    return set(cosa,-sina, 0.,
	       sina, cosa, 0.,
	       0.,     0., 1.); 
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationXYZ(double a, double b, double c)
  {
    double c1 = cos(a), c2 = cos(b), c3 = cos(c);
    double s1 = sin(a), s2 = sin(b), s3 = sin(c);
    return set(c2*c3, s1*s2*c3-c1*s3, c1*s2*c3+s1*s3, 
	       c2*s3, s1*s2*s3+c1*c3, c1*s2*s3-s1*c3, 
	      -s2,    s1*c2,          c1*c2);
  }
  
//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationEulerZXZ(double a, double b, double c) {
    double c1 = cos(a), c2 = cos(b), c3 = cos(c);
    double s1 = sin(a), s2 = sin(b), s3 = sin(c);
    return set(-s1*c2*s3+c1*c3, -s1*c2*c3-c1*s3,  s1*s2, 
	        c1*c2*s3+s1*c3,  c1*c2*c3-s1*s3, -c1*s2, 
	           s2*s3,           s2*c3,           c2);
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationEulerZYZ(double a, double b, double c) {
    double c1 = cos(a), c2 = cos(b), c3 = cos(c);
    double s1 = sin(a), s2 = sin(b), s3 = sin(c);
    return set(c1*c2*c3-s1*s3, -c1*c2*s3-s1*c3, c1*s2, 
	       s1*c2*c3+c1*s3, -s1*c2*s3+c1*c3, s1*s2, 
	         -s2*c3,           s2*s3,          c2);
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationAngleAxis(double a, const vector & axis)
  {
    if (a != 0.) {
      double ll = axis.mag();
      if (ll == 0.) {
	std::cerr
	  << "Rotation3D::setRotationAngleAxis() : zero axis"
	  << std::endl;
	return *this;
      }else{
	double sa = sin(a), ca = cos(a);
	double dx = axis.x()/ll, dy = axis.y()/ll, dz = axis.z()/ll;   
	return set(
          ca+(1-ca)*dx*dx,          (1-ca)*dx*dy-sa*dz,    (1-ca)*dx*dz+sa*dy,
	     (1-ca)*dy*dx+sa*dz, ca+(1-ca)*dy*dy,          (1-ca)*dy*dz-sa*dx,
	     (1-ca)*dz*dx-sa*dy,    (1-ca)*dz*dy+sa*dx, ca+(1-ca)*dz*dz);
      }
    }else{
      return setIdentity();
    }
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationAxes(const vector & newX,
					   const vector & newY,
					   const vector & newZ)
  {
    Rotation3D m(newX.x(), newY.x(), newZ.x(),
		 newX.y(), newY.y(), newZ.y(),
		 newX.z(), newY.z(), newZ.z());
    if (m.isRotation()) {
      return *this = m;
    }else{  
      std::cerr
	<< "Rotation3D::setRotationAxes() : bad axis vectors"
	<< std::endl;
      return *this;
    }
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::setRotationPoints(const point & fr1,
					     const point & fr2,
					     const point & to1,
					     const point & to2)
  {
    double tolerance = 0.000001;

    //   C H E C K   D I S T A N C E S   F R O M   T H E   O R I G I N

    if (abs(fr1.mag2()-to1.mag2()) > tolerance ||
	abs(fr2.mag2()-to2.mag2()) > tolerance ) {
      std::cerr
	<< "Rotation3D::setRotationPoints() : "
	<< "distances from the origin are different"
	<< std::endl;
      return setIdentity();
    }  
    
    //   C H E C K   A N G L E S
    
    vector x1 = vector(fr1).unit();
    vector y1 = vector(fr2).unit();
    vector x2 = vector(to1).unit();
    vector y2 = vector(to2).unit();
    double cos1 = x1.dot(y1);
    double cos2 = x2.dot(y2);
    
    if (abs(1.0-cos1) < tolerance || abs(1.0-cos2) < tolerance) {
      std::cerr
	<< "Rotation3D::setRotationPoints() : "
	<< "points lay on one line with the origin"
	<< std::endl;
      return setIdentity();
    }  
    
    if (abs(cos1-cos2) > tolerance) {
      std::cerr
	<< "Rotation3D::setRotationPoints() : "
	<< "angles between points are not equal"
	<< std::endl;
      return setIdentity();
    }
    
    //   G E N E R A T E   B A S I S   V E C T O R S
    
    vector z1 = vector(x1.cross(y1)).unit();
    vector z2 = vector(x2.cross(y2)).unit();
    y1  = z1.cross(x1);
    y2  = z2.cross(x2);
    
    //   S E T   R O T A T I O N   M A T R I X :
    //   | x2.x() y2.x() z2.x() |   | x1.x() x1.y() x1.z() |
    //   | x2.y() y2.y() z2.y() | x | y1.x() y1.y() y1.z() |
    //   | x2.z() y2.z() z2.z() |   | z1.x() z1.y() z1.z() |
    
    return set(x2.x()*x1.x() + y2.x()*y1.x() + z2.x()*z1.x(),
	       x2.x()*x1.y() + y2.x()*y1.y() + z2.x()*z1.y(),
	       x2.x()*x1.z() + y2.x()*y1.z() + z2.x()*z1.z(),
	       x2.y()*x1.x() + y2.y()*y1.x() + z2.y()*z1.x(),
	       x2.y()*x1.y() + y2.y()*y1.y() + z2.y()*z1.y(),
	       x2.y()*x1.z() + y2.y()*y1.z() + z2.y()*z1.z(),
	       x2.z()*x1.x() + y2.z()*y1.x() + z2.z()*z1.x(),
	       x2.z()*x1.y() + y2.z()*y1.y() + z2.z()*z1.y(),
	       x2.z()*x1.z() + y2.z()*y1.z() + z2.z()*z1.z());
  }

//----------------------------------------------------------------------------
  void Rotation3D::getAngleAxis(double & angle, vector & axis) const
  {
    double cosa  = 0.5*(xx()+yy()+zz()-1);
    double cosa1 = 1. - cosa;
    if (cosa1 <= 0.) {
      angle = 0.;
      axis.set(0., 0., 1.);
    }else{
      double x = 0., y = 0., z = 0.;
      if (xx() > cosa) x = sqrt((xx()-cosa)/cosa1);
      if (yy() > cosa) y = sqrt((yy()-cosa)/cosa1);
      if (zz() > cosa) z = sqrt((zz()-cosa)/cosa1);
      if (zy() < yz()) x = -x;
      if (xz() < zx()) y = -y;
      if (yx() < xy()) z = -z;
      angle = (cosa < -1.) ? acos(-1.) : acos(cosa);
      axis.set(x, y, z);
    }
  }

//----------------------------------------------------------------------------
  void Rotation3D::getXYZ(double & a, double & b, double & c) const
  {
    double cosb = sqrt(xx()*xx() + yx()*yx()); 
    if (cosb > 16*FLT_EPSILON) {
      a = atan2( zy(), zz());
      b = atan2(-zx(), cosb);
      c = atan2( yx(), xx());
    }else{
      a = atan2(-yz(), yy());
      b = atan2(-zx(), cosb);
      c = 0.;
    }
  }

//----------------------------------------------------------------------------
  void Rotation3D::getEulerZXZ(double & a, double & b, double & c) const
  {
    double sinb = sqrt(zx()*zx() + zy()*zy()); 
    if (sinb > 16*FLT_EPSILON) {
      c = atan2( zx(), zy());
      b = atan2( sinb, zz());
      a = atan2( xz(),-yz());
    }else{
      c = atan2(-xy(), xx());
      b = atan2( sinb, zz());
      a = 0;
    }
  }

//----------------------------------------------------------------------------
  void Rotation3D::getEulerZYZ(double & a , double & b, double & c) const
  {
    double sinb = sqrt(zy()*zy() + zx()*zx()); 
    if (sinb > 16*FLT_EPSILON) {
      c = atan2( zy(),-zx());
      b = atan2( sinb, zz());
      a = atan2( yz(), xz());
    }else{
      c = atan2( yx(), yy());
      b = atan2( sinb, zz());
      a = 0;
    }
  }

//----------------------------------------------------------------------------
  Rotation3D & Rotation3D::transform(const Rotation3D & m) {
    return set(m.xx()*xx()+m.xy()*yx()+m.xz()*zx(),
	       m.xx()*xy()+m.xy()*yy()+m.xz()*zy(),
	       m.xx()*xz()+m.xy()*yz()+m.xz()*zz(),
	       m.yx()*xx()+m.yy()*yx()+m.yz()*zx(),
	       m.yx()*xy()+m.yy()*yy()+m.yz()*zy(),
	       m.yx()*xz()+m.yy()*yz()+m.yz()*zz(),
	       m.zx()*xx()+m.zy()*yx()+m.zz()*zx(),
	       m.zx()*xy()+m.zy()*yy()+m.zz()*zy(),
	       m.zx()*xz()+m.zy()*yz()+m.zz()*zz());
  }

} /* namespace gblib */
