#ifndef GBLIB_ROTATION3D_HH
#define GBLIB_ROTATION3D_HH

#include "BasicVector3D.hh"

namespace gblib {

  template<class T> class Point3D;

  /**
   * Class for rotation of 3D objects.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  class Rotation3D {
  protected:
    BasicVector3D<double> xrow_;
    BasicVector3D<double> yrow_;
    BasicVector3D<double> zrow_;

    // Protected constructor.
    Rotation3D (double xx, double xy, double xz,
		double yx, double yy, double yz,
		double zx, double zy, double zz)
      : xrow_(xx, xy, xz), yrow_(yx, yy, yz), zrow_(zx, zy, zz) {}

    // Set elements of the rotation matrix.
    Rotation3D & set(double xx, double xy, double xz,
		     double yx, double yy, double yz,
		     double zx, double zy, double zz)
    {
      xrow_.set(xx, xy, xz);
      yrow_.set(yx, yy, yz);
      zrow_.set(zx, zy, zz);
      return *this;
    }

  public:
    // Default constructor.
    Rotation3D() : xrow_(1.,0.,0.), yrow_(0.,1.,0.), zrow_(0.,0.,1.) {}

    // Copy constructor.
    Rotation3D(const Rotation3D & m)
      : xrow_(m.xrow_), yrow_(m.yrow_), zrow_(m.zrow_) {}

    // Destructor.
    ~Rotation3D() {}

    // Assignment.
    Rotation3D & operator=(const Rotation3D & m) {
      xrow_ = m.xrow_; yrow_ = m.yrow_; zrow_ = m.zrow_; return *this;
    }

    // Get elements of the rotation matrix
    double xx() const { return xrow_.x(); }
    double xy() const { return xrow_.y(); }
    double xz() const { return xrow_.z(); }
    double yx() const { return yrow_.x(); }
    double yy() const { return yrow_.y(); }
    double yz() const { return yrow_.z(); }
    double zx() const { return zrow_.x(); }
    double zy() const { return zrow_.y(); }
    double zz() const { return zrow_.z(); }

    // Checkings.
    bool isIdentity() const {
      return (xx() == 1. && xy() == 0. && xz() == 0. &&
	      yx() == 0. && yy() == 1. && yz() == 0. &&
	      zx() == 0. && zy() == 0. && zz() == 1.);
    }
    bool isRotation(double tolerance = 1.e-5) const;

    // Invert.
    Rotation3D inverse() const {
      return Rotation3D(xx(),yx(),zx(), xy(),yy(),zy(), xz(),yz(),zz());
    }
    Rotation3D & invert() {
      return set(xx(),yx(),zx(), xy(),yy(),zy(), xz(),yz(),zz());
    }

    // Set the Identity rotation.
    Rotation3D & setIdentity() {
      set(1.,0.,0., 0.,1.,0., 0,0.,1.); return *this;
    } 

    // Set rotations around x-axis.
    Rotation3D & setRotationX(double);

    // Set rotations around y-axis.
    Rotation3D & setRotationY(double);

    // Set rotations around z-axis.
    Rotation3D & setRotationZ(double);

    // Set result of three consequentive rotations around the axes.
    Rotation3D & setRotationXYZ(double, double, double);

    // Set rotation defined by Euler angles. 
    Rotation3D & setRotationEulerZXZ(double, double, double);
    Rotation3D & setRotationEulerZYZ(double, double, double);

    // Set rotation on given angle along given axis 
    Rotation3D &
    setRotationAngleAxis(double, const BasicVector3D<double> &);

    // Set rotation defined by positions of rotated axes. 
    Rotation3D &
    setRotationAxes(const BasicVector3D<double> & newX,
		    const BasicVector3D<double> & newY,
		    const BasicVector3D<double> & newZ);

    // Set rotation defined by original and rotated positions of two points.
    Rotation3D &
    setRotationPoints(const Point3D<double> & fr1,
		      const Point3D<double> & fr2,
		      const Point3D<double> & to1,
		      const Point3D<double> & to2);

    // Get the rotation angle and the rotation axis.
    void getAngleAxis(double &, BasicVector3D<double> &) const;

    // Get angles of three consequentive rotations around the axes.
    void getXYZ(double &, double &, double &) const;

    // Get Euler angles.
    void getEulerZXZ(double &, double &, double &) const;
    void getEulerZYZ(double &, double &, double &) const;

    // Transformation.
    Rotation3D & transform(const Rotation3D &);
  };

  // Composite rotation.
  inline
  Rotation3D
  operator*(const Rotation3D & a, const Rotation3D & b) {
    Rotation3D m(b);
    return m.transform(a);
  }

} /* namespace gblib */

#endif /* GBLIB_ROTATION3D_HH */
