#ifndef GBLIB_SCREENPROJECTION3D_HH
#define GBLIB_SCREENPROJECTION3D_HH

#include "Projection3D.hh"

namespace gblib {

  /**
   * Provides transformation between different coordinate systems which are
   * in use during visualization.
   *
   * Transformation pipeline includes four coordinate systems: 
   * "user", "normalized", "view" and "projected" spaces.
   *
   * The "user" space is a coordinate system where original data (the model)
   * is given.
   *
   * The "normalized" space is constructed by shifting and scaling "user"
   * space to fit the model into the "globe" - a unit sphere with center
   * at the origin. The axes of the "normalized" space are parallel to the 
   * corresponding axes in the "user" space. Type of scaling is defined
   * by ScalingType. Position of the "equator" and "zero meridian" on 
   * the "globe" is defined by AxisDirection. 
   *
   * The "view" space is similar to the "normalized" space but has different
   * axes. Axes in this space are defined relatively to the screen and view
   * point: X axis is horizontal, Y axis is vertical, Z axis is directed to
   * the view point. Transformation between the "normalized" and "view"
   * spaces is just a rotation, without shifting or scaling. The "view" space
   * is convenient for positioning ligth sources as well as for definition of
   * rotations respective to the screen.
   *
   * The "projected" space is the final coordinate system in the
   * transformation pipeline. Coordinates of vertices in this space should be
   * used for real drawing. Rendering algorithims like hidden lines/surfaces
   * removal usually work in this space. In case of PARALLEL projection
   * the "view" and "projected" spaces are the same: corresponding points
   * in these systems have the same coordinates. 
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  class ScreenProjection3D {
  public:
    /**
     * Projection types.
     */
    enum ProjectionType { PARALLEL, PERSPECTIVE };

    /**
     * Type of scaling during transformation from "user" to "normalized"
     * space.
     */
    enum ScalingType { NORMALIZED, PROPORTIONAL };

    /**
     * Direction of axes in the "normalized" space.
     * AxisOrientation defines orientation of the model inside the "globe"
     * containing the model: first axis points to the cross point of
     * the "equator" and "zero meridian", third axis points to the
     * "north pole". Thus first two axes lie in the "equator" plane and
     * the "zero meridian" goes through third and first axes.
     */
    enum AxisOrientation { XYZ, ZXY, YZX };

    /**
     * Constructor.
     * By default creates Parallel Projection on YZ plane from +X direction
     * (AxisOrientation = XYZ). Y-axis is horizonal, Z-axis is vertical,
     * X-axis is directed to the view point.
     * Min-max for all directions is [-1;+1]. View distance is equal to 3.
     *
     * @see ProjectionType
     * @see ScalingType
     * @see AxisOrientation
     */
    ScreenProjection3D(ProjectionType  p = PARALLEL,
		       ScalingType     s = NORMALIZED,
		       AxisOrientation a = XYZ)
      : ptype(p), stype(s), axdir(a),
	oX(0), oY(0), oZ(0), sizeX(2), sizeY(2), sizeZ(2), vdist(3)
    {
      setUserToNorm(); setNormToView(0,0); setViewToProj();
    }

    /**
     * Sets projection type.
     */
    void setProjectionType(ProjectionType p) { ptype = p; }

    /**
     * Defines dimensions of the model.
     * Dimensions are given in the "user" space. 
     */
    void userSetMinMax(double xmin, double xmax,
		       double ymin, double ymax,
		       double zmin, double zmax) {
      oX = (xmin+xmax)/2; oY = (ymin+ymax)/2; oZ = (zmin+zmax)/2;
      sizeX = xmax-xmin; sizeY = ymax-ymin; sizeZ = zmax-zmin;
      setUserToNorm();
    }

    /**
     * Sets view point in the "normalized" space.
     */
    void normSetViewPoint(double x, double y, double z);

    /**
     * Sets view direction in the "normalized" space.
     */
    void normSetViewDirection(double longitude, double latitude) {
      setNormToView(longitude, latitude);
    }

    /**
     * Sets view distance for PERSPECTIVE projection.
     * It defines a distance between center of the "globe" and
     * the view point. Must be greater than 1. 
     */
    void normSetViewDistance(double dist) { vdist = dist; setViewToProj(); }

    /**
     * Horizontal rotation (respective to the screen).
     */
    void viewRotateH(double ang) {
      n2v = Projection3D().setRotationY(ang)*n2v;
    }

    /**
     * Vertical rotation (respective to the screen).
     */
    void viewRotateV(double ang) {
      n2v = Projection3D().setRotationX(ang)*n2v;
    }
 
    /**
     * Counterclockwise rotation (respective to the screen).
     */
    void viewRotateC(double ang) {
      n2v = Projection3D().setRotationZ(ang)*n2v; }
 
    /**
     * Returns transformation from "user" to "normalized" space.
     */
    Projection3D userToNorm() const { return u2n; } 

    /**
     * Returns transformation from "user" to "view" space.
     */
    Projection3D userToView() const { return n2v*u2n; }

    /**
     * Returns transformation from "user" to "projected" space.
     */
    Projection3D userToProj() const {
      return (ptype == PARALLEL) ? userToView() : v2p*n2v*u2n;
    }

    /**
     * Returns transformation from "normalized" to "view" space.
     */
    Projection3D normToView() const { return n2v; }

    /**
     * Returns transformation from "normalized" to "projected" space.
     */
    Projection3D normToProj() const {
      return (ptype == PARALLEL) ? normToView() : v2p*n2v;
    }

    /**
     * Returns transformation from "view" to "projected" space.
     */
    Projection3D viewToProj() const { 
      return (ptype == PARALLEL) ? Projection3D() : v2p*n2v;
    }

  private:
    ProjectionType  ptype;
    ScalingType     stype;
    AxisOrientation axdir;   
    double oX, oY, oZ;            // center of the min-max box
    double sizeX, sizeY, sizeZ;   // dimensions of the min-max box
    double vdist;                 // distance from view point
    Projection3D u2n, n2v, v2p;   // transformation between the spaces

  private:
    void setUserToNorm();       // sets "user" to "normalized" transformation 
    void setNormToView(double lon, double lat);
    void setViewToProj();       // sets matrix for PERSPECTIVE transformation
  };

} /* namespace gblib */

#endif /* GBLIB_SCREENPROJECTION3D_HH */
