#include <math.h>
#include <iostream>
#include "Point2D.hh"
#include "Transformation2D.hh"

namespace gblib {
//----------------------------------------------------------------------------
  Transformation2D Transformation2D::inverse() const
  /***********************************************************************
   *                                                                     *
   * Name: Transformation2D::inverse                   Date:    01.07.01 *
   * Author: E.Chernyaev (IHEP/Protvino)               Revised:          *
   *                                                                     *
   * Function: Find inverse transformation                               *
   *                                                                     *
   ***********************************************************************/
  {
    double det = xx()*yy()-yx()*xy();
    if (det == 0.) {
      std::cerr
	<< "Transformation2D::inverse() : zero determinant"
	<< std::endl;
      return Transformation2D();
    }
    double detxx = yy()/det, detxy = yx()/det;
    double detyx = xy()/det, detyy = xx()/det;
    return Transformation2D(detxx, -detyx, -detxx*dx()+detyx*dy(),
			   -detxy,  detyy,  detxy*dx()-detyy*dy());
  }

//----------------------------------------------------------------------------
  void
  Transformation2D::getDecomposition(double & sx, double & sy,
				     Rotation2D & r, Translation2D & t) const
  /***********************************************************************
   *                                                                     *
   * Name: Transformation2D::getDecomposition          Date:    01.07.01 *
   * Author: E.Chernyaev (IHEP/Protvino)               Revised:          *
   *                                                                     *
   * Function: Gets decomposition of general transformation on           *
   *           three consequentive particular transformations:           *
   *           Scale, then Rotation, then Translation.                   *
   *           If there was a reflection, then sy will be negative.      *
   *                                                                     *
   ***********************************************************************/
  {
    sx = sqrt(xx()*xx() + yx()*yx());
    sy = sqrt(xy()*xy() + yy()*yy());
    if (xx()*yy()-yx()*xy() < 0.) sy = -sy;
    r = Transformation2D(xx()/sx, xy()/sy, 0, yx()/sx, yy()/sy, 0);
    t = *this;
  }

//----------------------------------------------------------------------------
  Transformation2D &
  Transformation2D::setTransformationPoints(const Point2D<double> & fr0,
					    const Point2D<double> & fr1,
					    const Point2D<double> & to0,
					    const Point2D<double> & to1)
  {
    Rotation2D::setRotationPoints(fr1-fr0, to1-to0);
    Translation2D::setTranslation(to0.x() - xx()*fr0.x() - xy()*fr0.y(),
				  to0.y() - yx()*fr0.x() - yy()*fr0.y());
    return *this;
  }

//----------------------------------------------------------------------------
  Transformation2D &
  Transformation2D::setReflection(double a, double b, double c)
  /***********************************************************************
   *                                                                     *
   * Name: Transformation2D::setReflection             Date:    01.07.01 *
   * Author: E.Chernyaev (IHEP/Protvino)               Revised:          *
   *                                                                     *
   * Function: Sets 2D reflection in a line a*x+b*y+c=0                  *
   *                                                                     *
   ***********************************************************************/
  {
    double ll = a*a+b*b;
    if (ll == 0) {
      std::cerr
	<< "Transformation2D::setReflection : zero normal"
	<< std::endl;
      setIdentity();
    }else{
      ll = 1./ll;
      double aa = a*a*ll, ab = a*b*ll, ac = a*c*ll, bb = b*b*ll, bc = b*c*ll;
      Rotation2D::set(-aa+bb, -ab-ab, -ab-ab,  aa-bb);
      Translation2D::setTranslation(-ac-ac, -bc-bc);
      return *this;
    }
    return *this;
  }
} /* namespace gblib */
