#ifndef GBLIB_TRANSFORMATION2D_HH
#define GBLIB_TRANSFORMATION2D_HH

#include "Translation2D.hh"
#include "Rotation2D.hh"

namespace gblib {

  /**
   * Class for transformation of 2D objects.
   * It allows different translations, rotations, scalings and reflections.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  class Transformation2D : Rotation2D, Translation2D {
  protected:
    // Protected constructor.
    Transformation2D(double xx, double xy, double dx,
		     double yx, double yy, double dy)
      : Rotation2D(xx,xy, yx,yy), Translation2D(dx,dy) {}  
    
    // Sets matrix elements.
    Transformation2D & set(double xx, double xy, double dx,
			   double yx, double yy, double dy)
    {
      Rotation2D::set(xx,xy, yx,yy);
      Translation2D::setTranslation(dx,dy);
      return *this;
    }

  public:
    // Default constructor.
    Transformation2D() {}

    // Constructor from Translation2D.
    Transformation2D(const Translation2D & t)
      : Rotation2D(), Translation2D(t) {}

    // Constructor from Rotation2D.
    Transformation2D(const Rotation2D & r)
      : Rotation2D(r), Translation2D() {}

    // Constructor from Rotation2D and Translation2D.
    Transformation2D(const Rotation2D & r, const Translation2D & t)
      : Rotation2D(r), Translation2D(t) {}

    // Copy constructor.
    Transformation2D(const Transformation2D & m)
      : Rotation2D(m), Translation2D(m) {}

    // Destructor.
    ~Transformation2D() {}

    // Assignment.
    Transformation2D & operator=(const Transformation2D & m) {
      Rotation2D::operator=(m); Translation2D::operator=(m); return *this;
    }
			       
    // Get elements of the Transformation matrix
    double xx() const { return Rotation2D::xx(); }
    double xy() const { return Rotation2D::xy(); }
    double yx() const { return Rotation2D::yx(); }
    double yy() const { return Rotation2D::yy(); }
    double dx() const { return Translation2D::dx(); }
    double dy() const { return Translation2D::dy(); }
    
    // Invert.
    Transformation2D inverse() const;
    Transformation2D & invert() { *this = this->inverse(); return *this; }

    // Get decomposition.
    void getDecomposition(double & sx, double & sy,
			  Rotation2D & r, Translation2D & t) const;

    // Checking.
    bool isIdentity() const {
      return (Rotation2D::isIdentity() && Translation2D::isIdentity());
    }

    // Set the Identity transformation.
    Transformation2D & setIdentity() {
      Rotation2D::setIdentity(); Translation2D::setIdentity(); return *this;
    } 

    // Set translations.
    Transformation2D & setTranslationX(double x) {
      Rotation2D::setIdentity(); Translation2D::setTranslationX(x);
      return *this;
    }
    Transformation2D & setTranslationY(double y) {
      Rotation2D::setIdentity(); Translation2D::setTranslationY(y);
      return *this;
    }
    Transformation2D & setTranslation(double x, double y) {
      Rotation2D::setIdentity(); Translation2D::setTranslation(x, y);
      return *this;
    }

    // Set rotations.
    Transformation2D & setRotation(double a) {
      Rotation2D::setRotation(a); Translation2D::setIdentity(); return *this;
    } 
    Transformation2D &
    setRotationAxes(const BasicVector2D<double> & newX,
		    const BasicVector2D<double> & newY) {
      Rotation2D::setRotationAxes(newX, newY);
      Translation2D::setIdentity();
      return *this;
    }  
    Transformation2D &
    setRotationPoints(const Point2D<double> & fr, const Point2D<double> & to)
    { 
      Rotation2D::setRotationPoints(fr, to);
      Translation2D::setIdentity();
      return *this;
    }  

    // Set reflections.
    Transformation2D & setReflectionX(double x=0.) {
      Rotation2D::set(-1,0, 0,1);
      Translation2D::setTranslationX(x+x);
      return *this;
    }
    Transformation2D & setReflectionY(double y=0.) {
      Rotation2D::set(1,0, 0,-1);
      Translation2D::setTranslationY(y+y);
      return *this;
    }
    Transformation2D & setReflection(double, double, double);

    // Set scales
    Transformation2D & setScaleX(double x) {
      Rotation2D::set(x,0, 0,1); Translation2D::setIdentity(); return *this;
    }
    Transformation2D & setScaleY(double y) {
      Rotation2D::set(1,0, 0,y); Translation2D::setIdentity(); return *this;
    }
    Transformation2D & setScale(double a) {
      Rotation2D::set(a,0, 0,a); Translation2D::setIdentity(); return *this;
    } 
    Transformation2D & setScale(double x, double y) {
      Rotation2D::set(x,0, 0,y); Translation2D::setIdentity(); return *this;
    }

    // Set transformation
    Transformation2D &
    setTransformationPoints(const Point2D<double> & fr0,
			    const Point2D<double> & fr1,
			    const Point2D<double> & to0,
			    const Point2D<double> & to1);

    // Transformations
    Transformation2D & transform(const Translation2D & t) {
      Translation2D::transform(t); return *this;
    }
    Transformation2D & transform(const Rotation2D & m) {
      Rotation2D::transform(m);
      Translation2D::setTranslation(m.xx()*dx()+m.xy()*dy(),
				    m.yx()*dx()+m.yy()*dy());
      return *this;
    }
    Transformation2D & transform(const Transformation2D & m) {
      Rotation2D::transform(m);
      Translation2D::setTranslation(m.xx()*dx()+m.xy()*dy()+m.dx(),
				    m.yx()*dx()+m.yy()*dy()+m.dy());
      return *this;
    }
  };

  // Composite transformations
  inline Transformation2D
  operator*(const Translation2D & a, const Rotation2D & b) {
    return Transformation2D(b, a);
  }
  inline Transformation2D
  operator*(const Translation2D & a, const Transformation2D & b) {
    return Transformation2D(b).transform(a);
  }
  inline Transformation2D
  operator*(const Rotation2D & a, const Translation2D & b) {
    Translation2D t(a.xx()*b.dx()+a.xy()*b.dy(), a.yx()*b.dx()+a.yy()*b.dy());
    return Transformation2D(a, t);
  }

  inline Transformation2D
  operator*(const Rotation2D & a, const Transformation2D & b) {
    return Transformation2D(b).transform(a);
  }
 
  inline Transformation2D
  operator*(const Transformation2D & a, const Translation2D & b) {
    Transformation2D m(a);
    Translation2D t(a.xx()*b.dx()+a.xy()*b.dy(), a.yx()*b.dx()+a.yy()*b.dy());
    return m.transform(t);
  }

  inline Transformation2D
  operator*(const Transformation2D & a, const Rotation2D & b) {
    return Transformation2D(b).transform(a);
  }

  inline Transformation2D
  operator*(const Transformation2D & a, const Transformation2D & b) {
    return Transformation2D(b).transform(a);
  }

} /* namespace gblib */

#endif /* GBLIB_TRANSFORMATION2D_HH */
