#include <math.h>
#include <iostream>
#include "Point3D.hh"
#include "Transformation3D.hh"

namespace gblib {
//----------------------------------------------------------------------------
  Transformation3D Transformation3D::inverse() const
  /***********************************************************************
   *                                                                     *
   * Name: Transformation3D::inverse                   Date:    24.09.96 *
   * Author: E.Chernyaev (IHEP/Protvino)               Revised:          *
   *                                                                     *
   * Function: Find inverse transformation                               *
   *                                                                     *
   ***********************************************************************/
  {
    double detxx = yy()*zz()-yz()*zy();
    double detxy = yx()*zz()-yz()*zx();
    double detxz = yx()*zy()-yy()*zx();
    double det   = xx()*detxx - xy()*detxy + xz()*detxz;
    if (det == 0.) {
      std::cerr
	<< "Transformation3D::inverse() : zero determinant"
	<< std::endl;
      return Transformation3D();
    }
    det = 1./det; detxx *= det; detxy *= det; detxz *= det;
    double detyx = (xy()*zz()-xz()*zy())*det;
    double detyy = (xx()*zz()-xz()*zx())*det;
    double detyz = (xx()*zy()-xy()*zx())*det;
    double detzx = (xy()*yz()-xz()*yy())*det;
    double detzy = (xx()*yz()-xz()*yx())*det;
    double detzz = (xx()*yy()-xy()*yx())*det;
    return Transformation3D
      (detxx, -detyx,  detzx, -detxx*dx()+detyx*dy()-detzx*dz(),
      -detxy,  detyy, -detzy,  detxy*dx()-detyy*dy()+detzy*dz(),
       detxz, -detyz,  detzz, -detxz*dx()+detyz*dy()-detzz*dz());
  }

//----------------------------------------------------------------------------
  void
  Transformation3D::getDecomposition(double & sx, double & sy, double & sz,
				     Rotation3D & r, Translation3D & t) const
  /***********************************************************************
   *                                                                     *
   * Name: Transformation3D::getDecomposition          Date:    09.06.01 *
   * Author: E.Chernyaev (IHEP/Protvino)               Revised:          *
   *                                                                     *
   * Function: Gets decomposition of general transformation on           *
   *           three consequentive particular transformations:           *
   *           Scale, then Rotation, then Translation.                   *
   *           If there was a reflection, then sz will be negative.      *
   *                                                                     *
   ***********************************************************************/
  {
    sx = sqrt(xx()*xx() + yx()*yx() + zx()*zx());
    sy = sqrt(xy()*xy() + yy()*yy() + zy()*zy());
    sz = sqrt(xz()*xz() + yz()*yz() + zz()*zz());
    if (xx()*(yy()*zz()-yz()*zy()) -
	xy()*(yx()*zz()-yz()*zx()) +
	xz()*(yx()*zy()-yy()*zx()) < 0.) sz = -sz;
    r = Transformation3D(xx()/sx, xy()/sy, xz()/sz,0,
			 yx()/sx, yy()/sy, yz()/sz,0,
			 zx()/sx, zy()/sy, zz()/sz,0); 
    t = *this;
  }

//----------------------------------------------------------------------------
  Transformation3D &
  Transformation3D::setTransformationPoints(const Point3D<double> & fr0,
					    const Point3D<double> & fr1,
					    const Point3D<double> & fr2,
					    const Point3D<double> & to0,
					    const Point3D<double> & to1,
					    const Point3D<double> & to2)
  {
    Rotation3D::setRotationPoints(fr1-fr0, fr2-fr0, to1-to0, to2-to0);
    Translation3D::setTranslation
      (to0.x() - xx()*fr0.x() - xy()*fr0.y() - xz()*fr0.z(),
       to0.y() - yx()*fr0.x() - yy()*fr0.y() - yz()*fr0.z(),
       to0.z() - zx()*fr0.x() - zy()*fr0.y() - zz()*fr0.z());
    return *this;
  }

//----------------------------------------------------------------------------
  Transformation3D &
  Transformation3D::setReflection(double a, double b, double c, double d)
  /***********************************************************************
   *                                                                     *
   * Name: Transformation3D::setReflection             Date:    24.09.96 *
   * Author: E.Chernyaev (IHEP/Protvino)               Revised:          *
   *                                                                     *
   * Function: Sets 3D reflection in a plane a*x+b*y+c*z+d=0             *
   *                                                                     *
   ***********************************************************************/
  {
    double ll = a*a+b*b+c*c;
    if (ll == 0) {
      std::cerr
	<< "Transformation3D::setReflection : zero normal"
	<< std::endl;
      setIdentity();
    }else{
      ll = 1/ll;
      double aa = a*a*ll, ab = a*b*ll, ac = a*c*ll, ad = a*d*ll,
	     bb = b*b*ll, bc = b*c*ll, bd = b*d*ll,
	     cc = c*c*ll, cd = c*d*ll;
      Rotation3D::set(-aa+bb+cc, -ab-ab,    -ac-ac,
		      -ab-ab,     aa-bb+cc, -bc-bc,
		      -ac-ac,    -bc-bc,     aa+bb-cc);
      Translation3D::setTranslation(-ad-ad, -bd-bd, -cd-cd);
    }
    return *this;
  }

//----------------------------------------------------------------------------
  Transformation3D &
  Transformation3D::transform(const Rotation3D & m)
  {
    Rotation3D::transform(m);
    Translation3D::setTranslation(m.xx()*dx() + m.xy()*dy() + m.xz()*dz(),
				  m.yx()*dx() + m.yy()*dy() + m.yz()*dz(),
				  m.zx()*dx() + m.zy()*dy() + m.zz()*dz());
    return *this;
  }

//----------------------------------------------------------------------------
  Transformation3D &
  Transformation3D::transform(const Transformation3D & m)
  {
    Rotation3D::transform(m);
    Translation3D::setTranslation
      (m.xx()*dx() + m.xy()*dy() + m.xz()*dz() + m.dx(),
       m.yx()*dx() + m.yy()*dy() + m.yz()*dz() + m.dy(),
       m.zx()*dx() + m.zy()*dy() + m.zz()*dz() + m.dz());
    return *this;
  }
} /* namespace gblib */
