#ifndef GBLIB_TRANSFORMATION3D_HH
#define GBLIB_TRANSFORMATION3D_HH

#include "Translation3D.hh"
#include "Rotation3D.hh"

namespace gblib {

  /**
   * Class for transformation of 3D objects.
   * It allows different translations, rotations, scalings and reflections.
   *
   * @author Evgeni Chernyaev <Evgueni.Tcherniaev@cern.ch>
   */
  class Transformation3D : public Rotation3D, public Translation3D {
  protected:
    // Protected constructor.
    Transformation3D(double xx, double xy, double xz, double dx,
		     double yx, double yy, double yz, double dy,
		     double zx, double zy, double zz, double dz)
      : Rotation3D(xx,xy,xz, yx,yy,yz, zx,zy,zz), Translation3D(dx,dy,dz) {}  

    // Sets matrix elements.
    Transformation3D & set(double xx, double xy, double xz, double dx,
			   double yx, double yy, double yz, double dy,
			   double zx, double zy, double zz, double dz)
    {
      Rotation3D::set(xx,xy,xz, yx,yy,yz, zx,zy,zz);
      Translation3D::setTranslation(dx,dy,dz);
      return *this;
    }

  public:
    // Default constructor.
    Transformation3D() {}

    // Constructor from Translation3D.
    Transformation3D(const Translation3D & t)
      : Rotation3D(), Translation3D(t) {}

    // Constructor from Rotation3D.
    Transformation3D(const Rotation3D & r)
      : Rotation3D(r), Translation3D() {}

    // Constructor from Rotation3D and Translation3D.
    Transformation3D(const Rotation3D & r, const Translation3D & t)
      : Rotation3D(r), Translation3D(t) {}

    // Copy constructor.
    Transformation3D(const Transformation3D & m)
      : Rotation3D(m), Translation3D(m) {}

    // Destructor.
    ~Transformation3D() {}

    // Assignment.
    Transformation3D & operator=(const Transformation3D & m) {
      Rotation3D::operator=(m); Translation3D::operator=(m); return *this;
    }
			       
    // Get elements of the Transformation matrix
    double xx() const { return Rotation3D::xx(); }
    double xy() const { return Rotation3D::xy(); }
    double xz() const { return Rotation3D::xz(); }
    double yx() const { return Rotation3D::yx(); }
    double yy() const { return Rotation3D::yy(); }
    double yz() const { return Rotation3D::yz(); }
    double zx() const { return Rotation3D::zx(); }
    double zy() const { return Rotation3D::zy(); }
    double zz() const { return Rotation3D::zz(); }
    double dx() const { return Translation3D::dx(); }
    double dy() const { return Translation3D::dy(); }
    double dz() const { return Translation3D::dz(); }
    
    // Invert.
    Transformation3D inverse() const;
    Transformation3D & invert() { *this = this->inverse(); return *this; }

    // Get decomposition.
    void getDecomposition(double & sx, double & sy, double & sz,
			  Rotation3D & r, Translation3D & t) const;

    // Checking.
    bool isIdentity() const {
      return (Rotation3D::isIdentity() && Translation3D::isIdentity());
    }

    // Set the Identity transformation.
    Transformation3D & setIdentity() {
      Rotation3D::setIdentity(); Translation3D::setIdentity(); return *this;
    } 

    // Set translations.
    Transformation3D & setTranslationX(double x) {
      Rotation3D::setIdentity(); Translation3D::setTranslationX(x);
      return *this;
    }
    Transformation3D & setTranslationY(double y) {
      Rotation3D::setIdentity(); Translation3D::setTranslationY(y);
      return *this;
    }
    Transformation3D & setTranslationZ(double z) {
      Rotation3D::setIdentity(); Translation3D::setTranslationZ(z);
      return *this;
    }
    Transformation3D & setTranslation(double x, double y, double z) {
      Rotation3D::setIdentity(); Translation3D::setTranslation(x, y, z);
      return *this;
    }

    // Set rotations.
    Transformation3D & setRotationX(double a) {
      Rotation3D::setRotationX(a); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D & setRotationY(double a) {
      Rotation3D::setRotationY(a); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D & setRotationZ(double a) {
      Rotation3D::setRotationZ(a); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D & setRotationXYZ(double a, double b, double c) {
      Rotation3D::setRotationXYZ(a, b, c); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D & setRotationEulerZXZ(double a, double b, double c) {
      Rotation3D::setRotationEulerZXZ(a, b, c); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D & setRotationEulerZYZ(double a, double b, double c) {
      Rotation3D::setRotationEulerZYZ(a, b, c); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D &
    setRotationAngleAxis(double a, const BasicVector3D<double> & v) {
      Rotation3D::setRotationAngleAxis(a, v); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D &
    setRotationAxes(const BasicVector3D<double> & newX,
		    const BasicVector3D<double> & newY,
		    const BasicVector3D<double> & newZ) {
      Rotation3D::setRotationAxes(newX, newY, newZ);
      Translation3D::setIdentity();
      return *this;
    }  
    Transformation3D &
    setRotationPoints(const Point3D<double> & fr1,
		      const Point3D<double> & fr2,
		      const Point3D<double> & to1,
		      const Point3D<double> & to2) {
      Rotation3D::setRotationPoints(fr1,fr2, to1,to2);
      Translation3D::setIdentity();
      return *this;
    }  

    // Set reflections.
    Transformation3D & setReflectionX(double x=0.) {
      Rotation3D::set(-1,0,0, 0,1,0, 0,0,1);
      Translation3D::setTranslationX(x+x);
      return *this;
    }
    Transformation3D & setReflectionY(double y=0.) {
      Rotation3D::set(1,0,0, 0,-1,0, 0,0,1);
      Translation3D::setTranslationY(y+y);
      return *this;
    }
    Transformation3D & setReflectionZ(double z=0.) {
      Rotation3D::set(1,0,0, 0,1,0, 0,0,-1);
      Translation3D::setTranslationZ(z+z);
      return *this;
    }
    Transformation3D & setReflection(double, double, double, double);

    // Set scales
    Transformation3D & setScaleX(double x) {
      Rotation3D::set(x,0,0, 0,1,0, 0,0,1); Translation3D::setIdentity();
      return *this;
    }
    Transformation3D & setScaleY(double y) {
      Rotation3D::set(1,0,0, 0,y,0, 0,0,1); Translation3D::setIdentity();
      return *this;
    }
    Transformation3D & setScaleZ(double z) {
      Rotation3D::set(1,0,0, 0,1,0, 0,0,z); Translation3D::setIdentity();
      return *this;
    }
    Transformation3D & setScale(double a) {
      Rotation3D::set(a,0,0, 0,a,0, 0,0,a); Translation3D::setIdentity();
      return *this;
    } 
    Transformation3D & setScale(double x, double y, double z) {
      Rotation3D::set(x,0,0, 0,y,0, 0,0,z); Translation3D::setIdentity();
      return *this;
    }

    // Set transformation
    Transformation3D &
    setTransformationPoints(const Point3D<double> & fr0,
			    const Point3D<double> & fr1,
			    const Point3D<double> & fr2,
			    const Point3D<double> & to0,
			    const Point3D<double> & to1,
			    const Point3D<double> & to2);

    // Transformations
    Transformation3D & transform(const Translation3D & t) {
      Translation3D::transform(t); return *this;
    }
    Transformation3D & transform(const Rotation3D & r);
    Transformation3D & transform(const Transformation3D & m);
  };

  // Composite transformations
  inline Transformation3D
  operator*(const Translation3D & a, const Rotation3D & b) {
    return Transformation3D(b, a);
  }
  inline Transformation3D
  operator*(const Translation3D & a, const Transformation3D & b) {
    return Transformation3D(b).transform(a);
  }
  inline Transformation3D
  operator*(const Rotation3D & a, const Translation3D & b) {
    Translation3D t(a.xx()*b.dx()+a.xy()*b.dy()+a.xz()*b.dz(),
		    a.yx()*b.dx()+a.yy()*b.dy()+a.yz()*b.dz(),
		    a.zx()*b.dx()+a.zy()*b.dy()+a.zz()*b.dz());
    return Transformation3D(a, t);
  }

  inline Transformation3D
  operator*(const Rotation3D & a, const Transformation3D & b) {
    return Transformation3D(b).transform(a);
  }
 
  inline Transformation3D
  operator*(const Transformation3D & a, const Translation3D & b) {
    Transformation3D m(a);
    Translation3D t(a.xx()*b.dx()+a.xy()*b.dy()+a.xz()*b.dz(),
		    a.yx()*b.dx()+a.yy()*b.dy()+a.yz()*b.dz(),
		    a.zx()*b.dx()+a.zy()*b.dy()+a.zz()*b.dz());
    return m.transform(t);
  }

  inline Transformation3D
  operator*(const Transformation3D & a, const Rotation3D & b) {
    return Transformation3D(b).transform(a);
  }

  inline Transformation3D
  operator*(const Transformation3D & a, const Transformation3D & b) {
    return Transformation3D(b).transform(a);
  }

} /* namespace gblib */

#endif /* GBLIB_TRANSFORMATION3D_HH */
