#ifndef QPAXIS_H
#define QPAXIS_H

/** %%%
  Usage:

  TODO:

  (--) Missing grids. We first need a better structure to store ticks positions.
  (--) Still using qwt_math and qwt_dimapv ( qpqwt.h as well??? )
  (--) Orthogonal and SameSide option must be improved (or done at all)
  (--) drawLabels: set the label font size so that the text is readable
  
  areas requiring improvements are emphasized by [] comments.
*/


#include "Qplotter/qp/qpzone.h"
#include "Qplotter/qp/qpmultiline.h"
#include "Qplotter/qp/qptext.h"
#include "Qplotter/qp/qpnode.cpp"


class QpAxis : public QpNode<QpObject>
{
public:
  /** Scientific: multiply factor (such as "x 10e6") and only one figure before the dot.
      Engineer: 
      Normal: labels as they are defined by optimization
  */
  enum TickLabel {Normal, Scientific, Engineer};
  /** Options for scales :
      - Logarithmic : logarithmic scale, default is linear
      - Arrow : an arrow is drawn at the end of the axis
      - Tick[Positive/Negative] : tick marks are drawn on the
        positive/negative side of the axis
      - TickVertical : tick marks are drawn vertically
        (default is perpendicular to axis)
      - TickHorizontal : tick marks are drawn horizontally 
        (not compatible with TickVertical)
      - Unlabeled : unlabeled axis (default is labeled)
      - [ SameSide : labels are drawn on the same side as the tick marks] 
      - Orthogonal : labels are drawn orthogonal to the axis
      - Optimized : optimize axis, default is optimized
      - GridMajor : Draw the grid on the level 1 tick marks.
      - GridMinor : Draw the grid on the level 2 tick marks.
  **/
  enum Options {                             None           =   0,
    Logarithmic  =   1, Arrow        =    2, TickPositive   =   4, 
    TickNegative =   8, TickVertical =   16, TickHorizontal =  32, 
    Unlabeled    =  64, SameSide     =  128, Orthogonal     = 256, 
    Optimized    = 512, GridMajor    = 1024, GridMinor      = 2048,
    Axis3d       = 4096				    
  };
public:
  /// Constructors
  /// (_x1,_y1) (_x2,_y2) = origin and end of the axis
  /// _minVal/_maxVal : lowest and highest value for the tick marks labels
  /// nbdiv : number of  divisions computed according to the following convention :
  /// nbdiv = (nb of primary divisions)+100*(secondary)+10000*(third order divisions)
  /// (modulo operations decode the number, e.g. 
  /// 30910 -> 10 primary level, 9 secondary level, 3 tertiary level
  QpAxis(double _x1, double _y1, double _x2, double _y2,
	 double _minVal, double _maxVal, 
	 int opt=0, unsigned int nbdiv=10,
	 QpObject *z=0, bool addOnParent=false,
	 QpTransformation::tStrategy st=QpTransformation::Page);
  /// _step : distance between 2 major tick marks
  QpAxis(double _x1, double _y1, double _x2, double _y2,
	 double _minVal, double _maxVal, double _step, int opt=0,
	 QpObject *z=0, bool addOnParent=false,
	 QpTransformation::tStrategy st=QpTransformation::Page);
  /// _labels : list of tick marks labels
  QpAxis(double _x1, double _y1, double _x2, double _y2,
	 QPtrList<QString> *labels_, int opt=0,
	 QpObject *z=0, bool addOnParent=false,
	 QpTransformation::tStrategy st=QpTransformation::Page);
  /// Destructor [has to be completed]
  ~QpAxis() ;

  /// Accessors
  void getOrigin(double &x, double &y) const {x=x1;y=y1;}
  void getEnd(double &x, double &y) const {x=x2; y=y2;}
  void getValues(double &vmin, double &vmax) const {vmin=minVal; vmax=maxVal;}
  int getMaxL1Divs() const {return maxL1Divs;}
  int getMaxL2Divs() const {return maxL2Divs;}
  int getMaxL3Divs() const {return maxL3Divs;}
  int getL1Divs() const {return L1Divs;}
  int getL2Divs() const {return L2Divs;}
  int getL3Divs() const {return L3Divs;}
  double getStep() const {return stepl1;}
  double getStepL1() const {return stepl2;}
  double getStepL2() const {return stepl3;}
  Options getOptions() const {return options;}
  TickLabel getLabelMode() const {return labelMode;}
  QString getFindHint() const {return findHint;}
  QPtrList<QString> *getLabels() const {return labels;}
  QString getMulti() const {return multi;}
  QString getTitle() const {return title;}
  QString getLegend() const {return legend;}
  double getLabelDist() const {return labelDist;}
  double getTitleDist() const {return titleDist;}
  double getLegendDist() const {return legendDist;}
  double getMultiDist() const  {return multiDist;}
  double getL1Height() const {return l1Height;}
  double getL2Height() const {return l2Height;}
  double getL3Height() const {return l3Height;}
  double getArrowLength() const {return arrowLength;}
  double getArrowWidth() const {return arrowWidth;}
  double getGridLenght() const {return gridLength;}

  /// Manipulators
  void setLimits(double _x1, double _y1, double _x2, double _y2);
  void setValues(double vmin, double vmax) {minVal=vmin; maxVal=vmax;}
  void setMaxL1Divs(int nb) {if (nb>=1) maxL1Divs=nb;}
  void setMaxL2Divs(int nb) {maxL2Divs=nb; changed();}
  void setMaxL3Divs(int nb) {maxL3Divs=nb; changed();}
  void setStep(int nb) {stepl1=nb; changed();}
  void setStepL1(int nb) {stepl2=nb; changed();}
  void setStepL2(int nb) {stepl3=nb; changed();}
  /// [De]activate the option opt on this axis (according to value)
  void setOption(Options opt, bool value);
  /// Set a precise list of options
  void setOptions(Options opt) {options=opt; changed();}
  void setOptions(int opt);
  void setLabelMode(TickLabel lm) {labelMode=lm; changed();}
  void setLabelAlignment(int _align);
  void setFindHint(QString fh) {findHint=fh; changed();}
  void setLabels(QPtrList<QString> *l) {delete labels; labels=l; changed();}
  void setMulti(QString m) {multi=m; changed();}
  void setTitle(QString t) {title=t; changed();}
  void setLegend(QString l) {legend=l; changed();}
  void setLabelDist(double d) {labelDist=d; changed();}
  void setTitleDist(double d) {titleDist=d; changed();}
  void setLegendDist(double d) {legendDist=d; changed();}
  void setMultiDist(double d) {multiDist=d; changed();}

  void setL1Height(double d) {l1Height=d; changed();}
  void setL2Height(double d)  {l2Height=d; changed();}
  void setL3Height(double d)  {l3Height=d; changed();}
  void setArrowWidth(double d)  {arrowWidth=d; changed();}
  void setArrowLength(double d)  {arrowLength=d; changed();}
  void setGridLength(double d)  {gridLength=d; changed();}

  /// Change the limits of the axis
  void update(double _minVal, double _maxVal);
  /** [Is this implemented at all???] */
  QpObject *touched(double _x, double _y, double errorX, double errorY, QpObject *&node);

protected:
  /// Variables initialization
  void init();
  /** Creation of the labels string list by using an equidistant distribution
      Uses minVal, maxVal and maxL1Div
      [Should use label optimization]
   **/
  void computeLabels();
  /** Computation of the tick marks 
   **/
  void createTickMarks();
  /** Computation of the linear tick marks 
   **/
  void createLinearTickMarks();
  /** Computation of the Log tick marks 
   **/
  void createLogTickMarks();
  /** Draw the axis according to all the data structure defined
   **/
  void drawAxis();
  /// Optimization function taken from CernLib/Root
  void optimize(double oldMin, double oldMax, int oldN, double &BinLow, double &BinHigh, int &nbins, double &BinWidth);
  /** Rotation and translation of one point from the horizontal axis 
      representation
      Phi = atan2((y2-y1),(x2-x1))  angle of the axis according to horizontal
      PhiDeg = Phi in degrees
      cosPhi = cos(Phi)   sinPhi = sin(Phi)
      (x1,y1) -> (x2,y2) : limits of the axis
   **/
  void rotation(double &x, double &y);
  void computeEndPoint(double &x, double &y);
  /** Creation of a QpMultiLine with nbPoints whose coordinates are taken from tabx and taby
      Parameters : nbPoints = how many points to link 
                   tabx and taby = arrays of coordinates
   **/
  QpMultiLine* createMultiLine(int nbPoints, double *tabx, double *taby);
  /** Create and place the labels according to the coordinates 
      of the tick marks (from tabx and taby)
      Parameters : nbPoints = how many coordinates in these arrays
                   tabx and taby = arrays of tick marks' points coordinates
		   cosPhis, sinPhi = cos(Phi) and sin(Phi) where Phi is the angle of the axis
  **/      
  void drawLabels(int nbPoints, double *tabx, double *taby);
  /** Create and place the QpMultiLine for the axis arrow
      Uses arrowLength, arrowWidth
  **/
  void drawArrow();
  /// Function to call whenever the axis must be computed again
  void changed() { computeLabels(); 
                   createTickMarks(); 
		   drawAxis();}

private:

  /// Coordinates of extremities of the scale
  double x1, y1, x2, y2;
  
  /// Min and Max of the axis labels
  double minVal, maxVal;

  /// Optimized Min and Max of the axis labels
  double Optimized_minVal, Optimized_maxVal;

  /// Max number of differents ticks (for automatic tick placement)
  int maxL1Divs, maxL2Divs, maxL3Divs;

  /// Number of differents ticks after optimization. If the optimization is
  /// off these value are equal to the maximums.
  int L1Divs, L2Divs, L3Divs;

  /// Distance between ticks
  double stepl1, stepl2, stepl3;

  /// Graphics data structures to be drawn
  int NbTickLevel2, NbTickLevel3;
  double *TickLevel1X, *TickLevel1Y;
  double *TickLevel2X, *TickLevel2Y;
  double *TickLevel3X, *TickLevel3Y;
  double *GridX1, *GridY1;
  double *GridX2, *GridY2;
  double Phi, PhiDeg, sinPhi, cosPhi;
  double l1TickHeight, l2TickHeight, l3TickHeight;
  double projection;
  double axisLength;
  void fillTickArray(double xi, double yi, int i, double *TickX, double *TickY);

  /// Options for this axis (see the enum named Options)
  Options options;
  /// [Not used atm]
  TickLabel labelMode;

  /** To find a particular axis from an axis group.
      The default axis (handled by the zone) will be named "x1" and "y1"
  **/
  QString findHint;
  /// List of strings which will be used for the labels
  QPtrList<QString> *labels;

  /** The three strings which define the axis: title is on the middle, label at one extreme
    and multi is  a multiplication factor for the label values.
    */

  QString title, legend, multi;
  /** Text formatting : the bitwise OR of the Qt's alignment flags
      By default : Qt::AlignVCenter|Qt::AlignRight for vertical axis and
      Qt::AlignHCenter|Qt::AlignTop for others
      Keep the 0 value to use the default alignment.
  **/
  int labelAlignment; 
  /// [Will be used to center a QpText on a point (from x-centerVar , to x+centerVar)]
  double centerVar;
  /// Distance between strings and axis in percentage of the axis size
  double labelDist, titleDist, legendDist, multiDist;
  /// Height of different types of ticks in percentage of the axis size
  double l1Height, l2Height, l3Height;
  /// Height and width of the optional arrow in percentage of the axis size
  double arrowLength, arrowWidth;
  /// Grid lenght in page coordinates
  double gridLength;

};

/** The axis container class. QpAxis allows to group several axis for one zone.
    This class inherits from QpNode with QpAxis as a template element.
*/
class QpAxisNode : public QpNode<QpAxis>
{
public:
  /// Constructor taking a Zone pointer
  QpAxisNode(QpObject *z, bool addOnParent=true):QpNode<QpAxis>(new QString("QpAxisNode"),z,addOnParent) {}
  /// Destructor
  ~QpAxisNode() {}

  /// Return the first Axis which has this orientation/findHint
  QpAxis * const getAxis(QString findHint);
  /// Update the range of the i-th axis
  void update(int index, double min, double max) {
    ((QpAxis *) at(index))->update(min,max);
  }
  /// Set the opt option to val on all scales
  /// [has to be done with the new system]
  /*void changeOptions(int opt, bool val) {
    QPtrListIterator<QpScale> iGO(list); 
    for( ; iGO.current(); ++iGO ) 
      iGO.current()->getAutoScale()->changeOptions(opt,val);
  }*/
  /// Comparison function usable in QpNode::searchElement
  static bool cmpQpAxisNode(QpObject *obj)   { 
    return (obj->isA() == QString("QpAxisNode")); 
  }
};

#endif
