#ifndef QPCOMMONTEXT_H
#define QPCOMMONTEXT_H


/** %%%

    The basic text object.
    
  Usage:

   QpCommonText stores information to support drawing  text at a position 
   (x,y) or  aligned in a (x,y,w,h) rectangle. The different 
   types of alignment for Qt are the bitwise OR 
   of the following flags :
       AlignLeft aligns to the left border. 
       AlignRight aligns to the right border. 
       AlignHCenter aligns horizontally centered. 
       AlignTop aligns to the top border. 
       AlignBottom aligns to the bottom border. 
       AlignVCenter aligns vertically centered 
       AlignCenter (= AlignHCenter | AlignVCenter) 
       SingleLine ignores newline characters in the text. 
       ExpandTabs expands tabulators. 
       ShowPrefix displays "&x" as "x" underlined. 
       WordBreak breaks the text to fit the rectangle.    

   By default, the type of alignment is : AlignTop | AlignLeft.| DontClip
  
*/


#include "Qplotter/qp/qpobject.h"

class QpCommonText : public QpObject 
{
public:
  /// Constructor
  QpCommonText(QString _data,QpObject *z,double _x,double _y, double _w,
	       double _h , bool addOnParent=true,
	       QpTransformation::tStrategy st=QpTransformation::Zone);
  /// Destructor
  virtual ~QpCommonText() {}

  void setFontSize(int s)
    {
      fontSize = s;
    }
  
  /// Retrieve X coordinate
  double getX() const {return xC;}
  /// Retrieve Y coordinate
  double getY() const {return yC;}
  /// Retrieve width of the bounding rectangle
  double getW() const {return bRectWidth;}
  /// Retrieve height of the bounding rectangle
  double getH() const {return bRectHeight;}
  /// Retrieve the QString
  QString getData() const { return data; }
  /// Append data
  void appendData(QString );

  /** Return the width of the text in pixel according to its painter. */
  virtual int getPixWidth(QPainter *p) const  = 0;
  /** Return the height of the text in pixel according to its painter. */
  virtual int getPixHeight(QPainter *p) const = 0;
 
  /** Return the width of the text in page coords according to its painter. */
  double getWidth(QPainter *painter) const ;
  /** Return the height of the text in page coords according to its painter.*/
  double getHeight(QPainter *painter) const ;
  /// Set coordinates of the object. w and h are optional
  void setCoords(double _x, double _y, double _w=0.0, double _h=0.0);
  /// Set type of alignment
  void setAlignment(int _align) {align=_align | Qt::DontClip;}
  /// Get type of alignment
  int getAlignment() const {return align;}
  /// Set angle of rotation (Qt uses clockwise increasing angles...)
  void setAngle(double _angle) {angle = 360 - _angle;}
  /// Get angle of rotation
  double getAngle() const {return angle;}
  /** Apply rotation to painter. Returns coordinates transformed according
    to the rotation
  */
  void rotatePainter(QPainter *painter,int &transfX, int &transfY);
  /// Undo  painter rotation
  void restorePainter(QPainter *painter);
  /// Manage horizontal alignment on our own
  int getXOffset (int, QPainter *) const;
  /// Manage vertical alignment on our own
  int getYOffset (int, QPainter *) const ;
  /** Do we want baseline correction (i.e. shift text upwards so that
    letters like g or y are properly displayed if the text is aligned on
    some graphic entity)? As an example the zone title needs baseline
    correction since it is aligned on the zone bounding box
    */
  void setBaseline (bool v) {addBaseLine=v;} 
  /// Get baseline correction status
  bool getBaseline () const {return addBaseLine;} 
private:
  QString data;
  int fontSize;
  
  // Coordinates of the object
  double xC;
  double yC;
  // Bounding rectangle
  double bRectWidth;
  double bRectHeight;
  // Alignment properties
  int align;
  // Rotation angle 
  double angle;
  // Take into account adjustment for baseline
  bool addBaseLine;

};




#endif



