#ifndef QPLISTENER_H
#define QPLISTENER_H

#if defined( __sun ) && __SUNPRO_CC < 0x500
	#define true 1
	#define false 0
	typedef int bool;
#endif

/** %%%

    Event listener
    
  Usage:

    The class to provide information about mouse events and their
  context.  This class is updated by the QpPage every time a mouse
  event on the page is detected. The class will in turn notify all its
  observers that such an event has happened, so that they can react
  appropriately.  The class keeps information about the event and the
  place it happened, i.e. event type, page coordinates, zone
  coordinates, object picked, node picked.  Together with the
  QpEvObserver class it implements the "Observer" design pattern (See
  "Design Patterns" by Gamma et. al).
  
  TODO:

*/



#include <qptrlist.h>
#include <qstring.h>
#include <qframe.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpainter.h>

class QpEvObserver;
class QpObject;
template <class Element>
class QpNode;

/**
*/

class QpListener
{
public:
  enum EventType {None=0, LeftPressed=1, LeftReleased=2, RightPressed=4, RightReleased=8, LeftDouble=16, Move=32, Display=64};

  QpListener() ;
  ~QpListener() ;

  /// Setter methods
  void setEventType(EventType t)         {listenerType=t; }
  void setPageCoords(double _xp, double _yp) {xp=_xp; yp=_yp; }
  void setZoneCoords(double _xz, double _yz) {xz=_xz; yz=_yz; }
  void setZoneTitle(QString title)       {zone=title; }  
  void setPicked(QpObject *o)            {picked=o; }
  void setPickedNode(QpNode<QpObject>* n){pnode=n; }
  void setPixelCoords(int _x, int _y)    {x=_x; y=_y;}

  /// Getter methods
  EventType getEventType() const {return listenerType;}
  // Warning ! If the zone is void, xz and yz contain meaningless value
  double getXp()           const {return xp;}
  double getYp()           const {return yp;}
  double getXz()           const {return xz;}
  double getYz()           const {return yz;}
  int    getX()            const {return x;}
  int    getY()            const {return y;}
  QString getZoneTitle()   const {return zone;}
  QpObject *getPicked()    const {return picked;}
  QpNode<QpObject> *getPickedNode() const {return pnode;}

  /// Add an observer to the internal list
  void addObserver(QpEvObserver *o);
  /// Remove an observer to the internal list
  void removeObserver(QpEvObserver *o) {    
    observers->removeRef(o);    
  }
  void end();
  /** Notify explicitly the observers. This approach requires an explicit call
    from the client, but it allows to cumulate several changes in one notification
    (this is exactly what we want most of the time: an atomic snapshot of the event).
    See Implementation notes in Design Pattern book
  */
  void notify();
private:
  EventType listenerType;
  double xp,yp,xz,yz;
  int x, y;
  QString zone;
  QpObject *picked;
  QpNode<QpObject> *pnode;
  QPtrList<QpEvObserver> *observers;
};

class QpPage;

/** This class implements the "consumer side" of the Observer pattern. The class
  registers with a QpLister specifying the type of events it's interested in.
  Whenever such an event happens, the Observer is notified by executing its update()
  method. Users willing to manage events will subclass it and implement the pure
  virtual method update().
*/


class QpEvObserver
{
public:
  QpEvObserver(QpListener::EventType t);
  virtual ~QpEvObserver() {}
  /// Update the observer
  virtual void update() = 0;
  /// Register itself on a page QpListener
  virtual void lRegister(QpPage *p);
  /// Register itself on a QpListener
  virtual void lRegister(QpListener *p);
  virtual void end();
  void setEventType(QpListener::EventType t) {type=t;}
  QpListener::EventType getType() const {return type;}
protected:
  QpListener *listener;
  QpListener::EventType type;
};



class QpZone;

/** A class to superimpose an outline over the zone. Using this class it is possible
  to draw a cross, a rectangle or an ellipse over a zone without interfering with
  the underlying picture (a Xor operator is used to "negate" the outline.
  The type of outline can be selected at construction time. The class is an observer,
  thus depending on the ouline chosen it will examine the event context and take the
  appropriate action.
  If the type is {VLine, HLine, Cross}, the outline is displayed and follows the cursor
  according to mouse move ebvents.
  If the type is {Rectangle, Ellipse}, the class will intercept the first mouse click
  as the origin of the shape, then enlarge the shape as long as the mouse moves and
  finally freeze the shape on mouse release event.
*/

class QpOutLine : public QpEvObserver
{
public:
  enum Type {Off=0, VLine=1, HLine=2, Cross=3, Rectangle=4, Ellipse=8};
  enum State {NoSelection, Ready, PickedFirst};
public:
  QpOutLine(Type type=Off);
  virtual ~QpOutLine();
  virtual void update();
  virtual void updateLimit();
  virtual void end();
  virtual void setType(Type t);
  virtual void setSelectionState(State s=Ready) {outlineState=s;}
  Type getType() const {return type;}
protected:
  void init(QpZone *z);
  void updateMove();
  void updateLeftPressed();  
  void updateZone(QpZone *z);
  
  void drawType(Type type, State state);
  void drawRect();
  void drawEllipse();
  void getLimit();
  
  // Coordinates spanning the whole X range
  int xmin,xmax;
  // Coordinates spanning the whole Y range
  int ymin,ymax;
  // Coordinates of point where Left Press happens (anchor point)
  int x1, y1;
  // Coordinates of point where Left Release happens (end selection)
  int x2, y2;
  // Zone where the interaction starts
  QpZone *z1;
  Type type;
  // Internal state of the outline
  State outlineState;
  QPainter *p;
  bool visibleOutline;
  
};

/** A class to manage zooming on a zone. Using this class it is possible to select 
  a rectangular zoom area and apply a zoom() operation on the zone so that the
  selected area is magnified.
  The class allows to unzoom() a magnified image as well. Since the zone allows
  unlimited zoom/unzoom, it is possible to apply the zoom opeartion several times
  and then backtrace it.
*/

class QpZoom : public QpEvObserver
{
public:
  /** Valid states. Transitions are:
    Zoom -------(RightPressed)--------------------------> Zoom
    Zoom -------(LeftPressed && pressed on a zone)------> PickedFirst
    Zoom -------(LeftPressed && not pressed on a zone)--> Off
    Unzoom -----(LeftPressed)---------------------------> Off or UnZoom
    PickedFirst-(LeftReleased)--------------------------> Off or Zoom
  
    Zoom and Unzoom states are set by external entities (such as QpPage signals).
    Default state is Off.
  */
  enum State {Off, Zoom, PickedFirst, UnZoom};
  QpZoom(QpOutLine *o,bool fixedstate);
  virtual ~QpZoom();  
  virtual void update();
  virtual void end();
  void setState(State s);
  State getState() const {return state;}
protected:
  // Coordinates of point where Left Press happens (zone coordinates!)
  double x1, y1;
  // Coordinates of point where Left Press happens (anchor point)
  int xm1, ym1;
  // Coordinates of point where Left Release happens (end selection)
  int xm2, ym2;

  // For 3D zoom (Todo: How to zoom in 3d with a mouse ?)
  double longit;
  double lat;
  double norm;
  
  
  // Zone where the interaction starts
  QpZone *z1;
  // Internal state of the zoom
  State state;
  // Fixed means we continue zooming after each zoom
  bool fixed ;  
  // The QpOutline instance used to draw the selection area
  QpOutLine *ol;
};


/** A class to provide information on what's below the cursor in a zone. 
  This class is a Qt frame, thus it will show up as a separate window displaying
  all the information attached to the last mouse move event.
  This class is an example on how to exploit the event context to retrieve e.g.
  the name of the object pointed by the mouse (as one would do i.e. to implement
  picking).
*/

class QpFrameLocator : public QFrame, public QpEvObserver
{
public:
  QpFrameLocator(QWidget *p=0, const char *name=0, WFlags f=Qt::WStyle_StaysOnTop);
  virtual ~QpFrameLocator();

  /// Update the widget
  virtual void update();
  virtual void end();

  virtual void setXp(double _xp=0.0)  { xp->setNum(_xp); }
  virtual void setYp(double _yp=0.0)  { yp->setNum(_yp); }
  virtual void setXz(double _xz=0.0)  { xz->setNum(_xz); }
  virtual void setXzInvalid() { xz->setText(QString("--"));}
  virtual void setYz(double _yz=0.0)  { yz->setNum(_yz); }
  virtual void setYzInvalid() { yz->setText(QString("--"));}
  virtual void setZone(QString _zone) { zone->setText(_zone); }
  virtual void setPicked(QString _picked) { picked->setText(_picked); }

protected:
  void closeEvent( QCloseEvent *e );

private:
  QLabel *xpl, *ypl, *xzl, *yzl, *zonel, *pickedl;
  QLabel *xp, *yp, *xz, *yz, *zone, *picked;
  QGridLayout* grid;
  bool toClose;
};


#endif
