#include "Qplotter/qp/qpmap.h"
#include "Qplotter/qp/qpqwt_math.h"
#include <float.h>

const double QpMap::LogMin = 1.0e-150;
const double QpMap::LogMax = 1.0e150;
//------------------------------------------------------------
//.H QpMap 
// From QwtDiMap -> Qwt Widget Library
//------------------------------------------------------------
//------------------------------------------------------------
//
//.U NAME
//	QpMap - Mapping between two double intervals 
//
//.U SYNOPSIS
//	#include <qpmap.h>
//
//.U DESCRIPTION
//	The QpMap class maps an interval of type double into an interval of
//	type double. It consists
//	of two intervals D = [d1, d2] (double) and I = [i1, i2] (double), which are
//	specified with the @QpMap::setContentRange@ and @QpMap::setContainerRange@
//	members. The point d1 is mapped to the point i1, and d2 is mapped to i2. 
//	Any point inside or outside D can be mapped to a point inside or outside
//	I using @QpMap::transform@ or @QpMap::limTransform@ or vice versa
//	using @QwtPlot::invTransform@. D can be scaled linearly or
//	logarithmically, as specified with @QpMap::setDblRange@.
//
//.U PUBLIC MEMBERS
//.R
//
//    QpMap::QpMap (1) --	Constructor
//    QpMap::QpMap (2) -- Constructor
//    QpMap::~QpMap -- Destructor
//    QpMap::contains -- value contained in one interval?
//    QpMap::setContainerRange -- Specify container interval
//    QpMap::setContentRange -- Specify content interval
//    QpMap::transform   -- transform content double -> container double
//    QpMap::invTransform -- transform container -> content
//    QpMap::limTransform(double x) -- transform content -> container
//    QpMap::content1 -- first border of content interval 
//    QpMap::content2 -- second border of content interval
//    QpMap::container1 -- first border of container interval
//    QpMap::container2 -- second border of container interval
//    QpMap::logarithmic -- content double interval scaled logarithmically?
//
//.U  USAGE
//.c
//      #include <qpmap.h>
//
//      QpMap map;
//      double cval;
//      double Cval;
//
//      map.setContentRange(0.0, 3.1415);   // Assign an interval of type double with
//                                      // linear mapping
//      map.setContainerRange(0.0,100.0);         // Assign an container interval
//
//      Cval = map.transform(1.0);      // obtain container value corresponding to 1.0
//      cval = map.invTransform(77);    // obtain content value corresponding to 77
//.- 
//------------------------------------------------------------

//------------------------------------------------------------
//.U MEMBER FUNCTION DESCRIPTION
//------------------------------------------------------------

//------------------------------------------------------------
//
//.F	QpMap::QpMap (1)
//	Construct a QpMap instance.
//
//.u	Syntax
//.f	 QpMap::QpMap()
//
//.u	Description
//	The content and container intervals are both set to [0,1].
//
//------------------------------------------------------------
QpMap::QpMap()
{
  ContainerEmpty = true;
  ContentEmpty = true;
  content1 = DBL_MAX;
  content2 = -DBL_MAX;
  container1 = 0.0;
  container2 = 1.0; 
  d_cnv = 1.0;
  d_log = false;
}


//------------------------------------------------------------
//
//.F	QpMap::QpMap (2)
//	Construct a QpMap instance with initial container
//	and content intervals
//
//.u	Syntax
//.f	 QpMap::QpMap(double i1, double i2, double d1, double d2, bool logarithmic)
//
//.u	Parameters
//.p	double i1		--	first border of container interval
//	double i2		--	second border of container interval
//	double d1	--	first border of content interval
//	double d2	--	second border of content interval
//	bool logarithmic  -- logarithmic mapping, TRUE or FALSE. Defaults
//					to FALSE.
//
//------------------------------------------------------------
QpMap::QpMap(double _content1, double _content2, double _container1, double _container2, bool lg)
{
  /* Maybe this is better to use the initial values of
     content/container and to convert on the fly ?*/
  
    d_log = lg;
    container1 = container1_ = _container1;
    container2 = container2_ = _container2;
    content1_ = _content1;
    content2_ = _content2;
    ContainerEmpty = false;

    setContentRange(_content1, _content2,lg);
}

QpMap *QpMap::copy()
{
  return new QpMap(content1_,content2_,container1_,container2_,d_log);  
}


//------------------------------------------------------------
//
//.F	QpMap::~QpMap
//	Destroy a QpMap instance.
//
//.u	Syntax
//.f	 QpMap::~QpMap()
//
//------------------------------------------------------------
QpMap::~QpMap()
{
}

//------------------------------------------------------------
//
//.F	QpMap::contains 
//	Returns TRUE if a value x lies inside or at the border of the
//	map's content range.
//
//.u	Syntax
//.f	bool QpMap::contains(double x, char *interval)
//
//.u	Parameters
//.p	double x -- value 
//.p    char *interval -- concerned interval
//
//------------------------------------------------------------
/*bool QpMap::contains(double x, char *interval) const
{
  if (!strcmp(interval, "content"))
    return ( (x >= qwtMin(content1, content2)) && (x <= qwtMax(content1, content2)));
  else
    return ( (x >= qwtMin(container1, container2)) && (x <= qwtMax(container1, container2)));
}*/

//------------------------------------------------------------
//
//.F	QpMap::setContentRange
//	Specify the borders of the content interval 
//
//.u	Syntax
//.f	void QpMap::setContentRange(double d1, double d2, bool lg = FALSE)
//
//.u	Parameters
//.p	double d1	--	first border
//	double d2	--	second border
//	bool lg		--	logarithmic (TRUE) or linear (FALSE)
//				scaling. Defaults to FALSE.
//
//------------------------------------------------------------

void QpMap::setContentRange(double d1, double d2, bool lg) {
  // Upper limit must be greater than lower limit. 
  // Upper limit must be less than the maximum logarithmic value allowed
  if ((d2 >= d1) && (d2 < LogMax)) {
    if (lg) {      
      d_log = TRUE;
      if (d2 < LogMin) 
	d2 = 1000*LogMin;
      // If lower limit is zero or negative adjust to 1/1000 of upper limit
      if (d1 < LogMin) 
	d1 = d2/1000.;	
      content1 = log10(d1);
      content2 = log10(d2);
    } else {
	d_log = FALSE;
	content1 = d1;
	content2 = d2;
    }
    newFactor();
  }
  ContentEmpty = false;
}

void QpMap::setLog( bool lg)
{
  if (lg)    {
    // Switching from lin to log
    if (! d_log) 
      setContentRange(content1,content2,lg);
  }
  else if (d_log)
    {
      // Switching from log to lin
      d_log = FALSE;
      double d1 = pow(10,content1);
      double d2 = pow(10,content2);
      setContentRange(d1,d2,lg);
    }
  newFactor();
}

//------------------------------------------------------------
//
//.F	QpMap::setContainerRange
//	Specify the borders of the container interval
//
//.u	Syntax
//.f	void QpMap::setContainerRange(double i1, double i2)
//
//.u	Parameters
//.p	double i1	--	first border
//	double i2       --	second border
//
//------------------------------------------------------------
void QpMap::setContainerRange(double i1, double i2)
{
  ContainerEmpty = false;
  container1 = i1;
  container2 = i2;
  newFactor();
}



//------------------------------------------------------------
//
//.F	QpMap::transform
//	Transform a point in content interval into an point in the
//	container interval
//
//.u	Syntax
//.f	double QpMap::transform(double x)
//
//.u	Parameters
//.p	double x
//
//.u    Note
//	The specified point is allowed to lie outside the intervals. If you
//	want to limit the returned value, use @QpMap::limTransform@.
//------------------------------------------------------------
double QpMap::transform(double x) const
{
  if (d_log) {
    if (x > 0)
      return (container1 + (log10(x) - content1) * d_cnv );
    else
      return (container1 ); 
  }
    else
       return (container1 + (x - content1) * d_cnv );
}

//------------------------------------------------------------
//
//.F	QpMap::invTransform
//	Transform an container value into a content value
//
//.u	Syntax
//.f	double QpMap::invTransform(double y)
//
//.u	Parameters
//.p	double y	--	container value to be transformed
//
//.u	Return Value
//.t
//	linear mapping:	-- d1 + (d2 - d1) / (i2 - i1) * (y - i1)
//	logarithmic mapping: -- d1 + (d2 - d1) / log10(i2 / i1) * log10(y / i1)
//
//
//------------------------------------------------------------
double QpMap::invTransform(double y) const 
{
    if (d_cnv == 0.0)
       return 0.0;
    else
    {
      if(d_log)
	return pow(10.0, content1 + (y - container1) / d_cnv );
      else
	return ( content1 + (y - container1) / d_cnv );	
    }
}


//------------------------------------------------------------
//
//.F	QpMap::limTransform
//	Transform and limit
//
//.u	Syntax
//.f	double QpMap::limTransform(double x)
//
//.u	Parameters
//.p	double x
//
//.u	Return Value
//		transformed value
//
//.u	Description
//	The function is similar to @QpMap::transform@, but limits the input value
//	to the nearest border of the map's content interval if it lies outside
//	that interval.
//
//------------------------------------------------------------
double QpMap::limTransform(double x) const
{
    if ( x > qwtMax(content1, content2) )
       x = qwtMax(content1, content2);
    else if ( x < qwtMin(content1, content2))
       x = qwtMin(content1, content2);
    return transform(x);
}


double QpMap::widthTransform(double w) const
{
  // In the linear case the computation can be optimized as : w * d_cnv
  return (transform(content1 + w) - transform(content1));
}

double QpMap::invWidthTransform(double w) const
{
  return (invTransform(container1 + w) - invTransform(container1));
}


//------------------------------------------------------------
//.-
//.F	QpMap::newFactor
//	Re-calculate the conversion factor.
//------------------------------------------------------------
void QpMap::newFactor()
{
    if (content2 != content1)
       d_cnv = (container2 - container1) / (content2 - content1); 
    else 
       d_cnv = 0.0;
}

//------------------------------------------------------------
//
//.F	QpMap::d1
//	Returns the first border of the content interval
//
//.u	Syntax
//.f	double QpMap::d1()
//
//------------------------------------------------------------

//------------------------------------------------------------
//
//.F	QpMap::d2
//	Returns the second border of the content interval
//
//.u	Syntax
//.f	double QpMap::d2()
//
//------------------------------------------------------------

//------------------------------------------------------------
//
//.F	QpMap::i1
//	Returns the first border of the container interval
//
//.u	Syntax
//.f	double QpMap::i1()
//
//------------------------------------------------------------

//------------------------------------------------------------
//
//.F	QpMap::i2
//	Returns the second border of the container interval
//
//.u	Syntax
//.f	double QpMap::i2()
//
//------------------------------------------------------------


//------------------------------------------------------------
//
//.F	QpMap::logarithmic
//	Returns TRUE if the content interval is scaled logarithmically
//
//.u	Syntax
//.f	bool QpMap::logarithmic()
//
//------------------------------------------------------------




