#ifndef QPMARKER_H
#define QPMARKER_H

/** %%%
  Usage:

    The QpMarker class is the base class for markers drawn as a line,
  error line, triangle, rectangle,etc.

  TODO:

*/

#include "Qplotter/qp/qpobject.h"


class QpMarker : public QpObject
{
public:
  enum SymStyle { None, Ellipse, Rect, Diamond, Triangle, DTriangle, UTriangle,
		  LTriangle, RTriangle, Cross, XCross }; 
  enum LineStyle {NoLine, HLine, VLine, ErrorCross};

  /// Constructor without a marker style
  QpMarker(QpObject *z, bool addOnParent=false,
	   QpTransformation::tStrategy st=QpTransformation::Zone);
  /// Constructor with a marker style
  QpMarker(SymStyle ss, LineStyle ls, QpObject *z, bool addOnParent=false,
	   QpTransformation::tStrategy st=QpTransformation::Zone);
  /// Destructor
  virtual ~QpMarker() {}

  /// Get the marker symbol
  SymStyle getSymStyle()   const {return symStyle;}
  /// Get the line style
  LineStyle getLineStyle() const {return lineStyle;}
  /// Change X coordinate of marker
  void setX(double val) { x = val; }
  /// Change Y coordinate of marker
  void setY(double val) { y = val; }
  /// Gets the x coordinate
  double getX() const { return x; }
  /// Gets the y coordinate
  double getY() const { return y; }
  /// Gets the width
  double getW() const { return w; }
  /// Gets the height
  double getH() const { return h; }
  /// Set the symbol size (w and h). Used for FixedSizeMarkers.
  void setSymbolSize(double _w, double _h);
  /** Sets x,y coordinates, width and length of the object. By default, 
    the (x,y) point is the center of the rect. Center determines if the 
    (x,y) point is in the rect's up&left corner and must be moved. 
  */
  void setGeometry(double _x, double _y, double _w=0., double _h=0., bool center=true);
  /// Sets center and bounding rectangle of the error cross (or other lineStyle).
  void setLineGeometry(double _xc, double _yc, double _wc, double _hc);
  /// Draw the marker. Pure virtual method.
  virtual void draw() = 0;
  /// Print the marker. Pure virtual method.
  virtual void print() = 0;
  /// Change marker symbol
  void setSymStyle(SymStyle _style) {symStyle=_style;}
  /// Change line style
  void setLineStyle(LineStyle _style) {lineStyle=_style;}
  /// Is this object touched by this point ?
  virtual QpObject *touched(double _x, double _y, double errX, double errY, QpObject *&node);
protected:
  void drawSymbol(int _x, int _y, int w, int h, QPainter *p);
  void drawLineMarker(int _x, int _y, QPainter *p);
  QPointArray d_pa;
  SymStyle symStyle;
  LineStyle lineStyle;
  // Symbol center
  double x,y;
  // Symbol width,heigth
  double w,h;
  // Coords of upper left corner of the rectangle bounding the error bars
  double xc,yc;
  // Width, height of error bars
  double wc,hc;
};

/** Class to implement a scalable marker. A scalable marker is  a marker 
  with size (lenght/width) expressed in the Zone coordinate system.
*/

class QpScalableMarker : public QpMarker
{
public:
  /// Constructor without a marker style
  QpScalableMarker(QpObject *z, bool addOnParent=false,
	   QpTransformation::tStrategy st=QpTransformation::Zone) : 
    QpMarker(z, addOnParent, st ) {}
  /// Constructor with a marker style
  QpScalableMarker(SymStyle ss, LineStyle ls, QpObject *z, bool addOnParent,
	   QpTransformation::tStrategy st=QpTransformation::Zone) : 
    QpMarker(ss, ls, z, addOnParent , st) {}
  /// Draw the marker. 
  virtual void draw();
  /// Print the marker. 
  virtual void print();
protected:
  void paint(QPainter *painter);
};

/** Class to implement a fixed size marker. A fixed size marker is  a marker 
  with size (lenght/width) expressed in the pixel coordinate system. This means
  the marker will keep its size independently of the Zone coordinate system.
*/

class QpFixedSizeMarker : public QpMarker
{
public:
  /// Constructor without a marker style
  QpFixedSizeMarker(QpObject *z, bool addOnParent=false,
	   QpTransformation::tStrategy st=QpTransformation::Zone) : 
    QpMarker(z, addOnParent,st ){}
  /// Constructor with a marker style
  QpFixedSizeMarker(SymStyle ss, LineStyle ls, QpObject *z, bool addOnParent,
	   QpTransformation::tStrategy st=QpTransformation::Zone) : 
    QpMarker(ss, ls, z, addOnParent, st) {}
  QpObject *touched(double _x, double _y, double errorX, double errorY, QpObject *&node);
  /// Draw the marker. 
  virtual void draw();
  /// Print the marker. 
  virtual void print();
protected:
  void paint(QPainter *painter);
};


#endif

