#include <iostream>

#include "Qplotter/qp/qpmath.h"
#include "Qplotter/qp/qpmathbox.h"
#include "Qplotter/qp/qptextmetrics.h"

const char *entityTag = "EnTiTy";

/* 
   This is the set of mathematical entities supported. They correspond to
   similar entities defined in the W3C MathML DTD, but each entity correspond
   to a glyph in the Symbol font rather than to a Unicode character.
   We did not use the original DTD since the Qt XML parser can't include
   external definitions of entities.
   Each number is prefixed by an arbitrary keyword (EnTiTy) that allows to
   keep tract of the fact an entity was substituted by the parser. So for 
   instance the greek letter lambda (lowercase) correspond to:

   <!ENTITY lambda "EnTiTy 108" > 
   
   where 108 is the code of the Symbol glyph representing the letter.
*/

const char *entitiesDef = "<?xml version=\"1.0\"?> <!DOCTYPE qpmath [<!ENTITY alpha  \"EnTiTy 97\" ><!ENTITY beta   \"EnTiTy 98\" ><!ENTITY chi    \"EnTiTy 99\" ><!ENTITY delta  \"EnTiTy 100\" ><!ENTITY Delta  \"EnTiTy 68\" ><!ENTITY epsi   \"EnTiTy 101\" ><!ENTITY eta    \"EnTiTy 104\" ><!ENTITY gamma  \"EnTiTy 103\" ><!ENTITY Gamma  \"EnTiTy 71\" ><!ENTITY iota   \"EnTiTy 105\" ><!ENTITY kappa  \"EnTiTy 107 \" ><!ENTITY lambda \"EnTiTy 108\" ><!ENTITY Lambda \"EnTiTy 76\" ><!ENTITY mu     \"EnTiTy 109\" ><!ENTITY nu     \"EnTiTy 110\" ><!ENTITY omega  \"EnTiTy 119\" ><!ENTITY Omega  \"EnTiTy 87\" ><!ENTITY omicron \"EnTiTy 111\" ><!ENTITY phi    \"EnTiTy 102\" ><!ENTITY Phi    \"EnTiTy 70\" ><!ENTITY phiv   \"EnTiTy 106\" ><!ENTITY pi     \"EnTiTy 112\" ><!ENTITY Pi     \"EnTiTy 80\" ><!ENTITY piv    \"EnTiTy 118\" ><!ENTITY psi    \"EnTiTy 121\" ><!ENTITY Psi    \"EnTiTy 89\" ><!ENTITY rho    \"EnTiTy 114\" ><!ENTITY sigma  \"EnTiTy 115\" ><!ENTITY Sigma  \"EnTiTy 83\" ><!ENTITY tau    \"EnTiTy 116\" ><!ENTITY theta  \"EnTiTy 113\" ><!ENTITY Theta  \"EnTiTy 81\" ><!ENTITY thetav \"EnTiTy 74\" ><!ENTITY upsi   \"EnTiTy 117\" ><!ENTITY Upsi   \"EnTiTy 161\" ><!ENTITY xi     \"EnTiTy 120\" ><!ENTITY Xi     \"EnTiTy 104\" ><!ENTITY zeta   \"EnTiTy 122\" ><!ENTITY infin  \"EnTiTy 165\" ><!ENTITY int    \"EnTiTy 242\" ><!ENTITY mnplus \"EnTiTy 177\" ><!ENTITY nabla  \"EnTiTy 209\" ><!ENTITY sum    \"EnTiTy 229\" ><!ENTITY nbsp   \"EnTiTy 32\" ><!ENTITY larr   \"EnTiTy 174\" ><!ENTITY rarr   \"EnTiTy 172\" ><!ENTITY horbar \"EnTiTy 190\" ><!ENTITY prod   \"EnTiTy 213\" ><!ENTITY ape    \"EnTiTy 64\" ><!ENTITY ap     \"EnTiTy 187\" ><!ENTITY lt     \"EnTiTy 60\" ><!ENTITY gt     \"EnTiTy 62\" ><!ENTITY le     \"EnTiTy 163\" ><!ENTITY ge     \"EnTiTy 179\" ><!ENTITY exist  \"EnTiTy 36\" ><!ENTITY forall \"EnTiTy 34\" ><!ENTITY plusmn \"EnTiTy 177\" ><!ENTITY part   \"EnTiTy 182\" ><!ENTITY middot \"EnTiTy 183\" ><!ENTITY hArr   \"EnTiTy 220\" ><!ENTITY prop   \"EnTiTy 181\" ><!ENTITY radic  \"EnTiTy 214\" ><!ENTITY OverBar  \"EnTiTy 96\" >]>";


/*
 */
QpMath::QpMath(const char *value, QpObject *z, bool addOnParent ,
			     QpTransformation::tStrategy st) : 
  QpCommonText(entitiesDef,z,0.,0.,0.,0.,addOnParent,st) , doc ("qpMath") ,
  subSupFactor(defSubSupFact) 
{
  // Add the user MathML code to the entities definition
  appendData(value);
  // ??
  doc.setContent(getData());
  // Create dictionnary entries  string -> QpMath tags
  dict["msub"] = QpMath::msub; 
  dict["msup"] = QpMath::msup;
  dict["msubsup"] = QpMath::msubsup;
  dict["munder"] = QpMath::munder;
  dict["mover"] = QpMath::mover;
  dict["munderover"] = QpMath::munderover;
  dict["msqrt"] = QpMath::msqrt;
  // ??
  listOfBoxes = new QpMathBoxList(z,false,st);
  // Make "first version" of boxes...
  initFont(*getQpFont());
  makeBoxes(); 
}

/* Destructor method
 */ 
QpMath::~QpMath()
{
  delete  symsubsup ;
  delete  stdfont ;
  delete  stdsubsup;
  delete  symfont;
}

/* Create QFont variables with respect of size, get the size variables
 */
void QpMath::initFont (QFont &standardFont)
{
  stdfont = new QFont(standardFont);
  stdsubsup = new QFont(*stdfont);
  symfont = new QFont ("Symbol", stdfont->pointSize(), QFont::Normal, false
		       /*::Qt 2:: ,QFont::AnyCharSet*/);
  symsubsup = new QFont(*symfont);
  subSupSize = static_cast<int> (stdfont->pointSize()*subSupFactor);
  // Not too small
  if (subSupSize < 8)
    subSupSize = 8;
  
  stdsubsup->setPointSize(subSupSize);
  symsubsup->setPointSize(subSupSize);
  ySub = subSupSize/2;
  ySup = stdfont->pointSize()-subSupSize;
  yUnder = subSupSize;
  yOver = stdfont->pointSize();  
}

/* initFont must be called before makeBox one time only ! */
void QpMath::makeBoxes ()
{
  int x = 0;
  int y = 0; 
  listOfBoxes->clearList();
  makeBoxes (x,y,getRoot(),stdfont,symfont);
}

bool QpMath::twoArgs(QDomElement e, QDomElement& e1, QDomElement& e2)
{
  bool retval = false;
  if (e.childNodes().count() == 2) {
    QDomNode n1 = e.childNodes().item(0);
    e1 = n1.toElement();
    QDomNode n2 = e.childNodes().item(1);
    e2 = n2.toElement();	    
    retval = true;
  } else
    QPSTD::cerr <<"Expecting two elements for "<< e.tagName() <<QPSTD::endl;
  return retval;
}

bool QpMath::threeArgs(QDomElement e, QDomElement& e1, QDomElement& e2,
			 QDomElement& e3)
{
  bool retval = false;
  if (e.childNodes().count() == 3) {
    QDomNode n1 = e.childNodes().item(0);
    e1 = n1.toElement();
    QDomNode n2 = e.childNodes().item(1);
    e2 = n2.toElement();	    
    QDomNode n3 = e.childNodes().item(2);
    e3 = n3.toElement();	
    retval = true;
  } else
    QPSTD::cerr <<"Expecting three elements for "<< e.tagName() <<QPSTD::endl;
  return retval;
}

QDomNodeList QpMath::getRoot()
{
  return doc.childNodes();
}

void QpMath::nlist(QDomNodeList qnl )
{
  size_t i;
  QDomNodeList qnl2;
  for (i = 0; i<qnl.length();i++) {
    QDomNode n = qnl.item(i);
    QPSTD::cout << "Node "<< n.nodeName() << " Nodetype="  <<
      ((int)n.nodeType()) << QPSTD::endl;
    switch (n.nodeType()) {
    case QDomNode::TextNode :
      {	
	QPSTD::cout << "Text "<<n.nodeValue() <<QPSTD::endl;
	break;
      }
    case QDomNode::ElementNode :
      {
	QDomElement e = n.toElement();
	qnl2 = e.childNodes();
	nlist(qnl2);
	break;
      }
    default:
      break;
    }
  }
}

/** This method converts a  DOM element into QpBoxes. Depending on the element
  tag the method invokes "recursively" other makeBoxes() methods on the
  subelements of the element, e.g. in the case of subscript (QpMath::msub) it
  invokes makeBoxes() for the base and for the subscript part.
*/

void QpMath::makeBoxes(int &x, int &y, QDomElement& e, QFont* stdF, QFont* symF) {
  QDomElement e1,e2,e3;
  int xBase=x;
  QpMath::tags myElement = dict[e.tagName()];
  switch (myElement) {
  case QpMath::msqrt :
    {
      // BUG here with position
      // Lower superscript a bit so that they fit under the square root
      int oldySup = ySup;
      ySup -= subSupSize/4;

      // Create the skew (Symbol 214)
      QpTextBox *skew = new QpTextBox(symF,QString(QChar(214)),listOfBoxes);
      skew->setCoords(x,y);
      QFontMetrics qfm (*symF);
      x += qfm.width(QString(QChar(214)));
      int skewX = x;
      
      // Create the inside box
      size_t lastBoxes = listOfBoxes->count();
      makeBoxes(x,y,e.childNodes(),stdF,symF);

      // Get the width added by this box
      int boxWidth = listOfBoxes->getWidth(lastBoxes) ; // - skew->getWidth();
	
      // Overline this box
      QString *lineString = new QString();
      int lineLength;
      for (lineLength = 0;
	   qfm.width(*lineString) < boxWidth;
	   lineLength ++)
	lineString->append(QChar(96));

      // lineString->fill(QChar(96), lineLength);
      
      QpTextBox *overline = new QpTextBox(symF,*lineString,listOfBoxes);
      overline->setCoords(skewX,y);

      // Backup
      x = skewX + qfm.width(*lineString) ;
      ySup = oldySup;
      break;

      /*  Old code (Bug in positioning)
	  
      // Create the square root
      size_t curBoxes = listOfBoxes->count();
      size_t i = lastBoxes;
      QpRootBox *qrb = new QpRootBox(symF,boxWidth,listOfBoxes);

      while (i <curBoxes) {
	listOfBoxes->getElement(i)->shiftX(qrb->getSkew());
	i++;
      }

      qrb->setCoords(xBase,y);
      x = xBase+qrb->getWidth();
      //
      
      // Backup
      ySup = oldySup;
      break;
      */
    }
  case QpMath::msub :
    {
      if (twoArgs(e,e1,e2)) {
	makeBoxes(x,y,e1.childNodes(),stdfont,symfont);
	int yy = y+ySub;
	makeBoxes(x,yy,e2.childNodes(),stdsubsup,symsubsup);
      }
      break;
    }
  case QpMath::msup :
    {
      if (twoArgs(e,e1,e2)) {
	makeBoxes(x,y,e1.childNodes(),stdfont,symfont);
	int yy = y-ySup;
	makeBoxes(x,yy,e2.childNodes(),stdsubsup,symsubsup);
      }
      break;
    }
  case QpMath::msubsup :
    {
      if (threeArgs(e,e1,e2,e3)) {
	makeBoxes(x,y,e1.childNodes(),stdfont,symfont);
	int xSub = x;
	int yy = y+ySub;
	makeBoxes(xSub,yy,e2.childNodes(),stdsubsup,symsubsup);
	yy = y-ySup;
	makeBoxes(x,yy,e3.childNodes(),stdsubsup,symsubsup);
	if (xSub > x) 
	  x=xSub; 
      }
      break;
    }
  case QpMath::munder :
    {
      if (twoArgs(e,e1,e2)) {
	makeBoxes(xBase,y,e1.childNodes(),stdfont,symfont);
	int yy = y+yUnder;
	makeBoxes(x,yy,e2.childNodes(),stdsubsup,symsubsup);
	if (xBase > x) 
	  x=xBase;
      }
      break;
    }
  case QpMath::mover :
    {
      if (twoArgs(e,e1,e2)) {
	makeBoxes(xBase,y,e1.childNodes(),stdfont,symfont);
	int yy = y-yOver;
	makeBoxes(x,yy,e2.childNodes(),stdsubsup,symsubsup);
      }
      break;
    }
  case QpMath::munderover :
    {
      if (threeArgs(e,e1,e2,e3)) {
	int xBase=x;
	int xS=xBase;
	makeBoxes(x,y,e1.childNodes(),stdfont,symfont);
	int yy = y+yUnder;
	makeBoxes(xS,yy,e2.childNodes(),stdsubsup,symsubsup);
	xS=xBase;
	yy = y-yOver;
	makeBoxes(xS,yy,e3.childNodes(),stdsubsup,symsubsup);
      }
      break;
    }
  default:
    makeBoxes(x,y,e.childNodes(),stdF,symF);	  
  }
}

/** This method converts a list of DOM nodes into QpBoxes. If the
  node is a QDomNode::TextNode (i.e. a leaf of the tree) it is either 
  converted to a plain text box (QpTextBox) or to a symbol box (QpSymbolBox)
  if it corresponds to an entity.
  If the node is a QDomNode::ElementNode (node of the tree) it is 
  "recursively" translated by calling an overloaded version of makeBoxes ()
*/

void QpMath::makeBoxes(int &x, int &y, QDomNodeList qnl, QFont* stdF, QFont* symF )
{
  size_t i;
  for (i = 0; i<qnl.length();i++) {
    QDomNode n = qnl.item(i);
    switch (n.nodeType()) {
    case QDomNode::TextNode :
      {
	QString nVal  = n.nodeValue();
	if (nVal.contains(entityTag,true)) {
	  QpSymbolBox *qpb = new QpSymbolBox(symF,nVal,listOfBoxes);
	  qpb->setCoords(x,y);
	  x += qpb->getWidth();
	} else {
	  QpTextBox *qpb = new QpTextBox(stdF,nVal,listOfBoxes);
	  qpb->setCoords(x,y);
	  x += qpb->getWidth();
	}	  
	break;
      }
    case QDomNode::ElementNode :
      {
	QDomElement e = n.toElement();
	makeBoxes(x,y,e,stdF,symF);
      }
    default:
      break;
    }
  }
}

void QpMath::prepareBoxes(int& xCoord,int& yCoord, QPainter *pai) {
  makeBoxes();
  // Transform coords in pixel 
  QpTransformation *transf = getTransformation(getTransfStrategy());  
  xCoord = transf->xTransform(getX());
  yCoord = transf->yTransform(getY());

  // Alignments
  int boundWidth = transf->xTransform(getX() + getW()) - xCoord;
  int boundHeight = yCoord - transf->yTransform(getY() + getH());
  // Rotate painter if necessary
  rotatePainter(pai,xCoord,yCoord);
  xCoord += getXOffset(boundWidth, getPainter());
  yCoord += getYOffset(boundHeight, getPainter());
}

void QpMath::draw() {
  size_t i;
  int xCoord;
  int yCoord;
  prepareBoxes(xCoord,yCoord, getPainter());
  for (i=0; i < listOfBoxes->count(); i++) {
    listOfBoxes->getElement(i)->shiftX(xCoord);
    listOfBoxes->getElement(i)->shiftY(yCoord);
    listOfBoxes->getElement(i)->draw();
  }
  // Restore painter if necessary
  restorePainter(getPainter());
}

void QpMath::print() {
  size_t i;
  int xCoord;
  int yCoord;
  prepareBoxes(xCoord,yCoord,getPrinter());
  for (i=0; i < listOfBoxes->count(); i++) {
    listOfBoxes->getElement(i)->shiftX(xCoord);
    listOfBoxes->getElement(i)->shiftY(yCoord);
    listOfBoxes->getElement(i)->print();
  }
  // Restore painter if necessary
  restorePainter(getPrinter());
}
