/*
  Usage:

   The QpNode class represents a list of QpObject.
   It is used by QpZone, QpPage, QpPolyMarker, ... Each class 
   of this list is a list of QpObject :
   e.g : QpPage is a list of QpZone, 
         QpZone is a list of QpObject, ...

   This class uses a template class as content. So every node 
   will be a list of a given type.

   QpNode uses a container class of QT : QPtrList.
   Graphical attributes are kept in QpNode, so if you want some
   attributes on an object, you should assign them to its direct 
   QpNode parent. 
   QpZone, QpPage, QpPolyMarkers, QpAxis, ... inherit from QpNode.

   This class is stocked in a .cpp (cause of its template use).

  TODO:

    (!!) Non thread-safe using list.next()
    (!!) QPtrList Sorting does not keep the initial order for objects
    with the same priority (zone in QpPage)
    -> Add a class container of same class elements and sort it instead of elements ?
    (!!) Review the coupling between visual attributes and nodes. Review
     graphics attributes/node problem (some class are just like polymarker
     are just graphics container)  In general these are classes like QpNode <type>
     with type !=  qpobject
    
*/


#ifndef QPNODE_H
#define QPNODE_H

 
#include <qptrlist.h>
#include <qdict.h>
#include "Qplotter/qp/qpobject.h"

#include <iostream>


class ElementList : public QPtrList<QpObject> {
public:
  ElementList(QDict<int> p)
  {
    priority = p;
  }
  
  ~ElementList() {;}
  
  int compareItems( QCollection::Item el1, QCollection::Item el2)
  {
    QpObject *obj1 = (QpObject *) el1;
    QpObject *obj2 = (QpObject *) el2;

    
    if (!priority[obj1->isA()])
      {
	std::cout << "ElementList : Unknown qpobject type :" << obj1->isA() << std::endl ;
	return 1;
      }

    if (!priority[obj2->isA()])
      {
	std::cout << "ElementList : Unknown qpobject type :" << obj2->isA() << std::endl ;
	return -1;
      }

    int p1, p2;

    p1 = *priority[obj1->isA()];
    p2 = *priority[obj2->isA()];

    if (p1 != p2)
      return p1 - p2;
    else
      {
	// If this is the same kind of object
	// Then we must keep the chronological order
	return 0;
      }
    
  }

  void addPriority(QString objectName, int pr)
    {
      // You have to sort the list after this call if needed
      priority.insert(objectName, new int(pr));    
    }

  private:
  QDict<int> priority;
};




template <class Element>
class QpNode : public QpObject
{
public:
  /// Constructor
  QpNode(QString *t, QpObject *p=0, bool addOnParent=true,
	 QpTransformation::tStrategy st=QpTransformation::Zone) : 
    QpObject(t,p,addOnParent,st), list(QDict<int>(173)) {
    init();
  }
  /** Destructor
      Warning : when the auto-delete list's option is activated, every
      component of the list is deleted at list destruction.
  */
  virtual ~QpNode() {delete pen; delete brush ;delete mFont;}

  /// Selectors for attributes managing
  virtual QPen     *getQpPen() const;
  virtual QFont    *getQpFont() const;
  virtual QBrush   *getQpBrush() const;

  virtual void setQpPen(QPen p);
  virtual void setQpFont(QFont f);
  virtual void setQpBrush(QBrush b);
  virtual void setFontSize(int s);
  virtual int  getFontSize();
  
  /// Returns the list
  //ElementList *getList() {return &list;}
  /// Insert an element at a fixed position in the list
  //  virtual void insertElement(const Element *obj, int position=0) { list.insert(position, obj); }
  virtual void insertElementFirst(const Element *obj)
  {
    list.insert(0, obj);
  }
  /// Adds an element to the list
  virtual void addElement(const Element *obj)
  {
    list.inSort(obj);
    
    /*
      //  inSort() is wrong as it inserts the new element in front of
      //  other elements with the same priority.
      uint index ;
    
      for (index = 0;
      index < list.count();
      index ++)
      {
      if (list.compareItems((void *) list.at(index),
      (void *) obj) > 0)
      {
      list.insert(index, obj );
      break;
      }
      }

      if (index == list.count())
      {
      list.insert(index, obj );
      }
    */
  }
  /// Removes the obj element from the list
  virtual void removeElement(Element *obj)    { list.removeRef(obj); }
  /// Set the AutoDelete list's flag
  virtual void setAutoDelete(bool a)    { list.setAutoDelete(a); }
  /// Clear the list
  virtual void clearList()    { list.clear(); }

  virtual unsigned int count()    { return list.count(); }
  /// Returns the element at index position 
  virtual Element* const getElement(int index) {return 	(Element *) list.at(index);}
  QPtrList<QpObject> /*const*/ search(QString name) 
  {
    QPtrList<QpObject> found;
    QPtrList<QpObject> subsearch;
	
    if (isA(name))
      found.append(this);    
	    
    for (list.first();
	 list.current() != 0;
	 list.next())
      {
	subsearch = (list.current())->search(name);

	for (subsearch.first();
	     subsearch.current() != 0;
	     subsearch.next())
	  {
	    found.append(subsearch.current());	    
	  }
      }
    
    return found ;    
  }
  
  /** Search the first element (from the list) for which the function will return TRUE
      See qsort() help to find how to use pointer on function. */
  virtual Element* const searchElement(bool function(Element*))
  {
    for (list.first();
	 list.current() != 0;
	 list.next())
      {
	if (function((Element*) list.current()))
	  return (Element*) list.current();
      }

    return (Element *)0;
  }

  virtual QPtrList<QpObject> findType(QString typeName)
  {
    QPtrList<QpObject> result ;
    
    for (list.first();
	 list.current() != 0;
	 list.next())
      {
	if ((list.current())->isA() == typeName)
	  {
	    result.append(list.current());
	  }
      }

    return result ;
  }

  QpObject* at ( int index ) const
  {
    QPtrListIterator<QpObject> iGO(list); 
    iGO+=index;
    return iGO.current();
  }

  virtual void print();
  virtual void printList(QString className="");
  virtual void draw();
  virtual void drawList(QString className="");
  

  /*
  /// Calling print on a QpNode will call the print signal on every element of its list
  virtual void print()
  {
    QPtrListIterator<QpObject> iGO(list); 
    for( ; iGO.current(); ++iGO ) {
      iGO.current()->print();
    }
  }
  
  /// Calling draw on a QpNode will call the draw signal on every element of its list
  virtual void draw()
  {
    QPtrListIterator<QpObject> iGO(list); 
    for( ; iGO.current(); ++iGO ) {
      iGO.current()->draw();
    }
  }
  */
  
  /// Is the node or one of its objects touched by this point ?
  virtual QpObject *touched(double x, double y, double errorX, double errorY, QpObject *&node);
  
  /// Change the node visual properties if necessary
  virtual void changeStyle(QPainter *painter);
  /// Restore the node visual properties if necessary
  virtual void restoreStyle(QPainter *painter);

  // Show all the subnode of a node
  virtual void treeWalking() 
  {
    std::cout << isA() << ":" << std::endl ;
    
    for (list.first();
	 list.current() != 0;
	 list.next())
      {
	std::cout << "Subnode is " << (list.current())->isA() << std::endl;	
      }    
  }
  
  
protected:
  /// Variables initialisation 
  void init() {
    pen=0; 
    mFont=0; 
    brush=0;

    list.setAutoDelete( TRUE ); // Beware of that..
    
    list.addPriority(QString("QpZone"),-10);
    list.addPriority(QString("QpPage"),-10);
    list.addPriority(QString("QpCurve1D"),0);
    list.addPriority(QString("QpCurve2D"),0);
    list.addPriority(QString("QpCurve3D"),0);
    list.addPriority(QString("QpScatterPlot2D"),0);
    list.addPriority(QString("QpGrid"),100);
    list.addPriority(QString("QpAxisNode"),200);
    list.addPriority(QString("QpAxis"),200);
    list.addPriority(QString("QpSummaryInfo"),30);
    list.addPriority(QString("QpSummaryBox"),30);
    list.addPriority(QString("QpFace"),30); // 30 ? 50 ?
    list.addPriority(QString("QpPolyFace"),30); // 30 ? 50 ?
    list.addPriority(QString("QpMultiLine"),40);
    list.addPriority(QString("QpPolyLine"),40);
    list.addPriority(QString("QpPolygon"),40);
    list.addPriority(QString("QpNode<QpPolyShape>"),50);
    list.addPriority(QString("QpPolyMarker"),60);
    list.addPriority(QString("QpMarker"),60);
    list.addPriority(QString("QpText"),70);
    list.addPriority(QString("QpCommonText"),70);
    list.addPriority(QString("QpMathBoxList"),70);
    list.addPriority(QString("QpMathBox"),70);
  
  }
  QPen    *pen;    /* Optional drawing/printing pen   */
  QFont   *mFont;  /* Optional drawing/printing font  */
  QBrush  *brush;  /* Optional drawing/printing brush */

private:
  ElementList list;

};


template <class Element> QPen* QpNode<Element>::getQpPen() const
{
  return (pen) ? pen : getParent()->getQpPen();
}

template <class Element> void QpNode<Element>::setQpPen(QPen p)
{
  delete pen;
  pen = new QPen(p); 
}

template <class Element> void QpNode<Element>::setQpBrush(QBrush b)
{
  delete brush;
  brush = new QBrush(b); 
}

template <class Element> QBrush* QpNode<Element>::getQpBrush() const
{
  return (brush) ? brush : getParent()->getQpBrush();
}

template <class Element> QFont* QpNode<Element>::getQpFont() const
{
  return (mFont) ? mFont : getParent()->getQpFont();
}

template <class Element> void QpNode<Element>::setQpFont(QFont f)
{
  QFont *mFontNew = new QFont(f);
  if (mFontNew != 0) {
    delete mFont;
    mFont = mFontNew;
  }
}

template <class Element> void QpNode<Element>::setFontSize(int s)
{
  getQpFont()->setPointSize (s);
}

template <class Element> int QpNode<Element>::getFontSize()
{
 return  getQpFont()->pointSize ();
}

template <class Element> void QpNode<Element>::changeStyle(QPainter *painter)
{
  if (pen != 0)   
    painter->setPen(*pen);
  if (brush != 0) 
    painter->setBrush(*brush);
  if (mFont != 0)  
    painter->setFont(*mFont);
}

template <class Element> void QpNode<Element>::restoreStyle(QPainter *painter)
{
  if (getParent()) {
    if ((pen != 0) && getParent()->getQpPen()) 
      painter->setPen(*(getParent()->getQpPen()));
    if ((brush != 0) && getParent()->getQpBrush()) 
      painter->setBrush(*(getParent()->getQpBrush()));
    if ((mFont != 0) && getParent()->getQpFont()) 
      painter->setFont(*(getParent()->getQpFont()));
  }
}

template <class Element> void QpNode<Element>::draw()
{
  changeStyle(getPainter());

  for (list.first();
       list.current() != 0;
       list.next())
    {
      (list.current())->draw();
    }
  
  restoreStyle(getPainter());
}

template <class Element> void QpNode<Element>::drawList(QString className)
{
  //  changeStyle(getPrinter());
  
  for (list.first();
       list.current() != 0;
       list.next())
    {
      // By default an empty string means to draw everything
      
      if ((className == "") || (list.current()->isA() == className))
	list.current()->draw();
    }

  //  restoreStyle(getPainter());
}

template <class Element> void QpNode<Element>::print()
{
  changeStyle(getPrinter());

  for (list.first();
       list.current() != 0;
       list.next())
    {
      list.current()->print();
    }
  
  restoreStyle(getPrinter());
}

template <class Element> void QpNode<Element>::printList(QString className)
{
  changeStyle(getPrinter());

  for (list.first();
       list.current() != 0;
       list.next())
    {
      if ((className == "") || (list.current()->isA() == className))
	list.current()->print();
    }
  
  restoreStyle(getPrinter());
}

template <class Element> QpObject *QpNode<Element>::touched(double x, double y, double errorX, double errorY, QpObject *&node)
{
  QpObject *res = 0 ;

  for (list.last();
       list.current() != 0;
       list.prev())
    {
      res = (list.current())->touched(x,y,errorX,errorY,node);
      if (res)
	break;      
    }
  
  return res;
}

#endif



