#ifndef QPOBJECT_H
#define QPOBJECT_H

#if defined( __sun )  && __SUNPRO_CC < 0x500
	#define true 1
	#define false 0
	typedef int bool;
#endif

/** %%%

    QpObject : Base (virtual) class of Qplotter.
    
  Usage:

   This object defines all common functions for every object of Qplotter.
   It is also used to simulate some recursive calls... A QpObject has got 
   a parent QpObject (e.g. a polyline can have its zone has a direct parent)
   which is its direct neighbour in the object hierarchy.

   The QpObject is linked to its parent during its construction. However,
   this is not true for QpZones which are linked during the first drawing 
   event.

   The QpObject's functions are virtual default functions (not pure virtual 
   functions) or "recursives" one made to get information on other objects 
   in the hierarchy.
   
   e.g. : you can have a QpPage containing two QpZones. In one of these QpZone,
   there is a QpObject (it may be a QpPolyLine or something like this). Calling
   getZone() on this QpObject will call the same method from his parent. The 
   QpZone object will return a pointer on itself.

   Moreover, these functions are really simple and mostly kept inline.

   TODO:

   (--) decouple between graphics properties and hierarchy needed
   (--) VisualNode, object set, DrawingGroup (node with only one class type)
   and shared GraphicsAttribute
   (--) Style class (XML ?) Each object must be able to describe itself
   to rebuild it if needed
   (--) Move/Select object, manipulate it graphically
*/


#include <iostream>

#include <qpainter.h>
#include <qstring.h>
#include <qptrlist.h>
#include <qdict.h>
#include "Qplotter/qp/qptransformation.h"
class QpZone;
class QpPage;
class SummaryTable;


class QpObject
{
 public:
  /// Constructor
  QpObject(QString *_type, QpObject *_parent=0, bool addOnParent=true,
	   QpTransformation::tStrategy st=QpTransformation::Zone);
  /// Destructor (virtual)
  virtual ~QpObject()
    {
      delete type;
    }
  /// The drawing function
  virtual void draw() { }
  /// The printing function
  virtual void print() { }
  /// Used to identify an object found in a list
  virtual QString isA() const
    {
      return *type;
    }
  /// Return Parent type
  virtual QString isChild() const         {return fatherType;}
  /// Is this object of this type ?
  virtual bool isA(QString t)         {return t == *type;}
  /// Search facilities
  QPtrList<QpObject> const search(QString name)
    {
      QPtrList<QpObject> result ;

      if (isA(name)) 	
	result.append(this);
      
      return result;
    }
  
    
  virtual QpObject *getParent() const {return oparent;}
  /// Recursive function to retrieve the zone the object is in
  virtual QpZone   *getZone() const
    {return (getParent()) ? getParent()->getZone() : 0;}
  /// Recursive function to retrieve the page the object is in
  virtual QpPage   *getPage() const
    {return (getParent()) ? getParent()->getPage() : 0;}
  /// Recursive function to retrieve the painter
  virtual QPainter *getPainter() const
    {return (getParent()) ? getParent()->getPainter() : 0;}
  /// Recursive function to retrieve the printer
  virtual QPainter *getPrinter() const
    {return (getParent()) ? getParent()->getPrinter() : 0;}
  /// Recursive function to retrieve the print device
  virtual QPrinter * const getPrintDevice()
    {return (getParent()) ? getParent()->getPrintDevice() : 0;}
  /// Recursive function to get the pen
  virtual QPen  *getQpPen() const
    {return (getParent()) ? getParent()->getQpPen() : 0;}
  /// Recursive function to get the brush
  virtual QBrush  *getQpBrush() const
    {return (getParent()) ? getParent()->getQpBrush() : 0;}
  /// Recursive function to get the font
  virtual QFont *getQpFont() const
    {return (getParent()) ? getParent()->getQpFont() : 0;}
  /// Used for relative coordinates
  enum QpRelative 
  { BottomLeft, BottomRight, TopLeft, TopRight , Left, Right, Top, Bottom };

  /** Determine the point [x +/- errorX, y +/- errorY] is in the object
      The node pointer will return a reference on the "touched object's" parent
  */
  virtual QpObject *touched(double /* x */, double /* y */, double /* errorX */, double /* errorY */, QpObject *&node) { 
    node=(QpObject*)getZone(); return 0; }
  /// Function to get the transformation strategy
  virtual QpTransformation::tStrategy getTransfStrategy() const {
    return strat;
  }

  /*
    void describeHierarchy()
    {
    cout << "This is a " << isA() << std::endl ;
    if (getParent()) {      
    std::cout << "Parent is " << std::endl ;
    getParent()->describeHierarchy() ;	
    } else {
    std::cout << "I am a root" << std::endl ;	
    }      
    }
  */

  virtual void selfDescribe(std::ostream &output) const
    {
      output << "(QpObject " << std::endl ;
      output << "(type " << type << " )" << std::endl ;
      output << "(parent " << std::endl ;
      if (oparent)
	oparent->selfDescribe(output);
      else
	output << "0" ;
      output << ")" << std::endl ;
      output << "(strategy " ;
      switch (strat) {
      case QpTransformation::Zone:
	output << "Zone" << std::endl ;
	break;
      case QpTransformation::Page:
	output << "Page" << std::endl ;
	break;
      case QpTransformation::ZrBottomLeft:
	output << "ZrBottomLeft" << std::endl ;
	break;
      case QpTransformation::ZrBottomRight:
	output << "ZrBottomRight" << std::endl ;
	break;
      case QpTransformation::ZrTopLeft:
	output << "ZrTopLeft" << std::endl ;
	break;
      case QpTransformation::ZrTopRight:
	output << "ZrTopRight" << std::endl ;
	break;
      default:
	output << "?? " << strat << std::endl ;
      }
      output << ")" ;
      output << ")" << std::endl ;
    }
  
  
 protected:
  /// Set the parent
  virtual void setParent(QpObject* _parent) 
    {if (oparent!=_parent) oparent=_parent;}
  /// Function to set the transformation strategy
  virtual void setTransfStrategy(QpTransformation::tStrategy s) {strat=s;}
  /// Recursive function to get the transformation object
  virtual QpTransformation *getTransformation(QpTransformation::tStrategy s) const
    {return (getParent()) ? getParent()->getTransformation(s) : 0;}

 private:
  QString *type;
  QString fatherType;
  
  QpObject *oparent;
  QpTransformation::tStrategy strat; 

};



#endif
