#ifndef QPPAGE_H
#define QPPAGE_H

/** %%%

    Page class

  Usage:

   A container for zones with its own system of coordinates.
  The page's system of coordinates can be in centimeters or 
  any unit not linked to pixels.
   
  TODO:

*/


#include <iostream>
#include <qwidget.h>
#include <qpixmap.h>
#include <qprinter.h>
#include <qarray.h>

#include "Qplotter/qp/qpzone.h"
#include "Qplotter/qp/qplistener.h"
#include "Qplotter/qp/qptext.h"


class QpPage : public QWidget,  public QpNode<QpObject>
{ 
  Q_OBJECT
    public:
  /** Constructor. Page dimensions corresponds to the largest square
      that fits in an A4 paper (in millimeters)
  */
  QpPage(QWidget *_w, QRect geo, double wp=210.0, double hp=210.0);
  /* [WARNING]
     If qpPage is deleted when Outline is on, there is a warning at
     exit from Qt
  */
  ~QpPage();
  /// Retrieve page width
  double   getWidth()   const    {return w;}
  /// Retrieve page height
  double   getHeight()  const    {return h;}
  /// Retrieve screen page width
  double   getScreenWidth()   const    {return screenWidth;}
  /// Retrieve screen page height
  double   getScreenHeight()  const    {return screenHeight;}
  /// Selectors
  /// Retrieve the zone indexed by index
  virtual QpZone   * const getZone(int index) const; 
  /// Recursive function which should not be called here. It returns the current zone to avoid null problems
  virtual QpZone   * getZone() const  {return getZone(0);}
  /// Retrieve pen (close "recursion")
  virtual QPen     *getQpPen()   const {return QpNode<QpObject>::pen;}
  /// Retrieve brush (close "recursion")
  virtual QBrush   *getQpBrush() const {return QpNode<QpObject>::brush;}
  /// Retrieve font (close "recursion")
  virtual QFont    *getQpFont()  const {return QpNode<QpObject>::mFont;}
  /// Retrieve painter
  virtual QPainter *getPainter() const             {return painter;}
  /// Retrieve printer
  virtual QPainter *getPrinter() const             {return printer;}
  /// Retrieve page
  virtual QpPage   *getPage()    const             {return (QpPage * const)this;}
  /// Retrieve print device and creates it if it was null
  virtual QPrinter * const getPrintDevice();
  /// Retrieve transformation
  virtual QpTransformation *
    getTransformation(QpTransformation::tStrategy s) const;
  /// Retrieve the underlying pixmap
  QPixmap *getBuffer() const {return _buffer;}
  /// Set the centered title of the zone
  QpText *setTitle(QString _title, int _align=Qt::AlignCenter);
  /// Set a text in the x, y coordinate.
  QpText *addText(double xPosition,double yPosition,QString text);
  /// Clear the pixmap
  void clear() { _buffer->fill(white); }
  /// Update the event listener infos
  void updateListener(QEvent *ev);
  /// Return a pointer on the QpListener object
  QpListener *getListener() const {return (QpListener * const)(listener);}
  /// Use cross outline
  void crossOutline(bool On = true)
    {
      if (On) {      
	outline->setType(QpOutLine::Cross);
      } else {
	outline->setType(QpOutLine::Off);
      }        
    }
  /// Activate or desactivate picking
  void setPicking(bool activate=true) { picking = activate; }
  /// Activate or desactivate zooming
  void setZooming(bool activate=true, bool fixedstate=false, QpZoom::State state=QpZoom::Zoom);
  /// Set finePicking policy
  void setFinePicking (bool policy) { finePicking = policy;}

  void addElement(const QpObject *obj)
    {
      // Todo
      // If a zone is added keep track on it
      QpNode<QpObject>::addElement(obj);
    }
  
  
  /// Set number of zones in X
  void setNbXZones(int nbX)     {nbXZones=nbX;}
  /// Set number of zones in Y
  void setNbYZones(int nbY)     {nbYZones=nbY;}
  /// Get number of zones in X
  int getNbXZones() const    {return nbXZones;}
  /// Get number of zones in Y
  int getNbYZones() const    {return nbYZones;}
  /// Set pen
  virtual void setQpPen(QPen p);
  /// Set font
  virtual void setQpFont(QFont f);
  /// Set brush
  virtual void setQpBrush(QBrush b);
  /// Print with a fixed ratio (true) or with the draw ratio
  void useFixedRatio(bool fixed=false) { fixedRatio=fixed; }

  /// Creates nbX * nbY zones, symmetrical placement 
  void createZones(int nbX, int nbY);
  /// Create a non-regular zone
  QpZone* createIrregularZone(double x, double y,
			      double w, double h);
  
  /** Replace the "index" zone with a new zone created
      with the same caracteristics **/
  void flushZone(int index);
  // Redraw this zone only
  void redrawZone(QpZone *z);

  /// Lauch the draw with the stocked painter and dest.
  void redraw();
  /// Draw the page
  void draw();
  /// Save the page to a picture file (PNG format only)
  bool save(const QString& file,const char* format="PNG");  
  /// Print the page
  void print(QString fichier, bool setup=false);
  /// Print the page
  void print() { print(QString("default.ps")); }
  // Setup the printer
  //void setupPrinting();

 private:
  /// Initialization
  void init();
  void initPainting(int geo_w,int geo_h);
  void initPage(int w,int h);
  // Check which zone exist at these coordinates
  QpZone *touchedZone(double _x, double _y); 
  // Pointer to the page title 
  QpText *pagetitle;
  // Size in page coordinates
  double w,h;
  double screenWidth, screenHeight;
  
  bool locate, picking, fixedRatio;
  QpListener *listener;
  QpOutLine *outline;
  QpZoom *zoom; 
  QPrinter *printdevice;  /** QPrinter for Postscript                   */
  QPainter *printer;      /** QPainter used to print                    */
  QPainter *painter;      /** QPainter used to draw                     */
  /// Zones / margins variables
  int nbXZones, nbYZones;
  QArray<QpZone *> zones ;
  
  
  double leftMargin, bottomMargin;
  /// Transformation object
  PageStrategy *transformation;
  int originalXmin, originalXmax, originalYmin, originalYmax;
  
  QPixmap *_buffer;
  /**Should the page collect information on the picked object or just
     coordinates? Fine picking can be very slow with large scenegraph 
     (eg 500X500  box plot) and is not really useful for locating/zooming.
     By default fine picking is disabled.
  */							       
  bool finePicking;

  void paintEvent(QPaintEvent*);
  void resizeEvent(QResizeEvent*);
  void mouseMoveEvent(QMouseEvent *e);
  void mousePressEvent(QMouseEvent *e);
  void mouseReleaseEvent(QMouseEvent *e);

  public slots:
    void slotRefresh() { redraw(); }
  void slotZoom()    { setZooming(true,false,QpZoom::Zoom); } // Transient zoom (zoom only once)
  void slotZooming(bool zooming)  // zooming/unzooming state
    {
      if (zooming) {
	setZooming(true,true);	
      } else {
	setZooming(false,true);	
      }            
    }        
  void slotUnZoom()  { setZooming(true,false,QpZoom::UnZoom); }
  void slotPrint()   { print(QString("example.ps"),true); }
  void slotSelect(bool On)   {
    if (On) {      
      outline->setType(QpOutLine::Cross);
    } else {
      outline->setType(QpOutLine::Off);
    }        
  }
};


#endif

