#include "Qplotter/qp/qpzone.h"
#include "Qplotter/qp/qppolyline.h"
#include "Qplotter/qp/qpsmoother.h"
#include "Qplotter/qp/qptransformation.h"
#include "Qplotter/qp/qpqwt_math.h"

QpPolyShape::QpPolyShape(const char *className, int nbElements, double *_x, 
			 double *_y, QpObject *z, bool addOnParent,
			 QpTransformation::tStrategy st)
: QpObject(new QString(className),z,addOnParent,st) 
{
  boolSmooth = 0;
  nb=nbElements;
  setCoords(nbElements,_x,_y);
}

/// Destructor
QpPolyShape::~QpPolyShape()
{ 
  delete [] x; 
  delete [] y; 
}

void QpPolyShape::setCoords(int nbElements, double *_x, double *_y)
{
  nb = nbElements;
  x= new double [nb]; 
  y= new double [nb];
  for(int i=0;i<nb;i++) {
    x[i]=_x[i]; 
    y[i]=_y[i];
  }
}

void QpPolyShape::paint(QPainter *painter)
{
  QpTransformation *transf = getTransformation(getTransfStrategy());  
  int i;
  if (boolSmooth) {
    QpSmoother mySmoother(getNb(),x,y);
    int npt = mySmoother.smooth();
    if (npt>1) 
      smoothDraw(painter,mySmoother.getX(),mySmoother.getY(),npt);
  } else {
    QPointArray *data = new QPointArray(getNb());
    for(i=0;i<getNb();i++) 
      data->setPoint(i,transf->xTransform(getX(i)), 
		     transf->yTransform(getY(i)));
    basicDraw(painter,*data);  
    delete data;
  }
}

void QpPolyShape::draw()
{
  paint(getPainter());
}

void QpPolyShape::print()
{
  paint(getPrinter());
} 

void QpPolyShape::computeCoeffs(double x1, double y1, double x2, double y2, double *a, double *b)
{
  *a = (y2-y1)/(x2-x1);
  *b = y1 - (*a * x1);
}

QpObject *QpPolyShape::touched(double _x, double _y, double errorX, double errorY, QpObject *& /* node */)
{
  double x1, y1, x2, y2, a, b;
  double xz, yz, errorXz, errorYz;
  ZoneStrategy *t = (ZoneStrategy*)getZone()->getTransformation(QpTransformation::Zone);
  xz = t->getXzMap()->invTransform(_x);
  errorXz = t->getXzMap()->invWidthTransform(errorX) ;
  yz = t->getYzMap()->invTransform(_y);
  errorYz = t->getYzMap()->invWidthTransform(errorY) ;

  x1 = getX(0); y1 = getY(0);
  for(int i=1;i<getNb();i++) {
    x2 = getX(i); y2 = getY(i);
    if ((x1==x2) && (fabs(x1-xz)<=errorXz)) {
      if ((yz-errorYz <= qwtMax(y1,y2)) && (yz+errorYz >= qwtMin(y1,y2))) 
	return this;
    }
    else if ((y1==y2) && (fabs(y1-yz)<=errorYz)) {
      if ((xz-errorXz <= qwtMax(x1,x2)) && (xz+errorXz >= qwtMin(x1,x2))) 
	return this;
    }
    else if ((qwtMin(x1,x2) <= xz+errorXz) && (xz-errorXz <= qwtMax(x1,x2))) {
      computeCoeffs(x1,y1,x2,y2,&a,&b);
      if (fabs(yz-(a*xz+b)) <= errorYz) return this;
    }
    x1=x2; y1=y2;
  }
  return 0;
}

QpPolyLine::QpPolyLine(int nbElements, double *_x, double *_y, QpObject *z, 
		       bool addOnParent,QpTransformation::tStrategy st) : 
  QpPolyShape("QpPolyLine",nbElements,_x,_y,z,addOnParent,st) 
{
}

void QpPolyLine::basicDraw(QPainter *painter,QPointArray &qpa) {
    painter->drawPolyline(qpa);  
}

void QpPolyLine::smoothDraw(QPainter *painter,double *_x,double *_y, int npt)
{
  QpTransformation *transf = getTransformation(getTransfStrategy());  
  QPointArray *data = new QPointArray(npt);
  for(int i=0;i<npt;i++) {
    data->setPoint(i,transf->xTransform(_x[i]),transf->yTransform(_y[i]));
  }
  painter->drawPolyline(*data);  
  delete data;
}

QpPolygon::QpPolygon(int nbElements, double *_x, double *_y, QpObject *z, 
		       bool addOnParent, QpTransformation::tStrategy st) : 
  QpPolyShape("QpPolygon",nbElements,_x,_y,z,addOnParent,st) 
{
}

void QpPolygon::basicDraw(QPainter *painter,QPointArray &qpa) {
    painter->drawPolygon(qpa, false);  
}

void QpPolygon::smoothDraw(QPainter *painter,double *_x,double *_y, int npt) 
{
  int i;
  QpTransformation *transf = getTransformation(getTransfStrategy());  
  QPointArray *data = new QPointArray(npt+2);
  for( i=0;i<npt;i++) {
    data->setPoint(i,transf->xTransform(_x[i]),transf->yTransform(_y[i]));
  }
  data->setPoint(i,transf->xTransform(_x[i-1]),transf->yTransform(0));
  data->setPoint(i+1,transf->xTransform(_x[0]),transf->yTransform(0));    
  painter->drawPolygon(*data, false,0,npt+2);  
  delete data;
}
