#ifndef QPTRANSFO_H
#define QPTRANSFO_H

#include "Qplotter/qp/qpqwt_dimap.h"
#include "Qplotter/qp/qpmap.h"

/** %%%

    
  Usage:

    QpTransformation is the virtual class from which every
  transformation strategy inherits.  It defines the common interface
  of these strategy classes.  It still uses Qp and/or Qwt maps but
  this point could be changed later.

  TODO:

*/


class QpTransformation
{
public:
  /** The enum which will be used to determine the strategy to use for each object.
      The QpObject will have a value from this enum. 
      Zone -> The object will be placed in zone's coordinates (according to its zone)
      Page -> The object will be placed in "absolute" page's coordinates
      Zr = ZoneRelative 
      ZrBottomLeft -> The object will be placed according to the bottom-left corner 
      of its zone. The (x,y) coordinates of this object are the offset in page's 
      coordinates between its position and the corner's position.
      ...
  */
  enum tStrategy { Zone, Page, ZrBottomLeft, ZrBottomRight, ZrTopLeft, ZrTopRight };

    /// Constructor
  QpTransformation(QpQwtDiMap *x, QpQwtDiMap *y) {xpmap=x; ypmap=y;}
  
  /// Destructor (maps are destroyed by the page)
  virtual ~QpTransformation() { }

  QpQwtDiMap *getXpMap() const {return xpmap;}
  QpQwtDiMap *getYpMap() const {return ypmap;}

  inline virtual int xTransform(double x) const = 0;
  inline virtual int yTransform(double y) const = 0;
  
  inline virtual double invXTransform(int x) const = 0;
  inline virtual double invYTransform(int x) const = 0;
  /* virtual inline transform(double min1, double max1, double min2, double max2, double x)
  { return (min1 + (x-min2) * ((max1-min1)/(max2-min2))); } */

  virtual void selfDescribe(std::ostream output)
    {
      /* Send to output a XML definition of this class */
      output << "QpTransformation" << std::endl ;
    }
  
 protected:
  QpQwtDiMap *xpmap, *ypmap;
};



/** The "absolute page coordinates" transformation strategy.
 */
class PageStrategy : public QpTransformation
{
public:
  PageStrategy(QpQwtDiMap *x, QpQwtDiMap *y):QpTransformation(x,y) {}
  inline virtual int xTransform(double x) const { return xpmap->transform(x); }
  inline virtual int yTransform(double y) const { 
    return ypmap->i1() + ypmap->i2() - ypmap->transform(y); 
  }
  /// Transform pixels -> page coordinates
  inline virtual double invXTransform(int x) const { return xpmap->invTransform(x); }
  inline virtual double invYTransform(int x) const { 
    return ypmap->d2() - ypmap->invTransform(x); 
  }
};

/** ZoneStrategy is the "absolute zone coordinates" transformation strategy
   This class can be inherited to create some zone relative strategies.
*/
class ZoneStrategy : public QpTransformation
{
 public:
  ZoneStrategy(QpQwtDiMap *x, QpQwtDiMap *y, QpMap *xz, QpMap *yz):QpTransformation(x,y) {
    xzmap=xz; 
    yzmap=yz;
  }
  /// Destructor (maps are destroyed by the page)
  virtual ~ZoneStrategy()  { }
  inline virtual int xTransform(double x) const { 
    return xpmap->transform( xzmap->transform(x) ); 
  }
  inline virtual int yTransform(double y) const { 
    return ypmap->i1() + ypmap->i2() - ypmap->transform( yzmap->transform(y) ); 
  }
  /// Transform pixels -> zone coordinates
  inline virtual double invXTransform(int x) const { 
    return xzmap->invTransform(xpmap->invTransform(x)); 
  }
  inline virtual double invYTransform(int x) const { 
    return yzmap->invTransform(ypmap->d2() - ypmap->invTransform(x)); 
  }
  inline virtual QpMap *getXzMap() const {return xzmap;}
  inline virtual QpMap *getYzMap() const {return yzmap;}

  inline double getXFactor() const {return xzmap->getFactor();}
  inline double getYFactor() const {return xzmap->getFactor();}

  bool isInvalid()
    {
      if (xzmap == 0) return true;
      if (yzmap == 0) return true;
      return ((xzmap->isInvalid()) || (yzmap->isInvalid()));
    }

  ZoneStrategy *copy()
    {
      return new ZoneStrategy(xpmap->copy(),ypmap->copy(),
			      xzmap->copy(),yzmap->copy());      
    }

  
 protected:
  QpMap *xzmap, *yzmap;
};

/* According to ZrBottomLeft, ZrBottomRight, ZrTopLeft, ZrTopRight */
class ZrStrategy : public ZoneStrategy
{
 public:
  ZrStrategy(QpQwtDiMap *x,QpQwtDiMap *y,QpMap *xz, QpMap *yz, tStrategy st):ZoneStrategy(x,y,xz,yz) {
    strategy = st;
  }

  ZrStrategy *copy()
    {
      return new ZrStrategy(xpmap->copy(),ypmap->copy(),
			    xzmap->copy(),yzmap->copy(),strategy);      
    }

  

  inline virtual int xTransform(double x) const {
    switch (strategy) {
    case ZrBottomLeft:
      return xpmap->transform( x+xzmap->getContainer1() ); 
      break ;
    case ZrBottomRight:
      return xpmap->transform( x+xzmap->getContainer2() ); 
      break ;
    case ZrTopLeft:
      return xpmap->transform( x+xzmap->getContainer1() ); 
      break ;
    case ZrTopRight:
      return xpmap->transform( x+xzmap->getContainer2() ); 
      break ;
    default :
      return 0;
    }
  }
  
  inline virtual int yTransform(double y) const { 
    switch (strategy) {
    case ZrBottomLeft:
      return ypmap->i1() + ypmap->i2() - ypmap->transform( y+yzmap->getContainer1() ); 
      break ;
    case ZrBottomRight:
      return ypmap->i1() + ypmap->i2() - ypmap->transform( y+yzmap->getContainer1() ); 
      break ;
    case ZrTopLeft:
      return ypmap->i1() + ypmap->i2() - ypmap->transform( y+yzmap->getContainer2() ); 
      break ;
    case ZrTopRight:
      return ypmap->i1() + ypmap->i2() - ypmap->transform( y+yzmap->getContainer2() ); 
      break ;
    default :
      return 0;
    }
  }
  
  void setStrategy(tStrategy st) 
    {
      strategy = st ;
    }
  
 private:
  tStrategy strategy ;
};

#endif
