#ifndef QPZONE_H
#define QPZONE_H

#include <qrect.h>
#include <qstack.h>
#include <qdict.h>

#include "Qplotter/qp/qpnode.cpp"
#include "Qplotter/qp/qppolyline.h"

#include "Qplotter/gblib/Projection3D.hh"
#include "Qplotter/gblib/ScreenProjection3D.hh"
#include "Qplotter/gblib/Point3D.hh"

typedef gblib::Point3D<float> Point;
typedef gblib::ScreenProjection3D Projection;
typedef gblib::Projection3D Transformation;

/** %%%

    The zone class.
    
  Usage:

    QpZone is a list of QpObject. It has got its own system of
  coordinates placed according to the page's one.  A zone can have a
  title.

  TODO:

  (--) better title zone in one corner (with a box or a shade)
  
*/


#include <iostream>

class QpAxisNode;
class QpAxis;
class QpScale;
class QpText;
class QpSummaryInfo;

class QpZone : public QpNode<QpObject>
{
 public:
  enum AxisPosition {Top,Bottom,Left,Right};
 public:
  /// Constructors
  QpZone(QpObject *p, QpMap *_xMap, QpMap *_yMap);  // No deepcopy
  /// Copy constructor : create a new zone placed like z 
  QpZone(const QpZone &z);
  /// Destructor
  virtual ~QpZone();
  /// Modify the zone to allow the painting even if it is null
  /// The params will be used as axis limits
  /// Warning : the fixedCoords will be activated
  void null(double xmin=0.0, double xmax=1.0, double ymin=0.0, double ymax=1.0);
  /// Zone selector
  QpZone   *getZone()       const {return (QpZone * const)this;}
  ///  Retrieve whether the coordinate system is fixed or not
  bool     getFixedCoords() const {return fixedCoords;}
  /// Retrieve the transformation strategy
  virtual QpTransformation *getTransformation(QpTransformation::tStrategy s) const;
  /// Returns a Qt rectangle defining the zone limits
  QRect getLimits() const;

  bool isInvalid() ;

  
  
  /** To add some text anywhere on the zone. You can use the different
      Zone Transformations.  */
  QpText *addText (double xPosition,double yPosition,QString text,QpTransformation::tStrategy st = QpTransformation::Zone);
  /// Set the title of the zone
  QpText *setTitle(QString _title, int _align=Qt::AlignCenter);
  /// Add Legend to a zone
  QpText *addLegend(QString text);
  
  /// Get the zone title
  QString getTitle() const;
  /** Determine if the zone system of coordinates is locked or not 
      (i.e. if fixedCoords is false you can change the system)
  */
  void setFixedCoords(bool f)     {fixedCoords=f;}
  /// Draw the limits of the zone
  void enableBoundingBox(bool b=true) {limits=b;}
  /// Get the range of the X map
  void getXRange(double& min, double& max) const;
  /// Get the range of the Y map
  void getYRange(double& min, double& max) const;
  /// Change the range of the X map
  void setXRange(double min, double max, bool log=false);
  /// Change the range of the Y map
  void setYRange(double min, double max, bool log=false);
  /// Update zone coordinates and axis on X
  void setThickness(int t) { thickness = t; }
  int getThickness() { return thickness ; }
  
  /// Update zone coordinates and axis on X
  QpAxisNode *updateX(double xmin, double xmax, bool force=false);
  /// Update zone coordinates and axis on Y
  QpAxisNode *updateY(double ymin, double ymax, bool force=false);
  /// Set lin/log on X
  void setXLog(bool log);
  /// Set lin/log on Y
  void setYLog(bool log);
  /// Create an axis on one of the edges
  QpAxis* makeAxis (AxisPosition edge);

  /// Activate or deactivate the grid on X
  void enableGridX(bool maj=true, bool min=false);
  /// Activate or deactivate the grid on Y
  void enableGridY(bool maj=true, bool min=false);
  /// Return a pointer on the SummaryInfo [maybe null]
  QpSummaryInfo *getSummaryInfo() const {return sinfo;}

  inline void drawLine(float x1, float y1, float x2, float y2, QColor color = Qt::black) 
    {
      double LineX[2];
      double LineY[2];
      LineX[0] = x1;
      LineX[1] = x2;
      LineY[0] = y1;
      LineY[1] = y2;
      QpNode<QpPolyShape> *pl=new QpNode<QpPolyShape>(new QString("QpMultiLine"),this,true);      
      pl->setQpPen(QPen(color));
      new QpPolyLine(2, LineX, LineY, pl, true);
    }
  
  inline void draw3dLine(float x1, float y1, float z1,
			 float x2, float y2, float z2,
			 Transformation m,
			 QColor color = Qt::black) {

    Point p1(x1, y1, z1);
    Point p2(x2, y2, z2);
    double LineX[2];
    double LineY[2];

    p1 = m*p1;
    p2 = m*p2;

    LineX[0] = p1.x();
    LineX[1] = p2.x();
    LineY[0] = p1.y();
    LineY[1] = p2.y();

    QpNode<QpPolyShape> *pl=new QpNode<QpPolyShape>(new QString("QpMultiLine"),this,true);
    pl->setQpPen(color);
    new QpPolyLine(2, LineX, LineY, pl, true);
  }


  inline void draw3dPolygon(double x[], double y[], double z[],
			    Transformation m,
			    QColor color = Qt::white) {
    Point p1(x[0], y[0], z[0]);
    Point p2(x[1], y[1], z[1]);
    Point p3(x[2], y[2], z[2]);
    Point p4(x[3], y[3], z[3]);

    double LineX[5];
    double LineY[5];

    p1 = m*p1;
    p2 = m*p2;
    p3 = m*p3;
    p4 = m*p4;

    LineX[0] = p1.x();
    LineX[1] = p2.x();
    LineX[2] = p3.x();
    LineX[3] = p4.x();
    LineX[4] = p1.x();
    LineY[0] = p1.y();
    LineY[1] = p2.y();
    LineY[2] = p3.y();
    LineY[3] = p4.y();
    LineY[4] = p1.y();

    QpNode<QpPolyShape> *pg=new QpNode<QpPolyShape>(new QString("QpMultiLine"),this,true);
    /*
      pg->setQpBrush(QBrush(Qt::white));
      pg->setQpPen(QPen(Qt::black));
    */
    pg->setQpBrush(QBrush(color));
    pg->setQpPen(QPen(Qt::black));
    new QpPolygon (4, LineX, LineY, pg, true);
    QpNode<QpPolyShape> *pl=new QpNode<QpPolyShape>(new QString("QpMultiLine"),this,true);
    pg->setQpBrush(QBrush(color));
    pl->setQpPen(QPen(color));
    new QpPolyLine(5, LineX, LineY, pl, true);
  }

  
  
  /// Zoom on a zone region between 2 points (in zone coordinates)
  void zoom(double x1, double y1, double x2, double y2);
  /** Unzoom... i.e. pop the old maps from the stack. If original is set to 
      true, the original maps will be reused (and stack cleaned).
  */
  void unZoom(bool original=false);
  /** Zoom "factor times" on the zone from one point.
      i.e. Got a (100.0,100.0) zone and zoom(20.0,20.0,4.0) 
      [newWidth = oldWidth / 4.0] -> will zoom from (20.0,20.0) -> (45.0,45.0)
  */
  void zoom(double x, double y, double factor);
  /// Get the zoom level
  unsigned int getZoomLevel () {return mapStack.count();}   
  /// Draw the zone
  void draw();
  /// Print the zone
  void print();
  /// Access the axis system defined on the zone
  QpAxisNode *getAxisNode();
  /// Do {_x,_y} page coordinates touch the zone?
  bool isTouched(double _x, double _y);
  /// Is X axis log or linear?
  bool getXLog() const;
  /// Is Y axis log or linear?
  bool getYLog() const;
  // Zone height (page coordinates)
  double getHeight() const ;
  // Zone width (page coordinates)
  double getWidth() const ;
  // Zone Screen height 
  double getScreenHeight() const ;
  // Zone Screen width 
  double getScreenWidth() const ;
  /// Draw zone bounding box
  void boundingBox(QPainter *painter, int frame = 0);

  bool is2D ;

 private:
  /// Variables initialisation
  void init();
  /// Update the grid according to axis
  void setGridX();
  void setGridY();
  /// Update the list of summary tables by scanning all children objects
  void updateSt(QPainter *);

  /// Reconstruct transformation objects
  void updateTransf(QpMap *xz, QpMap *yz);
  /// Check that limits are consistent, i.e. min < max
  bool checkLimits();
  /// Switch an axis to/from lin/log
  void axisLinLog (QpAxis *axis, QpMap *map, bool logV);  

  
 private:
  // Stack of maps used for zooms
  QStack<ZoneStrategy> mapStack;
  // Pointer to the SummaryInfo of the zone 
  QpSummaryInfo *sinfo;

  // Properties
  bool limits;
  bool fixedCoords;
  bool gridX;
  bool gridY;
  bool logX;
  bool logY;
  int thickness;
  
  // Pointer to the zone title 
  QpText *title;
  // Pointer to the list of axis kept in the zone
  QpAxisNode *qpaxis;
  // Transformation objects
  ZoneStrategy *transformation;
  ZrStrategy *zrtransformation[4];
};


#endif
