#include <signal.h>
#include "Qplotter/viewer/qpViewer.h"
#include "Qplotter/qp/qpaxis.h"
#include "Qplotter/viewer/applInstance.h"
#include "Qplotter/qp/qpscatterplot2d.h"
#include <algorithm>
#include <iterator>

#include "Interfaces/IAnnotation.h" 
#include "Interfaces/IVector.h"
#include "Interfaces/IScatter2D.h"

#include <qvector.h> 

#if (defined __sun && __SUNPRO_CC > 0x500)
# include <strstream>
#else
# include <sstream>
#endif

QpViewer::Viewer::Viewer () : window(0) , page(0) , pageManager(0)
{
  // Get a QApplication instance from the singleton
  window = QApplicationInstance::Instance();
#ifdef NO_THREADED_PLOTTER
  pageManager = new ViewFrame(*window);
  // Define properties
  viewerProperties.addProperty("timer","");
#else
  // Ignore signals
  sigset_t ens ;
  struct sigaction action ;

  sigemptyset (&ens);
  sigaddset (&ens, SIGTERM) ;
  sigaddset (&ens, SIGTSTP) ;
  sigaddset (&ens, SIGINT) ;
  action.sa_handler = SIG_IGN ;
  action.sa_mask = ens ;
  action.sa_flags = SA_RESTART ;
  sigaction (SIGTERM, &action, NULL) ;
  sigaction (SIGTSTP, &action, NULL) ;
  sigaction (SIGINT,  &action, NULL) ;
  
  
  window->lock();
  pageManager = new ViewFrame(*window);
  window->unlock();
#endif
}

QpViewer::Viewer::~Viewer () {
  window->deref();
  delete page;
  delete pageManager;
#ifdef NO_THREADED_PLOTTER
  //delete window;
#endif
}

void QpViewer::Viewer::clear () {
  window->lock();
  pageManager->getPage()->clear();
  window->unlock();
} 

void QpViewer::Viewer::show () {
#ifdef NO_THREADED_PLOTTER
  // Draw the empty page
  pageManager->show();
  pageManager->raise();
  pageManager->refresh();
  window->exec();
  // Now go full speed !
  pageManager->setTimerInterval(1);
#else
  // Draw the empty page
  window->lock();
  pageManager->show();
  window->wakeUpGuiThread();
  window->unlock();
#endif
} 

void QpViewer::Viewer::refresh () {
  // To refresh the window when it has been covered by others.
#ifdef NO_THREADED_PLOTTER
  pageManager->refresh();
  window->exec(); 
#endif
} 


IPage *QpViewer::Viewer::createPage (int nx, int ny) {
#ifndef NO_THREADED_PLOTTER
  window->lock();
#endif
  pageManager->getPage()->createZones(nx,ny);
#ifndef NO_THREADED_PLOTTER
  window->unlock();
#endif
  delete page;
  page = new QpViewer::Page(nx,ny,pageManager->getPage());
  return page;
}

void QpViewer::Viewer::makePS (const char *filename) {
#ifndef NO_THREADED_PLOTTER
  window->lock();
#endif
  page->draw();
  pageManager->getPage()->print(filename,false);   
#ifndef NO_THREADED_PLOTTER
  window->unlock();
#endif
}

void QpViewer::Viewer::makePicture (const char *filename) {
  window->lock();
  page->draw();
  pageManager->getPage()->save(filename);   
  window->unlock();
}
void QpViewer::Viewer::setProperty (QPSTD::string name, QPSTD::string value) {
  QPSTD::transform(name.begin(),name.end(),name.begin(),tolower);
  if (viewerProperties.find(name,value) == StringPropertySet::FOUND) {
    if (name == "timer" ) {
      int interval;
      QPSTD::istringstream is(value.c_str());
      is >> interval;
      pageManager->setTimerInterval(interval);
    }
  } else
    QPSTD::cerr << "Unknown property: " << name <<QPSTD::endl;
}

void QpViewer::Viewer::listOptions(QPSTD::ostream &os) {
  QpViewer::listOptions("Viewer",viewerProperties, os);
} 

void QpViewer::Viewer::draw () {
#ifdef NO_THREADED_PLOTTER
  page->draw();
  pageManager->show();
  pageManager->raise();
  pageManager->refresh();
  window->exec();
#else
  window->lock();
  page->draw();
  pageManager->show();
  pageManager->raise();
  pageManager->refresh();
  window->wakeUpGuiThread();
  window->unlock();
#endif
}

QpViewer::Page::Page(int nx,int ny,QpPage *_page) : 
  currentZone(0),qppage(_page) {
    int i ;
    zones.clear();
    for (i=0;i< nx*ny;i++)
      zones.push_back(new QpViewer::Zone(_page,i));
    theDrawStyle = new VDrawStyle(this);
    theTextStyle = new VTextStyle(this);
}

QpViewer::Page::~Page() {
  size_t i ;
  for (i=0;i<zones.size() ;i++)
    delete zones[i]; 
  for (i=0;i<allText.size() ;i++)
    delete allText[i]; 
  delete theDrawStyle;
  delete theTextStyle;
}

IText *QpViewer::Page::setTitle(QPSTD::string newTitle) {
  // Changing the title may trigger the creation of a QpMath object.
  // QpMath::QpMath calls QFontMetrics which queries the font manager... 
  // On Solaris we have to lock/unlock the resource
  (QApplicationInstance::retrieveInstance())->lock();
  QpText *qpto = qppage->setTitle(newTitle.c_str());
  (QApplicationInstance::retrieveInstance())->unlock();

  VText *title = new VText(qpto);
  allText.push_back(title);
  return title;
}

IText *QpViewer::Page::addText(double x, double y,QPSTD::string text) {
  // Changing the title may trigger the creation of a QpMath object.
  // QpMath::QpMath calls QFontMetrics which queries the font manager... 
  // On Solaris we have to lock/unlock the resource
  (QApplicationInstance::retrieveInstance())->lock();
  QpText *qpto = qppage->addText(x,y,text.c_str());
  (QApplicationInstance::retrieveInstance())->unlock();

  VText *newTxt = new VText(qpto);
  allText.push_back(newTxt);
  return newTxt;
}


void QpViewer::Page::draw(){
  size_t i ;
  for (i=0;i< zones.size();i++) 
    zones[i]->draw();
}


// New AIDA 2.2 methods

void QpViewer::Zone::clear()
{
  hiZone->clearList() ;
}

void QpViewer::Zone::createHistogram (Anaphe::IVector *data)
{
  if (data  != 0) {
    switch (data->dimension())
      {
      case 1:
	new QpCurve1D(hiZone, *data);
	break;
      case 2:
	new QpCurve2D(hiZone, *data);
	break;
      case 3:
	new QpCurve3D(hiZone, *data);
	break;
      default:
	std::cerr << "QpViewer::Zone::createHistogram> don't know how to deal with data of dim " 
		  << data->dimension() << std::endl;
	break;
      }
  } else {
	std::cerr << "QpViewer::Zone::createHistogram> Ignoring illegal data: NULL" << std::endl;

  }
}



void QpViewer::Zone::update(IStyleOwner::styleId /* sti */) {
  hiZone->setQpFont(theTextStyle->getFont());
  // Draw style color overrides text color!
  hiZone->setQpPen(theDrawStyle->getPen());
  hiZone->setQpBrush(theDrawStyle->getBrush());
}

QpViewer::Dataset::Dataset( Anaphe::IVector& VoP) : lVoP(VoP) ,summaryT(0) ,  
  qpc1(0) , qpc2(0) , curveCreated (false) {
  dataSetProperties.addProperty
    ("representation","error line histo marker errormark smooth hfilled box color");
  dataSetProperties.addProperty("legend","");
  dataSetProperties.addProperty("mtype","none rect diamond triangle dtriangle utriangle ltriangle rtriangle xcross ellipse");
  dataSetProperties.addProperty("msize","");
  dataSetProperties.addProperty("colorlevels","");
  if (lVoP.annotation() != 0)
    addAnnotation(lVoP.annotation());
  theCurveStyle = new VDrawStyle (this) ;
} 

QpViewer::Dataset::~Dataset() {
  delete summaryT;
  delete theCurveStyle;
}

QpViewer::Dataset & QpViewer::Dataset::operator= (const class QpViewer::Dataset &other) {
  if (this == &other)
    return *this;
  lVoP=other.lVoP;
  qpc1=0;
  qpc2=0;
  summaryT=0;
  return *this;
}

void QpViewer::Dataset::createCurve(QpZone *z1) {
  if (lVoP.dimension() == 1) 
    qpc1 = new QpCurve1D(z1);
  else 
    qpc2 = new QpCurve2D(z1);
}

void QpViewer::Dataset::addAnnotation(Anaphe::IAnnotation *anno)
{
  if (anno == 0) return;	// ... just in case ... 
  if (anno->size() > 0) {
    int cnt=0;
    summaryT = new SummaryTable;
    for (int index=0; index<anno->size(); index++) {
      cnt++;
      if ( anno->visibilityFlag(index) ) {
	summaryT->append( 
	   new StPair(anno->key(index).c_str(), anno->value(index).c_str()));
      }
    }
  }
}

// This method does not actually draw, it merely creates the "low-level"
// Qplotter graphics objects.
void QpViewer::Dataset::draw(QpZone *z1, bool stat, bool reverse) {
  if (!curveCreated) {
    if (qpc1 != 0) {
      qpc1->newCurve(lVoP,reverse);
      if ((summaryT != 0) && stat )
	qpc1->setSummaryTable(summaryT); 
    }
    if (qpc2 != 0) {
      qpc2->newCurve(lVoP,reverse);
      if ((summaryT != 0) && stat )
	qpc2->setSummaryTable(summaryT); 
    }
    curveCreated = true;
  }
}

void QpViewer::Dataset::setProperty(QPSTD::string name, QPSTD::string newValue)
{
  QPSTD::string value(newValue);
  QPSTD::transform(name.begin(),name.end(),name.begin(),tolower);
  QPSTD::transform(value.begin(),value.end(),value.begin(),tolower);
  StringPropertySet::Existance result = dataSetProperties.find(name,value);
  switch (result) {
  case(StringPropertySet::FOUND) :
    switch (lVoP.dimension()) {
    case (1) :
      if (name == "representation") {
	if(value == "error") qpc1->setType(QpCurve1D::ERROR);
	if(value == "line")  qpc1->setType(QpCurve1D::LINE);
	if(value == "histo") qpc1->setType(QpCurve1D::HISTO);
	if(value == "errormark") qpc1->setType(QpCurve1D::MARKER_ERROR);
	if(value == "marker") qpc1->setType(QpCurve1D::MARKER);
	if(value == "smooth") qpc1->setType(QpCurve1D::SMOOTH);
	if(value == "hfilled") qpc1->setType(QpCurve1D::FILLED_HISTO);
      }
      if (name == "mtype") {
	qpc1->setSymStyle(MapMarker(value));
      }
      if (name == "msize" ) {
	int markerSize;
	QPSTD::istringstream is(value.c_str());
	is >> markerSize;
	qpc1->setSymSize(markerSize);
      }
      break;
    case (2) :
      if (name == "representation") {
	if(value ==  "box") qpc2->setType(QpCurve2D::BOX);
	if(value ==  "color") qpc2->setType(QpCurve2D::COLOR);
      }
      if (name == "colorlevels" ) {
	int colorlevels;
	QPSTD::istringstream is(value.c_str());
	is >> colorlevels;
	qpc2->changePalette(QpPalette::GEOGRAPHICAL,colorlevels);
      }
      break;
    default:
      break;
    }
    if (name == "legend") {
      if (summaryT == 0) summaryT = new SummaryTable;
      // Remove if already there
      summaryT->removeByKey("legend");
      summaryT->append(new StPair("legend", newValue.c_str()));
    }
    break;
  case(StringPropertySet::KEYNOTFOUND):
    QPSTD::cerr << "Unknown property: " << name << QPSTD::endl;
    break;
  case(StringPropertySet::VALUENOTFOUND):
    QPSTD::cerr << "Unknown value: " << value << " for property" ;
    QPSTD::cerr<< name <<QPSTD::endl; 
  }
}

void QpViewer::Dataset::setProperty (QPSTD::string name, int value) { 
}

void QpViewer::Dataset::listOptions(QPSTD::ostream &os){
  QpViewer::listOptions("DataSet",dataSetProperties, os);
}

void QpViewer::Dataset::update(IStyleOwner::styleId sti) {
  if (sti == QpViewer::IStyleOwner::DrawStyle) {
    if (qpc1 != 0) {
      qpc1->setQpPen(theCurveStyle->getPen());
      qpc1->setQpBrush(theCurveStyle->getBrush());
    } else {
      qpc2->setQpPen(theCurveStyle->getPen());
      qpc2->setQpBrush(theCurveStyle->getBrush());
    }
  }
}

IDrawStyle* QpViewer::Dataset::getStyle () {
  return theCurveStyle;
}

/** Global functions */

void QpViewer::listOptions( QPSTD::string className, StringPropertySet &pSet, QPSTD::ostream &os) {
  os << QPSTD::endl;
  os << "--------------------- " <<  className;
  os << " Option List -------------------"<< QPSTD::endl;
  os << "The following options for ";
  os <<  className;
  os << " objects are available";
  os << QPSTD::endl;
  os << "via the setProperty method taking two strings as parameter:";
  os << QPSTD::endl;
  os << "setProperty (string name, string value)."<<QPSTD::endl;
  pSet.print(os);
  os << "------------------------------------------------------------" ;
  os << QPSTD::endl << QPSTD::endl;
  os << QPSTD::endl;
} 

const QColor &QpViewer::MapColor(QPSTD::string value) {
   if (value == "blue") return Qt::blue;
   if (value == "white") return Qt::white;
   if (value == "black") return Qt::black;
   if (value == "red") return Qt::red;
   if (value == "green") return Qt::green;
   if (value == "yellow") return Qt::yellow;
   if (value == "magenta") return Qt::magenta;
   if (value == "cyan") return Qt::cyan;   
   if (value == "darkgray") return Qt::darkGray;   
   if (value == "lightgray") return Qt::lightGray;   
   if (value == "gray") return Qt::gray;   
   if (value == "darkred") return Qt::darkRed;   
   if (value == "darkgreen") return Qt::darkGreen;   
   if (value == "darkblue") return Qt::darkBlue;   
   if (value == "darkcyan") return Qt::darkCyan;   
   if (value == "darkmagenta") return Qt::darkMagenta;   
   if (value == "darkyellow") return Qt::darkYellow;
   QPSTD::cerr << "Unknown color: " << value << QPSTD::endl;
   return Qt::black;
}

const QpMarker::SymStyle QpViewer::MapMarker(QPSTD::string value) {
   if (value == "none") return QpMarker::None;
   if (value == "rect") return QpMarker::Rect;
   if (value == "diamond") return QpMarker::Diamond;
   if (value == "triangle") return QpMarker::Triangle;
   if (value == "dtriangle") return QpMarker::DTriangle;
   if (value == "utriangle") return QpMarker::UTriangle;
   if (value == "ltriangle") return QpMarker::LTriangle;
   if (value == "rtriangle") return QpMarker::RTriangle;   
   if (value == "cross") return QpMarker::Cross;   
   if (value == "xcross") return QpMarker::XCross;   
   if (value == "ellipse") return QpMarker::Ellipse;   
   QPSTD::cerr << "Unknown marker style: " << value << QPSTD::endl;
   return QpMarker::Ellipse;
}

void QpViewer::Viewer::createCoordWindow()
{
  window->lock();
  pageManager->createCoordWindow();
  window->unlock();
}



// New AIDA 2 methods Version 2

void QpViewer::Viewer::createRegion  (double x, double y, double w, double h)
{
  if (page)
    page->createIrregularZone(x,y,w,h);
  else
    QPSTD::cerr << "QpViewer::Viewer::createRegion> Page not yet created" << QPSTD::endl;
}

void QpViewer::Viewer::createRegions (int columns, int rows)
{
  if (page)
    page->createZones(columns, rows);
  else
    QPSTD::cerr << "QpViewer::Viewer::createRegions> Page not yet created" << QPSTD::endl;
}

void QpViewer::Viewer::plot (Anaphe::IVector *data, int index)
{
  // hiZone  
  if (data) {
    switch (data->dimension())
      {
      case 1:
	new QpCurve1D(page->getZone(index), *data);
	break;
      case 2:
	new QpCurve2D(page->getZone(index), *data);
	break;
      case 3:
	new QpCurve3D(page->getZone(index), *data);
	break;
      default:
	std::cerr << "QpViewer::Viewer::plot> don't know how to deal with data of dim " 
		  << data->dimension() << std::endl;
	break;
      }
  } else
    {
	std::cerr << "QpViewer::Viewer::plot> Ignoring illegal data: NULL" << std::endl;
    }
}

void QpViewer::Viewer::clearRegion (int index)
{
  page->getZone(index)->clearList();
}

void QpViewer::Viewer::activateOutline(bool On)
{
  //  page->crossOutline(On);
}

IZone *QpViewer::Page::selectZone(unsigned int index)
{
  if ((index >= 0) && ( index < zones.size()))
    currentZone = index;
  return zones[currentZone];
}
  
ITextStyle* QpViewer::Page::getTextStyle()
{
  // Needed in AIDA 2 ?
  return theTextStyle;
}

IDrawStyle* QpViewer::Page::getDrawStyle()
{
  // Needed in AIDA 2 ?
  return theDrawStyle;
}


// -------------------------------------- //

void QpViewer::Page::update(IStyleOwner::styleId /* sti */) {
  qppage->setQpFont(theTextStyle->getFont());
  // Draw style color overrides text color!
  qppage->setQpPen(theDrawStyle->getPen());
  qppage->setQpBrush(theDrawStyle->getBrush());
}

QpViewer::Zone::Zone(QpPage *pm , int hiZ) : stat(true) , mirrorAxis(false) ,
reverseDraw(false) {
  zoneProperties.addProperty("option","xlinear xlog stats nostats ylinear ylog xaxisgrid yaxisgrid");
  zoneProperties.addProperty("coordinates","locked free");
  zoneProperties.addProperty("mirroraxis","yes no");
  zoneProperties.addProperty("draworder","normal reverse");
  hiZone = pm->getZone(hiZ);
  hiZone->setFixedCoords(true);
  xAxis = new VAxis();
  yAxis = new VAxis();
  theDrawStyle = new VDrawStyle(this);
  theTextStyle = new VTextStyle(this);
  datasets.clear();
}

QpViewer::Zone::~Zone() {
  size_t i ;
  for (i=0;i<datasets.size() ;i++)
    delete datasets[i]; 
  for (i=0;i<allText.size() ;i++)
    delete allText[i]; 
  delete xAxis;
  delete yAxis;
  delete theDrawStyle;
  delete theTextStyle;
}

IText *QpViewer::Zone::setTitle(QPSTD::string newTitle) {
  // Changing the title may trigger the creation of a QpMath object.
  // QpMath::QpMath calls QFontMetrics which queries the font manager... 
  // On Solaris we have to lock/unlock the resource
  (QApplicationInstance::retrieveInstance())->lock();
  QpText *qpto = hiZone->setTitle(newTitle.c_str());
  (QApplicationInstance::retrieveInstance())->unlock();

  VText *title = new VText(qpto);
  allText.push_back(title);
  return title;
}

IText *QpViewer::Zone::addText(double x, double y,QPSTD::string text) {
  // Changing the title may trigger the creation of a QpMath object.
  // QpMath::QpMath calls QFontMetrics which queries the font manager... 
  // On Solaris we have to lock/unlock the resource
  (QApplicationInstance::retrieveInstance())->lock();
  (QApplicationInstance::retrieveInstance())->unlock();

  QpText *qpto = hiZone->addText(x,y,text.c_str());
  VText *newTxt = new VText(qpto);
  allText.push_back(newTxt);
  return newTxt;
}

void  QpViewer::Zone::draw() {
  size_t i ;
  if (datasets.size() > 0) {
    for (i=0;i<datasets.size() ;i++) 
      datasets[i]->draw(hiZone,stat,reverseDraw);
  }
  // Titles/labels
  if (xAxis->getTitle() != "")
    hiZone->getAxisNode()->getAxis("x1")->setTitle
      (xAxis->getTitle().c_str());
  if (yAxis->getTitle() != "")
    hiZone->getAxisNode()->getAxis("y1")->setTitle
      (yAxis->getTitle().c_str());
  if (xAxis->getLabel() != "")
    hiZone->getAxisNode()->getAxis("x1")->setLegend
      (xAxis->getLabel().c_str());
  if (yAxis->getLabel() != "")
    hiZone->getAxisNode()->getAxis("y1")->setLegend
      (yAxis->getLabel().c_str());
  if (mirrorAxis && (hiZone->getAxisNode()->getAxis("x2") == 0)) {
    hiZone->makeAxis(QpZone::Top);
    hiZone->makeAxis(QpZone::Right);
  }
}

/// Retrieve axis
ViewerAxis *QpViewer::Zone::getAxis (QPSTD::string name) {
  if ((name == "X")||(name == "x"))
    return xAxis;
  else
    return yAxis;
}

void  QpViewer::Zone::xMinMax (double min, double max) {
  if (max > min)
     hiZone->updateX(min,max,true);
}

void  QpViewer::Zone::yMinMax (double min, double max) {
  if (max > min) 
    hiZone->updateY(min,max,true);
}

bool QpViewer::Zone::getStat() const {
  return stat;
}

void  QpViewer::Zone::setProperty (QPSTD::string name, QPSTD::string newValue) {
  QPSTD::string value(newValue);
  QPSTD::transform(name.begin(),name.end(),name.begin(),tolower);
  QPSTD::transform(value.begin(),value.end(),value.begin(),tolower);
  StringPropertySet::Existance result = zoneProperties.find(name,value);
  switch (result) {
  case(StringPropertySet::VALUENOTFOUND):
      QPSTD::cerr << "Unknown value: " << newValue;
      QPSTD::cerr << " for property" << name <<QPSTD::endl; 
      break;
  case(StringPropertySet::FOUND) :
      if (name == "option") {
	if (value == "xlinear") hiZone->setXLog(false);
	if (value == "xlog")   hiZone->setXLog(true);
	if (value == "stats") stat = true;
	if (value == "nostats") stat = false;
	if (value == "ylinear")  hiZone->setYLog(false) ;
	if (value == "ylog") hiZone->setYLog(true) ;
	if (value == "xaxisgrid") hiZone->enableGridX(true);
	if (value == "yaxisgrid") hiZone->enableGridY(true);
      }
      if (name == "coordinates") {
	if (value == "locked") hiZone->setFixedCoords(true);
	if (value == "free")   hiZone->setFixedCoords(false);
      }
      if (name == "mirroraxis") {
	if (value == "yes") mirrorAxis = true;
	if (value == "no")  mirrorAxis = false;
      }
      if (name == "draworder") {
	if (value == "normal")   reverseDraw = false;
	if (value == "reverse")  reverseDraw = true;
      }
      break;
  case(StringPropertySet::KEYNOTFOUND):
      QPSTD::cerr << "Unknown property: " << name << QPSTD::endl;
      break;
  }
}

void  QpViewer::Zone::listOptions(QPSTD::ostream &os){
  QpViewer::listOptions("Zone",zoneProperties, os);
}

IDataset*  QpViewer::Zone::addDataset (Anaphe::IVector& VoP) {
  Dataset *dS = new QpViewer::Dataset(VoP);
  dS->createCurve(hiZone);
  datasets.push_back(dS);
  return dS;
}

void  QpViewer::Zone::addScatter (Anaphe::IScatter2D *scat) {
  new QpScatterPlot2D(hiZone, scat );
}

IDataset*  QpViewer::Zone::getDataset (unsigned int ds) {
  if ((ds >= 0) && ( ds < datasets.size()))
    return datasets[ds];
  else 
    return 0;    
}

ITextStyle* QpViewer::Zone::getTextStyle() {
  return theTextStyle;
}

IDrawStyle* QpViewer::Zone::getDrawStyle(){
  return theDrawStyle;
}













