#ifndef _QPVIEWERH_
#define _QPVIEWERH_

#include "string"
#include "vector"

namespace Anaphe 
{
  class IAnnotation;
  class IScatter2D;
  class IVector;
}


#include "Qplotter/qp/qpcurve1D.h"
#include "Qplotter/qp/qpcurve2D.h"
#include "Qplotter/qp/qpcurve3D.h"
#include "Qplotter/qp/qpsummaryinfo.h"
#include "Qplotter/viewer/IViewer.h"
#include "Qplotter/viewer/applInstance.h"
#include "Qplotter/viewer/viewFrame.h"
#include "Qplotter/viewer/propertySet.h"
#include "Qplotter/qp/qplistener.h"
#include "Qplotter/viewer/qpViewerExt.h"


class Zone;
class Page;

#if defined __hpux || (defined __sun && __SUNPRO_CC < 0x500)
#define QPSTD
#else
#define QPSTD std
#endif

namespace QpViewer {

  /** Class to implement a generic dataset for Qplotter.
   */
  class Dataset : public IDataset , IStyleOwner {
  public:
    /// Constructor taking a Vector
    Dataset(Anaphe::IVector& VoP);
    /// Destructor
    virtual ~Dataset();
    /// Change dataset property
    void setProperty (QPSTD::string name, QPSTD::string value);
    /// Change integer property
    void setProperty (QPSTD::string name, int value) ;
    /// List supported dataset options
    void listOptions(QPSTD::ostream &os = QPSTD::cout); 
    /// Draw the dataset
    void draw(QpZone *,bool,bool);
    /// Overloaded assignment operator
    Dataset & operator =(const class Dataset &other);
    /// Update the style on property change (upon notification)
    void update(IStyleOwner::styleId) ;
    /// Get draw style
    IDrawStyle *getStyle();
    void createCurve(QpZone * z);
  protected:
    /// Add an annotation to the dataset
    void addAnnotation(Anaphe::IAnnotation *anno) ;
  private:
    // Reference to the Vector
    Anaphe::IVector& lVoP;
    SummaryTable *summaryT;
    QpCurve1D *qpc1;
    QpCurve2D *qpc2;
    StringPropertySet dataSetProperties;
    VDrawStyle *theCurveStyle;
    bool curveCreated;
  };  
  
  /** Class to implement a generic zone for Qplotter.
   */
  class Zone : public IZone , IStyleOwner {
  public:
    /// Constructor taking the QpPage and the sequence number of the zone
    Zone(QpPage *pm, int _hiZ) ;
    /// Destructor
    virtual ~Zone();
    /// Change zone property
    void setProperty (QPSTD::string name, QPSTD::string value);
    /// List supported zone properties
    void listOptions(QPSTD::ostream &os = QPSTD::cout); 
    /// Add a dataset to the zone
    IDataset *addDataset (Anaphe::IVector& VoP);
    /// Add a scatterplot to the zone
    void addScatter (Anaphe::IScatter2D *scat);
    /// Retrieve a dataset from the zone. Index is sequential from 0 to N-1
    IDataset *getDataset (unsigned int );
    /// Draw the Zone
    void draw();
    /// Set min/max on X axis
    void xMinMax (double min, double max) ;
    /// Set min/max on Y axis
    void yMinMax (double min, double max) ;
    /// Set title
    IText *setTitle(QPSTD::string newTitle);
    /// Add text in zone coordinates
    IText *addText(double x, double y,QPSTD::string text);
    /// Retrieve axis
    virtual ViewerAxis *getAxis (QPSTD::string name);
    /// Get text style
    ITextStyle* getTextStyle();
    /// Get draw style
    IDrawStyle* getDrawStyle();
    /// Update the style on property change (upon notification)
    void update(IStyleOwner::styleId) ;
    /// Get back statistics flag
    bool getStat() const ;

    /// Clear all objects
    void clear();

  private:
    void createHistogram  (Anaphe::IVector *data);
	
  private:
    
    bool stat;
    bool mirrorAxis;
    bool reverseDraw;
    VAxis *xAxis;
    VAxis *yAxis;
    QpZone *hiZone;
    // All the datasets in the zone
    QPSTD::vector <Dataset *> datasets;
    // Pointers to all textual objects.
    QPSTD::vector <VText *> allText;
    StringPropertySet zoneProperties;
    VDrawStyle *theDrawStyle;
    VTextStyle *theTextStyle;
  }; 

  /** Class to implement a generic page for Qplotter.
   */
  class Page : public IPage , IStyleOwner {
  public:
    /// Constructor taking the QpPage and the number of zones on X and Y
    Page(int nx, int ny,QpPage *_page);
    /// Destructor
    virtual ~Page();
    /// Retrieve a zone from the page. Index is sequential from 0 to N-1
    IZone *selectZone(unsigned int);
    /// Draw the Page
    void draw ();
    /// Set title
    virtual IText *setTitle(QPSTD::string newTitle);
    /// Add text in page coordinates
    IText *addText(double x, double y,QPSTD::string text);
    /// Get text style
    ITextStyle* getTextStyle();
    /// Get draw style
    IDrawStyle* getDrawStyle();
    /// Update the style on property change (upon notification)
    void update(IStyleOwner::styleId) ;

    ///

    QpZone *getZone(int index)
      {
	return qppage->getZone(index);
      }
    
    
    void createZones(int nbX, int nbY)
      {
	qppage->createZones(nbX, nbY);
      }
    
    IZone* createIrregularZone(double x, double y,
			       double w, double h)
      {
	qppage->createIrregularZone(x, y, w, h);
	return 0;
      }

    void clear()
      {
	qppage->clear();
      }
    
    void activateOutline(bool On)
      {
	qppage->crossOutline(On);
      }
    
    ///
	
  private:
    int currentZone;
    QpPage *qppage;
    // All the zones in the page
    QPSTD::vector <Zone *> zones;
    // Pointers to all textual objects.
    QPSTD::vector <VText *> allText;
    VDrawStyle *theDrawStyle;
    VTextStyle *theTextStyle;
  };

  /** Class to implement a generic viewer for Qplotter.
   */
  class Viewer : public IViewer {
  public:
    /// Constructor
    Viewer();
    /// Destructor
    virtual ~Viewer();
    /// Clear the Viewer
    void clear();
    /// Show the Viewer
    void show();
    /// Create a new page with nx*ny zones
    IPage *createPage(int nx, int ny);
    /// Dump viewer content to file as Postscript
    void makePS(const char *filename);
    /// Dump viewer content to file as PNG bitmap
    void makePicture(const char *filename);
    /// Draw the current page on the Viewer window
    void draw();
    /// Change viewer property
    void setProperty (QPSTD::string name,QPSTD::string value);
    /// List supported viewer properties
    void listOptions(QPSTD::ostream &os = QPSTD::cout); 
    /// Refresh the viewer
    void refresh();
    void reset() {;};
    
    /// Create coordinate window
    void createCoordWindow();  

    // New AIDA 2 methods

    void createRegion  (double x, double y, double w, double h);
    void createRegions (int columns, int rows);
    void plot (Anaphe::IVector *data, int index = 0);
    void clearRegion (int index);
    void activateOutline(bool On = true);
    
    
  private:
    QApplicationInstance *window;
    Page *page;
    ViewFrame *pageManager;
    StringPropertySet viewerProperties;
  };
 
  /// Global list options function
  void listOptions( QPSTD::string className, StringPropertySet &pSet, QPSTD::ostream &os = QPSTD::cout);

  /// Global color mapping function
  const QColor &MapColor(QPSTD::string value) ;

  /// Global marker mapping function
  const QpMarker::SymStyle MapMarker(QPSTD::string value);

}
#endif
