 /**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Andreas Pfeiffer, CERN/IT                       *
  *                    <andreas.pfeiffer@cern.ch>                      *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/

// implementation file for class AIDAVectorManager 
// 
// Created by: Andreas Pfeiffer
// 
// Last update: 
// 

#ifndef INCLUDED_AIDAVECTORMANAGER_H
#include "VectorOfPoints/AIDAVectorManager.h"
#endif

#include "VectorOfPoints/AIDAVectorFactory.h"

#include <dlfcn.h>
#include <stdio.h>

#include <algorithm>

Anaphe::IVectorManager * createIVectorManager() { return new Anaphe::AIDAVectorManager(0); }

Anaphe::AIDAVectorManager::AIDAVectorManager(Anaphe::IVectorFactory *f) : vFact(f)
{
  if (vFact == 0) {
    vFact = new Anaphe::AIDAVectorFactory();
    if (vFact == 0) {
      std::cerr << "AIDAVectorManager::AIDAVectorManager> could not create AIDAVectorFactory !" << std::endl;
      return;
    } 
  }
  // for call-backs (e.g., unregister)
  vFact->setManager(this);
}


Anaphe::IScalar * Anaphe::AIDAVectorManager::createScalar(double x, double xp, double xm)
{
  if (vFact == 0) {
    std::cerr << "AIDAVectorManager::createScalar> no factory set, don't know how to create Scalar ! " << std::endl;
    return 0;
  }
  Anaphe::IScalar *s = vFact->createScalar(x, xp, xm);
  return s;
}

Anaphe::IPoint * Anaphe::AIDAVectorManager::createPoint()
{
  if (vFact == 0) {
    std::cerr << "AIDAVectorManager::createPoint> no factory set, don't know how to create Point ! " << std::endl;
    return 0;
  }
  Anaphe::IPoint *p = vFact->createPoint();
  return p;
}

Anaphe::IVector * Anaphe::AIDAVectorManager::create()
{
  if (vFact == 0) {
    std::cerr << "AIDAVectorManager::create> no factory set, don't know how to create vector ! " << std::endl;
    return 0;
  }
  Anaphe::IVector *v = vFact->create();
  enroll(v);
  return v;
}

//-ap Anaphe::IVector * Anaphe::AIDAVectorManager::from1D(Anaphe::IHistogram1D *h)
//-ap {
//-ap 	if (vFact == 0) {
//-ap 	  std::cerr << "AIDAVectorManager::from1D> no factory set, don't know how to create vector ! " << std::endl;
//-ap 	  return 0;
//-ap 	}
//-ap 	Anaphe::IVector *v = vFact->from1D(h);
//-ap 	enroll(v);
//-ap 	return v;
//-ap }
//-ap 
//-ap Anaphe::IVector * Anaphe::AIDAVectorManager::from2D(Anaphe::IHistogram2D *h)
//-ap {
//-ap 	if (vFact == 0) {
//-ap 	  std::cerr << "AIDAVectorManager::from2D> no factory set, don't know how to create vector ! " << std::endl;
//-ap 	  return 0;
//-ap 	}
//-ap 	Anaphe::IVector *v = vFact->from2D(h);
//-ap 	enroll(v);
//-ap 	return v;
//-ap }
//-ap 
//-ap Anaphe::IVector * Anaphe::AIDAVectorManager::fromProfile(Anaphe::IProfileHistogram *h, const char * options)
//-ap {
//-ap 	if (vFact == 0) {
//-ap 	  std::cerr << "AIDAVectorManager::fromProf> no factory set, don't know how to create vector ! " << std::endl;
//-ap 	  return 0;
//-ap 	}
//-ap 	Anaphe::IVector *v = vFact->fromProfile(h, options);
//-ap 	enroll(v);
//-ap 	return v;
//-ap }


void Anaphe::AIDAVectorManager::clearAndDestroy() 
{
  for (size_t i=0; i<vectorList.size(); i++) {
    delete vectorList[i];
  }
  vectorList.clear();
}

void Anaphe::AIDAVectorManager::list(std::ostream &os) 
{
  os << std::endl;
  if (vectorList.empty()) {
    os << "no vectors stored\n" << std::endl;
    return;
  }

  os << vectorList.size() << " vectors:" << std::endl;
  for (size_t i=0; i<vectorList.size(); i++) {
    os << " " << i+1 
       << " dim " << vectorList[i]->dimension() 
       << " size " << vectorList[i]->nPoints() << std::endl;
  }
  os << std::endl;

}

Anaphe::IVector * Anaphe::AIDAVectorManager::retrieve1D(int i) 
{ 
  if ( i == 0 ) clearAndDestroy();

  if ( i<0 || i>int(vectorList.size()) ) return 0;
  if ( vectorList[i-1]->dimension() != 1 ) {
    std::cerr << "retrieve1D: requested vector number" << i << " is not 1-D" << std::endl;
    return 0;
  } 
  return (vectorList[i-1]); 
}

Anaphe::IVector * Anaphe::AIDAVectorManager::retrieve2D(int i) 
{ 
  if ( i == 0 ) clearAndDestroy();

  if ( i<0 || i>int(vectorList.size()) ) return 0;
  if ( vectorList[i-1]->dimension() != 2 ) {
    std::cerr << "retrieve2D: requested vector number" << i << " is not 2-D" << std::endl;
    return 0;
  } 
  return (vectorList[i-1]); 
}

void Anaphe::AIDAVectorManager::delete1D(int i)
{ 
  if ( i<=0 || i>int(vectorList.size()) ) return;
  if ( vectorList[i-1]->dimension() != 1 ) {
    std::cerr << "delete1D: requested vector number" << i << " is not 1-D" << std::endl;
  }
  vectorList.erase ( vectorList.begin()+i-1 ); 
}

void Anaphe::AIDAVectorManager::delete2D(int i)
{
  if ( i<=0 || i>int(vectorList.size()) ) return;
  if ( vectorList[i-1]->dimension() != 2 ) {
    std::cerr << "delete2D: requested vector number" << i << " is not 2-D" << std::endl;
  }
  vectorList.erase ( vectorList.begin()+i-1 ); 
}

int  Anaphe::AIDAVectorManager::size() const
{
  return static_cast<int> (vectorList.size());
}

void Anaphe::AIDAVectorManager::clearAll()
{
  clearAndDestroy();
}



// called at destruction of v ...
void Anaphe::AIDAVectorManager::unregister(Anaphe::IVector * v)
{
  std::vector<Anaphe::IVector *>::iterator vItr;

  vItr = std::find (vectorList.begin(), vectorList.end(), v);
  if (vItr != vectorList.end()) {
    vectorList.erase(vItr);
    return;
  }

}


