 /**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Andreas Pfeiffer, CERN/IT                       *
  *                    <andreas.pfeiffer@cern.ch>                      *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/

// implementation file for class Vector 
// 
// Created by: Andreas Pfeiffer
// 
// Last update: 
// 

#ifndef INCLUDED_VECTOR_H
#include "VectorOfPoints/Vector.h"
#endif

#include "Interfaces/IVectorFactory.h"

////  #include "HepFitting/fitmodels.h"
//-ap #include "Interfaces/IAxis.h"
//-ap #include "Interfaces/IHistogram1D.h"
//-ap #include "Interfaces/IHistogram2D.h"
//-ap #include "Interfaces/IProfileHistogram.h"

#include "Interfaces/IAnnotation.h"
#include "Interfaces/IAnnotationFactory.h"
	  

// #if ( ((defined __sun) && (__SUNPRO_CC > 0x0500)) )
// # include <strstream>
// #else
# include <sstream>
// #endif
#include <math.h>
#include <assert.h>


#include <string>

#ifndef AIDA_DONT_USE_STD
using std::string;
using std::ends;

using std::ofstream;
using std::ifstream;

using std::ostream;
using std::istream;

using std::ostringstream;
using std::istringstream;

using std::cerr;
using std::cout;
using std::endl;

using std::vector;
#endif

using Anaphe::IAnnotation;
using Anaphe::IScalar;
using Anaphe::IPoint;
using Anaphe::IVector;
using Anaphe::AIDAScalar;
using Anaphe::AIDAPoint;
using Anaphe::AIDAVector;

// ------------------------------------ Scalar --------------------------------------------

namespace Anaphe {
ostream & operator << (ostream & os, const AIDAScalar &p)  { 
  return os << p.value() << " " << p.ePlus() << " " << p.eMinus() << "      "; 
}
}
AIDAScalar::AIDAScalar(double _v, double _ep, double _em) : 
  val(_v), eP(_ep), eM(_em) { /* nop */; }

AIDAScalar::~AIDAScalar()
{
  // nop
}

void   AIDAScalar::scale(double s) { val *= s; eP *= s; eM *= s; }
void   AIDAScalar::shift(double s) { val += s; }

double AIDAScalar::value()  const { return val; }
double AIDAScalar::ePlus()  const { return eP; }
double AIDAScalar::eMinus() const { return eM; }

void   AIDAScalar::setValue (double valIn) { val = valIn; }
void   AIDAScalar::setEPlus (double valIn) { eP  = valIn; }
void   AIDAScalar::setEMinus(double valIn) { eM  = valIn; }

void AIDAScalar::fromAscii(istream &is)  { is >> val >> eP >> eM; }

//
// Arithmetic operations with other IScalars.
// In order to avoid creating new objects (and problems with ownership)
// the operations modify "this" like: "this" [+-*/]= "other" ...
// The implementation has to correctly propagate the errors (assuming independence).
// (this should be done using an external ErrorPropagationStrategy which can be
// re-defined by the user).
//
void AIDAScalar::addScalar(const IScalar * other)
{
  if ( other == 0 ) return;
 
  val += other->value();

  eP   = sqrt(eP*eP + other->ePlus()*other->ePlus()); 
  eM   = sqrt(eM*eM + other->eMinus()*other->eMinus()); 
}

void AIDAScalar::subScalar(const IScalar * other)
{
  if ( other == 0 ) return;
 
  val -= other->value();
  eP   = sqrt(eP*eP + other->ePlus()*other->ePlus()); 
  eM   = sqrt(eM*eM + other->eMinus()*other->eMinus()); 
}

void AIDAScalar::mulScalar(const IScalar * other)
{
  if ( other == 0 ) return;
 
  // here calculate the errors first, as they need the original value of this !
  eP   = sqrt(eP*eP * other->value()*other->value() + 
	      other->ePlus()*other->ePlus() * val*val); 

  eM   = sqrt(eM*eM * other->value()*other->value() + 
	      other->eMinus()*other->eMinus() * val*val); 

  // ... now modify the value ...
  val *= other->value();
}

// Division by 0 sets this.value to 0, 
// other->error == 0  sets  this.error = 0
void AIDAScalar::divScalar(const IScalar * other)
{
  if ( other == 0 ) return;
 
  // here calculate the errors first, as they need the original value of this !
  if ( other->ePlus() == 0 ) {
    eP = 0;
  } else {
    eP   = sqrt(eP*eP * other->value()*other->value() 
		+ other->ePlus()*other->ePlus() * val*val) 
           / (other->ePlus()*other->ePlus()); 
  }

  if ( other->eMinus() == 0 ) {
    eM = 0;
  } else {
    eM   = sqrt(eM*eM * other->value()*other->value() 
		+ other->eMinus()*other->eMinus() * val*val)
           / (other->eMinus()*other->eMinus()); 
  }

  // ... now modify the value ...
  if ( other->value() == 0 ) {
    val = 0;
  } else {
    val /= other->value();
  }
}

// ... and with doubles (implemented using scale/shift)
void AIDAScalar::add(double value) { shift(  value ); }
void AIDAScalar::sub(double value) { shift( -value ); }
void AIDAScalar::mul(double value) { scale(  value ); }
void AIDAScalar::div(double value) 
{ 
  if (value == 0) { 
    cerr << "AIDAScalar::div> attempt to divide by zero ignored." << endl;
  } else {
    scale(1./value);
  }
}

void AIDAScalar::merge(const IScalar * other)
{
  if ( other == 0 ) return;
 
  double v1 = fabs (eP+eM);
  double v2 = fabs (other->ePlus()+other->eMinus());
  if ((v1 + v2) != 0) {
    val = (v2*other->value() + v1*val) / (v1+v2);
  }else {
    val = (other->value() + val)/2.;
  }

  eP   = sqrt(eP*eP + other->ePlus()*other->ePlus()); 
  eM   = sqrt(eM*eM + other->eMinus()*other->eMinus()); 

}
// ------------------------------------ Point --------------------------------------------

AIDAPoint::AIDAPoint(int _dim, vector<AIDAScalar> _c, AIDAScalar _v) : dim(_dim), val(_v)
{
  if ( (_c.size() > 0) && (static_cast<size_t> (dim) != _c.size()) ) {
    cerr << " AIDAPoint::AIDAPoint> warning, dimension != _c.size, " 
	 << dim << " != " << _c.size() 
	 << " for non-empty point. Reset to _c.size()" << endl; 
    dim = _c.size();
  }

  coord.reserve(dim);

  for (size_t i=0; i<_c.size(); i++) {	// deep copy of list
    coord.push_back(_c[i]);
  }
}

int    AIDAPoint::dimension() const { return dim; }

void   AIDAPoint::setDimension(int _dim)  { 
  if (coord.size() != 0) return; // point already set ==> ignore
  else dim = _dim;
    
  coord.reserve(dim);

  AIDAScalar *pScal;
  for (int i=0; i<dim; i++) {	// create scalars according to dim
    pScal = new AIDAScalar();
    coord.push_back( *pScal );
    delete pScal;
  }
} 

double AIDAPoint::value()     const { return val.value(); }
double AIDAPoint::vPlus()     const { return val.ePlus(); }
double AIDAPoint::vMinus()    const { return val.eMinus(); }

void   AIDAPoint::scaleValue(double scale) { val.scale(scale); }
void   AIDAPoint::shiftValue(double scale) { val.shift(scale); }

double AIDAPoint::coordinate(int index) const { if (!checkDim()) return 0; else return coord[index].value(); }
double AIDAPoint::coordPlus (int index) const { if (!checkDim()) return 0; else return coord[index].ePlus(); }
double AIDAPoint::coordMinus(int index) const { if (!checkDim()) return 0; else return coord[index].eMinus(); }

void   AIDAPoint::scaleCoordinate(int index, double scale) { if (!checkDim()) return; else coord[index].scale(scale); }
void   AIDAPoint::shiftCoordinate(int index, double scale) { if (!checkDim()) return; else coord[index].shift(scale); }

IScalar * AIDAPoint::coordScalar(int index) { if (!checkDim()) return 0; else return (&coord[index]); }
IScalar * AIDAPoint::vScalar() { return &val; }

const IScalar * AIDAPoint::coordScalar(int index) const { if (!checkDim()) return 0; else return (&coord[index]); }
const IScalar * AIDAPoint::vScalar() const { return &val; }

//
// Arithmetic operations with other IPoints, acting on the "value" only,
// the coordinates are not updated.
// In order to avoid creating new objects (and problems with ownership)
// the operations modify "this" like: "this" [+-*/]= "other" ...
// The implementation has to correctly propagate the errors (assuming independence).
// These operations are only defined if the dimension (number of coordinates) is
// equal. 
// 

//-ap: toDo:  merging the coordinates ????

void AIDAPoint::fromAscii(istream &is)   
{
  if (dim < 0) {
    cerr << "AIDAPoint::fromAscii> ERROR: cannot read point for dimension < 0 ! Ignoring" << endl;
    return;
  }

  coord.clear();		// remove old stuff

  AIDAScalar c;			// create temporary
  for (int i=0; i<dimension(); i++) { // for all dimensions required/existing
    c.fromAscii(is);		// read one coord from stream (x,y,z,... in this order)
    coord.push_back(c);		// and add it to coord. list
  }

  val.fromAscii(is);		// read value from file
}

void AIDAPoint::addPoint(const IPoint * other)
{
  if ( other == 0 ) return;
 
  if (dimension() != other->dimension()) {
    cerr << "AIDAPoint::addPoint> cannot add point of dim " << other->dimension()
	 << " to point of dim " << dimension()
	 << endl;
    return;
  }

  val.addScalar(other->vScalar());

}

void AIDAPoint::subPoint(const IPoint * other)
{
  if ( other == 0 ) return;
 
  if (dimension() != other->dimension()) {
    cerr << "AIDAPoint::subPoint> cannot subtract point of dim " << other->dimension()
	 << " from point of dim " << dimension()
	 << endl;
    return;
  }

  val.subScalar(other->vScalar());

}
void AIDAPoint::mulPoint(const IPoint * other)
{
  if ( other == 0 ) return;
 
  if (dimension() != other->dimension()) {
    cerr << "AIDAPoint::mulPoint> cannot multiply point of dim " << other->dimension()
	 << " with point of dim " << dimension()
	 << endl;
    return;
  }

  val.mulScalar(other->vScalar());

}
void AIDAPoint::divPoint(const IPoint * other)
{
  if ( other == 0 ) return;
 
  if (dimension() != other->dimension()) {
    cerr << "AIDAPoint::divPoint> cannot divide point of dim " << dimension()
	 << " by point of dim " << other->dimension()
	 << endl;
    return;
  }

  // more error-checking (zero-divide) is done in the scalar ...
  val.divScalar(other->vScalar());

}

// ... and with doubles (implemented using scale/shift) acting on the "value"
void AIDAPoint::add(double value){ val.add(value); }
void AIDAPoint::sub(double value){ val.sub(value); }
void AIDAPoint::mul(double value){ val.mul(value); }
void AIDAPoint::div(double value)
{
  if (value == 0) { 
    cerr << "AIDAPoint::div> attempt to divide by zero ignored." << endl;
  } else {
    val.div(value);
  }
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// 
// ------------------------------------ Vector --------------------------------------------
// 
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

AIDAVector::AIDAVector() : dim(0), vmax(0), vmin(0), anno(0), myFact(0)
{
  // ensure the annotation is there, even for an empty vector
    IAnnotationFactory * aFact = createIAnnotationFactory();
    assert (aFact);
    anno = aFact->create();
    delete aFact;
}

//-ap AIDAVector::AIDAVector(IHistogram1D *h1, IHistogram2D *h2, IVectorFactory *f) : dim(0), vmax(0), vmin(0), anno(0), myFact(f)
//-ap {
//-ap 	if ((h1 != 0) && (h2!=0)) {
//-ap 	  cerr << "cannot create _one_ vector from 2 histograms ... " << endl;
//-ap 	  return;
//-ap 	}
//-ap 
//-ap 	if (h1 != 0) {
//-ap 	  fromHisto1D(h1);
//-ap 	} 
//-ap 	if (h2 != 0) {
//-ap 	  fromHisto2D(h2);
//-ap 	}
//-ap 
//-ap 	if (h1 == 0 && h2 == 0) {	// ensure the annotation is there, even for an empty vector
//-ap 	  IAnnotationFactory * aFact = createIAnnotationFactory();
//-ap 	  assert (aFact);
//-ap 	  anno = aFact->create();
//-ap 	  delete aFact;
//-ap 	}
//-ap 
//-ap }
//-ap 
//-ap AIDAVector::AIDAVector(IProfileHistogram *h1, const char * opt, IVectorFactory *f) : dim(0), vmax(0), vmin(0), anno(0), myFact(f)
//-ap {
//-ap 	if (h1 != 0) {
//-ap 	  fromProfile(h1, opt);
//-ap 	} 
//-ap 
//-ap 	if (h1 == 0) {	// ensure the annotation is there, even for an empty vector
//-ap 	  IAnnotationFactory * aFact = createIAnnotationFactory();
//-ap 	  assert (aFact);
//-ap 	  anno = aFact->create();
//-ap 	  delete aFact;
//-ap 	}
//-ap 
//-ap }


AIDAVector::AIDAVector(int _dim, int pointsX, int pointsY , vector<AIDAPoint *> p, IVectorFactory *f) :
  dim(_dim) ,vmax (0) ,vmin(0), anno(0), myFact(f)  {
  points = p;
  pointsPerDim.push_back(pointsX);
  pointsPerDim.push_back(pointsY);

  // create a fresh and empty annotation. 
  IAnnotationFactory * aFact = createIAnnotationFactory();
  assert (aFact);
  anno = aFact->create();
  delete aFact;

  computeMinMax();
}

IVector * AIDAVector::deepClone() 
{
  AIDAVector * clone = new AIDAVector();

  clone->deepCopy(*this);
  clone->setFactory(getFactory());

  return clone;
}

AIDAVector::~AIDAVector() 
{ 

//-ap this needs some more design ... if the factory is destroyed before 
//-ap the vector(s), the application crashes ...
//-ap
//-ap  if (myFact !=0) {
//-ap	 myFact->unregister(this);
//-ap  }
//-ap
//-ap end

  clearAndDestroy();
  delete anno;  // as clearAndDestroy only is allowed to reset the annotation ...
}

int  AIDAVector::dimension () const { return dim; }
void AIDAVector::setPointsPerDim(vector<int> v) { if (pointsPerDim.size() != 0) return; else pointsPerDim = v; }

IAnnotation * AIDAVector::annotation() { return anno; }
const IAnnotation * AIDAVector::annotation() const { return anno; }

int      AIDAVector::nPoints()          const { return points.size(); }
int      AIDAVector::nPoints(int index) const { if (size_t(index) <= pointsPerDim.size()) return pointsPerDim[index]; else return 0; } 

// --------------------------------------------------------------------------------

void AIDAVector::removePoint(int i)     
{ 
  if (dimension() != 1) {
    std::cerr << "AIDAVector::removePoint> "
	      << " operation not implemented for vectors with dimension != 1 sorry."
	      << std::endl;
    return;
  }

  if (i < 0 || size_t(i) > points.size()) return;

  points.erase(points.begin()+i);  // see Stroustrup, 3rd ed., p 454

  pointsPerDim[0]--;		// decrement counter for this dimension

}

void AIDAVector::addPoint(IPoint * newPoint) 
{
  if (newPoint == 0) return;

  if (dim > 1 || newPoint->dimension() != 1) {
    std::cerr << "AIDAVector::addPoint> "
	      << " operation not implemented for higher dimensional vectors, sorry."
	      << std::endl;
    return;
  }

  if (dim == 0 && points.size() != 0) {
    std::cerr << "AIDAVector::addPoint> "
	      << " OOOPS  inconsistency: points.size != 0 but dimension() == 0 !"
	      << std::endl;
  }

  // in case the vector was empty before, update these
  if (dim == 0) {
    dim = 1;
  }
  if ( pointsPerDim.empty() ) {
    pointsPerDim.push_back(0);
  }

  points.push_back( dynamic_cast<AIDAPoint *> (newPoint) );

  pointsPerDim[0]++; 

  computeMinMax();

}

// --------------------------------------------------------------------------------

bool AIDAVector::chkIndex(int i) const 
{
  // for other dimensions, this probably means linearized access, so don't check
  if (dimension() == 1) {
    if (i < 0 || i > nPoints(0) ) {
      std::cerr << "AIDAVector::chkIndex> index " << i 
		<< " is out of range [0," << nPoints(0) << "]. Ignoring."
		<< std::endl;
      return false;
    }
  }
  return true;
}

bool AIDAVector::chkIndex(int i, int j) const 
{
  if (dimension() != 2) {
    std::cerr << "AIDAVector::chkIndex> vector has dimension " << dimension() 
	      << " instead of 2. Ignoring. " << std::endl;
    return false;
  }

  if (i < 0 || i > nPoints(0) ||
      j < 0 || j > nPoints(1) ) {
    std::cerr << "AIDAVector::chkIndex> index " << i 
	      << " / " << j
	      << " is out of range [0," << nPoints(0) << "]/[0," << nPoints(1) << "]. Ignoring."
	      << std::endl;
    return false;
  }
  return true;
}

bool AIDAVector::chkIndex(int i, int j, int k) const 
{
  if (dimension() != 3) {
    std::cerr << "AIDAVector::chkIndex> vector has dimension " << dimension() 
	      << " instead of 3. Ignoring. " << std::endl;
    return false;
  }

  if (i < 0 || i > nPoints(0) ||
      j < 0 || j > nPoints(1) ||
      k < 0 || k > nPoints(2) ) {
    std::cerr << "AIDAVector::chkIndex> index " << i 
	      << " / " << j
	      << " / " << k
	      << " is out of range [0," << nPoints(0) << "] "
	      << " / [0," << nPoints(1) << "]"
	      << " / [0," << nPoints(2) << "]"
	      << ". Ignoring."
	      << std::endl;
    return false;
  }
  return true;
}
IPoint * AIDAVector::point(int i)                 
{ 
  if ( !chkIndex(i) ) return 0;
  return points[i]; 
}

IPoint * AIDAVector::point(int i, int j)          
{ 
  if ( !chkIndex(i, j) ) return 0;

  int linIndex = j * nPoints(0) + i;
  return points[linIndex];
}

IPoint * AIDAVector::point(int i, int j, int k ) 
{ 
  if ( !chkIndex(i, j, k) ) return 0;

  int linIndex = k * nPoints(0) * nPoints(1) + j * nPoints(0) + i;

  return points[linIndex];
  
}

const IPoint * AIDAVector::point(int i)                const 
{
  if ( !chkIndex(i) ) return 0;
  return points[i]; 
}

const IPoint * AIDAVector::point(int i, int j)         const 
{ 
  if ( !chkIndex(i, j) ) return 0;

  int linIndex = j * nPoints(0) + i;
  return points[linIndex];
  
}

const IPoint * AIDAVector::point(int i, int j, int k ) const
{ 
  if ( !chkIndex(i, j, k) ) return 0;

  int linIndex = k * nPoints(0) * nPoints(1) + j * nPoints(0) + i;

  return points[linIndex];
  
}

//-ap IPoint * AIDAVector::point(std::vector<int> ijk) const 
//-ap { 
//-ap 	// check if enough indices are given for dimension of vector ....
//-ap 	if (ijk.size() < dimension()) {
//-ap 	  std::cerr << "AIDAVector::point> not enough indices supplied. Ignoring. " << std::endl;
//-ap 	  return 0;
//-ap 	  // return point if valid ...
//-ap 	}
//-ap 
//-ap 	// now loop all dimensions and check if index is valid (i.e., > 0 and < pointsPerDim[index])
//-ap 	// calculate the linearized index assuming the first index runs fastest
//-ap 	int linearizedIndex = 0;
//-ap 	for (int i=0; i<dimension(); i++) {
//-ap 	  if (ijk[i] < 0 || ijk[i] > nPoints(i)) {
//-ap 	    std::cerr << "AIDAVector::point> index for dim " << i 
//-ap 		      << " is " << ijk[i]
//-ap 		      << " and therefore out of range [0," << nPoints(i) << "]. Ignoring."
//-ap 		      << std::endl;
//-ap 	    return 0;
//-ap 	  }
//-ap 	  if (i == 0) {
//-ap 	    linearizedIndex += ijk[i];
//-ap 	  } else {
//-ap 	    linearizedIndex += ijk[i] * nPoints(i-1);
//-ap 	  }
//-ap 	}
//-ap 
//-ap 	// ... finally we have found the requested point
//-ap 	return points[linearizedIndex]; 
//-ap }

// --------------------------------------------------------------------------------

IPoint * AIDAVector::maxValue() const { return vmax; }
IPoint * AIDAVector::minValue() const { return vmin; }

void AIDAVector::scaleV(double scale) 
{ 
  for (size_t i=0; i<points.size(); i++) {
    points[i]->scaleValue(scale);
  }
  computeMinMax();
}

void AIDAVector::scaleCoordinate(int iCoord, double scale) 
{ 
  for (size_t i=0; i<points.size(); i++) { 
    points[i]->scaleCoordinate(iCoord, scale); 
  } 
}

void AIDAVector::shiftV(double val) 
{ 
  for (size_t i=0; i<points.size(); i++) {
    points[i]->shiftValue(val);
  }
  computeMinMax();
}

void AIDAVector::shiftCoordinate(int iCoord, double val) 
{ 
  for (size_t i=0; i<points.size(); i++) { 
    points[i]->shiftCoordinate(iCoord, val); 
  } 
}

//
// Arithmetic operations with other IVectors, acting on the "value" only,
// the coordinates are not updated.
// In order to avoid creating new objects (and problems with ownership)
// the operations modify "this" like: "this" [+-*/]= "other" ...
// The implementation has to correctly propagate the errors (assuming independence).
// These operations are only defined if the dimension (number of coordinates) is
// equal. 
// 

void AIDAVector::addVector(const IVector * other)
{
  if ( !isCompatible(other) ) {
    cerr << "AIDAVector::addVector> cannot add incompatible vectors " << endl;
    return;
  }

  // next update the points
  for (size_t i=0; i<points.size(); i++) {
    points[i]->addPoint(other->point(i));
  }

  mergeAnnotation(other->annotation(), "add:");
  computeMinMax();
}

void AIDAVector::subVector(const IVector * other)
{
  if ( !isCompatible(other) ) {
    cerr << "AIDAVector::subVector> cannot subtract incompatible vectors " << endl;
    return;
  }


  // next update the points
  for (size_t i=0; i<points.size(); i++) {
    points[i]->subPoint(other->point(i));
  }

  mergeAnnotation(other->annotation(), "sub:");
  computeMinMax();
}
void AIDAVector::mulVector(const IVector * other)
{
  if ( !isCompatible(other) ) {
    cerr << "AIDAVector::mulVector> cannot multiply incompatible vectors " << endl;
    return;
  }

  // next update the points
  for (size_t i=0; i<points.size(); i++) {
    points[i]->mulPoint(other->point(i));
  }

  mergeAnnotation(other->annotation(), "mul:");
  computeMinMax();
}
void AIDAVector::divVector(const IVector * other)
{
  if ( !isCompatible(other) ) {
    cerr << "AIDAVector::divVector> cannot divide incompatible vectors " << endl;
    return;
  }

  // next update the points
  for (size_t i=0; i<points.size(); i++) {
    points[i]->divPoint(other->point(i));
  }
  mergeAnnotation(other->annotation(), "div:");
  computeMinMax();
}

void AIDAVector::mergeAnnotation(const IAnnotation * other, string prefix)
{
  // merge the two annotations
  if ( anno != 0 && other != 0 ) {
    for ( int index=0; index < other->size(); index++ ) {
      anno->add(prefix + other->key(index), other->value(index), other->visibilityFlag(index));
    }
  }
}

// ... and with doubles (implemented using scale/shift) acting on the "value"
// toDo: check/update/delete annotation !?
void AIDAVector::add(double value) { shiftV(     value ); }
void AIDAVector::sub(double value) { shiftV(    -value ); }
void AIDAVector::mul(double value) { scaleV(     value ); }
void AIDAVector::div(double value) 
{
  if (value == 0) { 
    cerr << "AIDAVector::div> attempt to divide by zero ignored." << endl;
  } else {
    scaleV(1./value);
  }
}

bool AIDAVector::isCompatible(const IVector * other)
{
  if ( other == 0 ) return false; // we're certainly NOT compatible with a 0-pointer

  // check dimensions first
  if ( dimension() != other->dimension() ) return false;

  // check on overall number of points (for speed)
  if ( nPoints() != other->nPoints() ) return false;

  // check coordinates 
  if (dimension() == 1) {
    int i;
    bool compatible = true;
    const double eps = 1e-6;
    for(i=0; i<nPoints(); i++){
      const IPoint *pt1D1 = point(i);
      const IPoint *pt1D2 = other->point(i); 
      double low1  = pt1D1->coordinate(0)-pt1D1->coordMinus(0);
      double high1 = pt1D1->coordinate(0)+pt1D1->coordPlus(0);
      double low2  = pt1D2->coordinate(0)-pt1D2->coordMinus(0);
      double high2 = pt1D2->coordinate(0)+pt1D2->coordPlus(0);
      if ((fabs(low1-low2) > eps)|| (fabs(high1-high2) > eps)) 
	compatible = false;
    }
    return compatible;
  } else {
    // finally check on number of points in each dimension
    // (dim == 1 is already covered by check on nPoints())
    for (int i=0; i<dimension(); i++) {
      if (pointsPerDim[i] != other->nPoints(i)) return false;
    }
  }
  return true;
}


void AIDAVector::deepCopy(const IVector& other)
{
  clearAndDestroy();		// remove old stuff

  dim = other.dimension();
  int chkSum = 0;
  pointsPerDim.clear();
  for (int i=0; i<dim; i++) {
    pointsPerDim.push_back(other.nPoints(i));
    chkSum += other.nPoints(i);
  }
  if (chkSum != other.nPoints()) {
    cerr << "AIDAVector::deepCopy> Warning: chkSum != nPoints : " << chkSum << " != " << other.nPoints() << endl;
  }

  points.clear();		// just in case ... 
  points.reserve(other.nPoints());

  AIDAPoint *p;
  for (size_t i1=0; i1 < static_cast<size_t> (other.nPoints()); i1++) {
    // to keep the compiler quiet, we cast the const away here ... should probably be fixed properly
    p = const_cast<AIDAPoint *> (dynamic_cast<const AIDAPoint *> (other.point(i1)) );
    if (p == 0) {
      cerr << " AIDAVector::deepCopy> could not dyn-cast to AIDAPoint !?!? index = " << i1 << endl;
    } else {
      p = new AIDAPoint ( *p );
      if (p == 0) {
	cerr << " AIDAVector::deepCopy> could not create new AIDAPoint !?!? index = " << i1 << endl;
      }
    }
    if (p != 0) points.push_back( p );
  }

  // create a fresh and empty annotation. 
  IAnnotationFactory * aFact = createIAnnotationFactory();
  assert (aFact);
  anno = aFact->create();
  delete aFact;
  computeMinMax();

  // implementation dependend ...
  const AIDAVector * av;
  av = dynamic_cast<const AIDAVector *> (&other);
  if (av != 0) {
    myFact = av->getFactory();
  } 
}

void AIDAVector::clearAndDestroy()
{
  for (size_t i=0; i<points.size(); i++) {
    delete points[i];
  }
  points.clear();
  vmax = vmin = 0;

  dim= 0;
  pointsPerDim.clear();

  anno->reset();
}

void AIDAVector::replaceFromFunction(int index, double val)
{
  AIDAPoint *point, *p1d;
  AIDAScalar *x, *v;
  if (dim == 1) {
    point = points[index];  //-ap this is done because SunCC-4.2 does not like: dynamic_cast<AIDAPoint *> (points[index]);
    p1d = dynamic_cast< AIDAPoint * > ( point );
    if (p1d != 0) {
      x = new AIDAScalar( *(dynamic_cast<const AIDAScalar *>(p1d->coordScalar(0))) );
      v = new AIDAScalar(val, 0., 0.);
      delete points[index];		// remove previous point
      point = new AIDAPoint(1,vector<AIDAScalar>(1, *x), *v);
      delete x;			// point is created, scalars no longer needed
      delete v;			// point is created, scalars no longer needed
      points[index] = point;
    }
  } else if (dim == 2) {
    cerr << "Vector::replaceFromFunction> not yet implemented for 2-D, sorry ... " << endl;
    // do whatever is needed here ...
  }
  computeMinMax();
}

void AIDAVector::toAscii(const char * file)  
{ 
  ofstream os(file); 
  if (!os) { 
    cerr << "AIDAVector::toAscii> error opening file " << file << endl; 
    return; 
  } 
  convertToAscii(os); 
}

void AIDAVector::fromAscii(const char *file) 
{ 
  ifstream is(file) ; 
  if (!is) { 
    cerr << "AIDAVector::fromAscii> error opening file " << file << endl; 
    return; 
  } 
  convertFromAscii(is); 
}

const int VectorVersionNumber = 1;

void AIDAVector::convertFromAscii(istream &is)
{
  // read in a few lines and check which version we have ...

  // for now: simply use old format
  convertFromAsciiV1(is);
}

void AIDAVector::convertToAscii(ostream &os)
{
  // write out in new format ...
  // convertToAsciiV2(os);
  // for now use old format
  convertToAsciiV1(os);
}

void AIDAVector::convertToAsciiV2(ostream &os)
{
  // new XML style output
}

void AIDAVector::convertToAsciiV1(ostream &os)
{
  // for a simple gnuplot interface, we start the non-data lines with #
  // this needs to be ignored when reading in the data ...

  os << "# This and the next three lines are needed to read in the vector. Change at own risk :-) " << endl;
  os << "# Vector: version, dimension and size (nX, nY for 2D)" << endl;
  os << "# " << VectorVersionNumber << " " << dim << " " << points.size();
  for (int i=0; i<dim; i++) {
    os << "   " << nPoints(i);
  }
  os << endl;

  // ================================================================================
  // version handling follows
  // ================================================================================

  if (VectorVersionNumber == 1) { // need to read/write the annotation and statistisc (if existent) from here
    if (anno != 0) {		// annotation exists, write out ...
      anno->toAscii(os);
    } else {			// write empty annotation to make sure there is something ...
      os << "# <AIDAAnnotation size=\"0\" />" << endl;
      os << "# </AIDAAnnotation>" << endl;
    }
  }

  // ================================================================================
  // end of version handling 
  // ================================================================================

  os << "# x ex+ ex-    ";
  if (dim >= 2) os << "y ey+ ey-  ";
  if (dim >= 3) os << "z ez+ ez-  ";
  if (dim  > 3) os << "   ....    ";
  os << "val ev+ ev-    " << endl;

  AIDAPoint *p1d;
  const AIDAScalar * c;
  const AIDAScalar * v;
  for (size_t i1=0; i1<points.size(); i1++) {
    p1d = points[i1];		//-ap this is done because SunCC-4.2 does not like: dynamic_cast<AIDAPoint *> (points[i1]);
    p1d = dynamic_cast<AIDAPoint *> (p1d);
    if (p1d == 0) break;
    for (int i=0; i<dim; i++) {
      c = dynamic_cast<const AIDAScalar *> (p1d->coordScalar(i));
      if (c != 0) os << *c << "    ";
      else cerr << "AIDAVector::convertToAscii> dynamic cast for coord " << i << " not possible" << endl;
    }
    v = dynamic_cast<const AIDAScalar *> (p1d->vScalar());
    if (c != 0) os << *v << "    ";
    else cerr << "AIDAVector::convertToAscii> dynamic cast for value not possible" << endl;
    os << endl;
  }
}

void AIDAVector::convertFromAsciiV2(istream &is)
{
}

void AIDAVector::convertFromAsciiV1(istream &is)
{

  clearAndDestroy();		// reset all internals.

  const int maxSize=1024;	// this is dangerous as lines could be longer 
  char cbuf[maxSize];

  is.getline(cbuf, maxSize);		// ignore first two lines
  is.getline(cbuf, maxSize);

  is.getline(cbuf, maxSize);
  int versionFound = -1;
  size_t size;
  {
    int tmpDim=0;
    istringstream buf( &cbuf[1] ); // ignore # sign at start of line (see convertToAscii)
    buf >> versionFound >> dim >> size;
    if (dim < 0) {
      std::cerr << "Vector::convertFromAscii> ERROR "
		<< "found illegal dimension =" << dim
		<< std::endl;
      return;
    }
    if (size == 0) {
      std::cerr << "Vector::convertFromAscii> ERROR "
		<< "found illegal size =" << size
		<< std::endl;
      return;
    }
    for (int i=0; i<dim; i++) {
      buf >> tmpDim;
      if (tmpDim == 0) {
	if (dim == 1) {
	  tmpDim = size;
	} else {
	  std::cerr << "Vector::convertFromAscii> ERROR "
		    << "found illegal size for dim=" << i 
		    << std::endl;
	  return;
	}
      }
      pointsPerDim.push_back(tmpDim);
    }
  }

  if (versionFound < 0 || versionFound > VectorVersionNumber) {
    cerr << "Vector::convertFromAscii> ERROR "
	 << "in version number."
	 << " Version found: " << versionFound
	 << " this version can handle only " << VectorVersionNumber << " or lower."
	 << endl;
    return;
  }

  if ( (dim != 1) && (dim !=2) ) {
    cerr << "Vector::convertFromAscii> ERROR wrong dimension : " << dim << " expected 1 or 2 " << endl;
    return;
  }

  // ================================================================================
  // version handling follows
  // ================================================================================

  if (versionFound == 1) { // need to read/write the annotation from here
    if (anno == 0) {		// annotation does not yet exist, but is required by version.
      IAnnotationFactory * aFact = createIAnnotationFactory();
      assert (aFact);
      anno = aFact->create();	// create empty annotation
      delete aFact;
    }
    anno->fromAscii(is);	// read in from file
  }

  if (is.eof()) {
    cerr << "Vector::convertFromAscii> ERROR "
	 << "end of file found while reading annotation. Is <AIDAAnnotation> ... </AIDAAnnotation> present ?"
	 << endl;
    return;
  }
  if (is.fail() || is.bad() || !is.good() ) {
    cerr << "Vector::convertFromAscii> ERROR "
	 << "while reading annotation."
	 << endl;
    return;
  }

  // ================================================================================
  // end of version handling 
  // ================================================================================

  is.getline(cbuf, maxSize);		// ignore next line : # x ex+ ex- ...

  if (is.eof()) {
    cerr << "Vector::convertFromAscii> ERROR "
	 << "end of file found while reading annotation. Is <AIDAAnnotation> ... </AIDAAnnotation> present ?"
	 << endl;
    return;
  }
  if (is.fail() || is.bad() || !is.good() ) {
    cerr << "Vector::convertFromAscii> ERROR "
	 << "while reading annotation."
	 << endl;
    return;
  }

  cerr << "reading in " << dim << "D-vector of size " << size;
  if (dim >= 1) {
    cerr << " (";
    for (int i=0; i<dim; i++) {
      cerr << nPoints(i) << ", ";
    }
    cerr << ") ";
  }
  cerr  << endl;

  points.reserve(size);

  AIDAPoint *point;
  for (size_t i=0; i<size; i++) {

    is.getline(cbuf, maxSize);
    if (is.eof()) {
      cerr << "ERROR: EOF found while reading in vector." << endl;
      break;
    }
    if (is.fail() || is.bad() || !is.good() ) {
      cerr << "Vector::convertFromAscii> ERROR "
	   << "while reading annotation."
	   << endl;
      break;
    }

    point = new AIDAPoint(dim);
    { 
      istringstream buf( cbuf );
      point->fromAscii(buf);
      //      cerr << "point i=" << i << ": "; point->toAscii(cerr); cerr << endl;
      points.push_back(point);
    }
  }
  computeMinMax();
}


//-ap void AIDAVector::fromHisto1D(IHistogram1D *h)
//-ap {
//-ap 	clearAndDestroy();		// get rid of possible old contents
//-ap 
//-ap 	dim = 1;
//-ap 
//-ap 	int nBins = h->xAxis()->bins();
//-ap 
//-ap 	points.reserve(nBins);
//-ap 
//-ap 	AIDAPoint *point = 0;
//-ap 	AIDAScalar *x, *v;
//-ap 	vector<AIDAScalar> coords;
//-ap 	for(int i = 0; i < nBins; i++ ) {
//-ap 	  v = new AIDAScalar(h->binHeight(i), h->binError(i), h->binError(i));
//-ap 	  // for gravity binned histograms (default in HTL) we need to calculate the e+/e- ...
//-ap 	  x = new AIDAScalar(h->xAxis()->binCentre(i), 
//-ap 			     h->xAxis()->binUpperEdge(i) - h->xAxis()->binCentre(i), 
//-ap 			     h->xAxis()->binCentre(i)    - h->xAxis()->binLowerEdge(i));
//-ap 	  coords.push_back(*x);
//-ap 	  point = new AIDAPoint(dim, coords, *v);
//-ap 	  delete x;			// point is created, scalars no longer needed
//-ap 	  delete v;			// point is created, scalars no longer needed
//-ap 	  coords.clear();
//-ap 
//-ap 	  points.push_back(point);
//-ap 	}
//-ap 
//-ap 	pointsPerDim.push_back( int(nBins) ); 
//-ap 	computeMinMax();
//-ap 
//-ap 	IAnnotationFactory * aFact = createIAnnotationFactory();
//-ap 	assert (aFact);
//-ap 	if (h->annotation() != 0) {
//-ap 	  anno = aFact->create( h->annotation() );
//-ap 	} else {
//-ap 	  anno = aFact->create();
//-ap 	}
//-ap 	delete aFact;
//-ap 
//-ap 	// add histogram title to the annotation (otherwise it gets lost ...)
//-ap 	anno->add("title", h->title(), false);    // default is: invisible ...
//-ap 
//-ap 	// add statistics info from histo to annotation
//-ap 	// separate scopes to re-set the ostrstream
//-ap 	{ ostrstream val; val << h->entries() << ends; anno->add("Entries", val.str(), true); }
//-ap 	{ ostrstream val; val << h->mean()    << ends; anno->add("Mean"   , val.str(), true); }
//-ap 	{ ostrstream val; val << h->rms()     << ends; anno->add("RMS"    , val.str(), true); }
//-ap 
//-ap 	{ ostrstream val; val << h->binEntries(IHistogram::OVERFLOW_BIN)  << ends; anno->add("Overflow" , val.str(), true); }
//-ap 	{ ostrstream val; val << h->binEntries(IHistogram::UNDERFLOW_BIN) << ends; anno->add("Underflow", val.str(), true); }
//-ap }
//-ap 
//-ap void AIDAVector::fromHisto2D(IHistogram2D *h)
//-ap {
//-ap 	clearAndDestroy();		// get rid of possible old contents
//-ap 
//-ap 	dim = 2;
//-ap 
//-ap 	int nBinsX = h->xAxis()->bins();
//-ap 	int nBinsY = h->yAxis()->bins();
//-ap 
//-ap 	double *bcX  = new double [nBinsX]; // Array of bin centers in X
//-ap 	double *hbwX = new double [nBinsX]; // Array of bin centers in Y
//-ap 	double *bcY  = new double [nBinsY]; // Array of half bin widht in X
//-ap 	double *hbwY = new double [nBinsY]; // Array of half bin widht in Y
//-ap 
//-ap 	// Compute once for all bin centers and half bin widths
//-ap 	int i,j;
//-ap 	for (i=0;i<nBinsX;i++) {
//-ap 	  bcX [i] = h->xAxis()->binCentre(i  );
//-ap 	  hbwX[i] = h->xAxis()->binWidth(i  )/2.; //-ap: need for binLowerEdge/binUpperEdge as in 1-D ???  (11may00)
//-ap 	}
//-ap 	for (j=0;j<nBinsY;j++) {
//-ap 	  bcY [j] = h->yAxis()->binCentre(j);
//-ap 	  hbwY[j] = h->yAxis()->binWidth(j)/2.;
//-ap 	}
//-ap 
//-ap 	// Reserve space for the STL vector
//-ap 	points.reserve(nBinsX*nBinsY);
//-ap 	AIDAPoint *point = 0;
//-ap 	vector<AIDAScalar> coords;
//-ap 	for(i = 0; i < nBinsX; i++ ) {
//-ap 	  AIDAScalar x(bcX[i], hbwX[i], hbwX[i]);   
//-ap 	  for(j = 0; j < nBinsY; j++ ) {
//-ap 	    AIDAScalar y(bcY[j],hbwY[j],hbwY[j]);
//-ap 	    AIDAScalar v(h->binHeight (i,j), h->binError (i,j), h->binError (i,j));
//-ap 
//-ap 	    coords.clear();
//-ap 	    coords.push_back(x);
//-ap 	    coords.push_back(y);
//-ap 
//-ap 	    point = new AIDAPoint(dim, coords, v);
//-ap 	    points.push_back(point);
//-ap 	  }
//-ap 	}
//-ap 
//-ap 	pointsPerDim.push_back( int(nBinsX) ); // order is important here (x/y) !
//-ap 	pointsPerDim.push_back( int(nBinsY) ); 
//-ap 	computeMinMax();
//-ap 	delete [] bcX;
//-ap 	delete [] hbwX;
//-ap 	delete [] bcY;
//-ap 	delete [] hbwY;
//-ap 
//-ap 	IAnnotationFactory * aFact = createIAnnotationFactory();
//-ap 	assert (aFact);
//-ap 	if (h->annotation() != 0) {
//-ap 	  anno = aFact->create( h->annotation() );
//-ap 	} else {
//-ap 	  anno = aFact->create();
//-ap 	}
//-ap 	delete aFact;
//-ap 
//-ap 	// add histogram title to the annotation (otherwise it gets lost ...)
//-ap 	anno->add("title", h->title(), false);    // default is: invisible ...
//-ap 
//-ap 	// add statistics info from histo to annotation
//-ap 	// separate scopes for re-init of ostrstream
//-ap 	{ ostrstream val; val << h->entries()      << ends; anno->add("Entries", val.str(), true); }
//-ap 	{ ostrstream val; val << h->extraEntries() << ends; anno->add("Sum Over/Underflow", val.str(), true); }
//-ap 
//-ap }
//-ap 
//-ap void AIDAVector::fromProfile(IProfileHistogram *h, const char * options)
//-ap {
//-ap 	clearAndDestroy();		// get rid of possible old contents
//-ap 
//-ap 	bool useSpread = false;
//-ap 	if (options != 0) {		// just about anything else will make use of spread
//-ap 	  if (string(options) != "") {
//-ap 	    useSpread = true;
//-ap 	  }
//-ap 	}
//-ap 
//-ap 	dim = 1;
//-ap 
//-ap 	int nBins = h->xAxis()->bins();
//-ap 
//-ap 	points.reserve(nBins);
//-ap 
//-ap 	AIDAPoint *point = 0;
//-ap 	AIDAScalar *x, *v;
//-ap 	vector<AIDAScalar> coords;
//-ap 	for(int i = 0; i < nBins; i++ ) {
//-ap 	  if (useSpread) {
//-ap 	    v = new AIDAScalar(h->binHeight(i), h->binSpread(i), h->binSpread(i));
//-ap 	  } else {
//-ap 	    v = new AIDAScalar(h->binHeight(i), h->binError(i), h->binError(i));
//-ap 	  }
//-ap 	  // for gravity binned histograms (default in HTL) we need to calculate the e+/e- ...
//-ap 	  x = new AIDAScalar(h->xAxis()->binCentre(i), 
//-ap 			     h->xAxis()->binUpperEdge(i) - h->xAxis()->binCentre(i), 
//-ap 			     h->xAxis()->binCentre(i)    - h->xAxis()->binLowerEdge(i));
//-ap 	  coords.push_back(*x);
//-ap 	  point = new AIDAPoint(dim, coords, *v);
//-ap 	  delete x;			// point is created, scalars no longer needed
//-ap 	  delete v;			// point is created, scalars no longer needed
//-ap 	  coords.clear();
//-ap 
//-ap 	  points.push_back(point);
//-ap 	}
//-ap 
//-ap 	pointsPerDim.push_back( int(nBins) ); 
//-ap 	computeMinMax();
//-ap 
//-ap 	IAnnotationFactory * aFact = createIAnnotationFactory();
//-ap 	assert (aFact);
//-ap 	if (h->annotation() != 0) {
//-ap 	  anno = aFact->create( h->annotation() );
//-ap 	} else {
//-ap 	  anno = aFact->create();
//-ap 	}
//-ap 	delete aFact;
//-ap 
//-ap 	// add histogram title to the annotation (otherwise it gets lost ...)
//-ap 	anno->add("title", h->title(), false);    // default is: invisible ...
//-ap 
//-ap 	// add statistics info from histo to annotation
//-ap 	// separate scopes to re-set the ostrstream
//-ap 	{ ostrstream val; val << h->entries() << ends; anno->add("Entries", val.str(), true); }
//-ap 
//-ap 	{ ostrstream val; val << h->binEntries(IHistogram::OVERFLOW_BIN)  << ends; anno->add("Overflow" , val.str(), true); }
//-ap 	{ ostrstream val; val << h->binEntries(IHistogram::UNDERFLOW_BIN) << ends; anno->add("Underflow", val.str(), true); }
//-ap 
//-ap 	if (useSpread) {
//-ap 	  { ostrstream val; val << h->rms()     << ends; anno->add("Errors are"    , "spread", true); }
//-ap 	} else {
//-ap 	  { ostrstream val; val << h->rms()     << ends; anno->add("Errors are"    , "errors", true); }
//-ap 	}
//-ap }

void AIDAVector::setDimension(int dimIn) 
{
  dim = dimIn;
}

//  set from external source ...
void AIDAVector::setData(std::vector<AIDAPoint *> data, std::vector<int> ppd )
{
  std::vector<AIDAPoint *>::const_iterator cItrData;
  std::vector<int>::const_iterator cItrDim;

  for (cItrData = data.begin(); cItrData != data.end(); cItrData++ ) {
    points.push_back( *cItrData );
  }

  for (cItrDim = ppd.begin(); cItrDim != ppd.end(); cItrDim++ ) {
    pointsPerDim.push_back( *cItrDim ); 
  }
  computeMinMax();
}


void AIDAVector::computeMinMax() 
{
  vmin=0;
  vmax=0;
  // imply use linearized access ...
  for(int i = 0; i < nPoints(); i++ ) {
    if (vmax == 0) vmax = points[i];
    if (vmin == 0) vmin = points[i];
    if (points[i]->value() > vmax->value()) vmax = points[i];
    if (points[i]->value() < vmin->value()) vmin = points[i];
  }
}

const IPoint * AIDAVector::computeNonZeroMin() 
{
  IPoint * nzMin = 0;

  if (dim == 1) {
    for(int i = 0; i < pointsPerDim[0]; i++ ) {
      if (points[i]->value () == 0 && 
	  points[i]->vPlus () == 0 && 
	  points[i]->vMinus() == 0 ) continue;

      if ( nzMin == 0 ) nzMin = points[i];
      if ( points[i]->value() < nzMin->value() ) nzMin = points[i];
    }
  }
  if (dim == 2) {
    int i,j,k=0;
    for(i = 0; i < pointsPerDim[0]; i++ ) {
      for(j = 0; j < pointsPerDim[1]; j++ ) {
	if (points[k]->value () == 0 && 
	    points[k]->vPlus () == 0 && 
	    points[k]->vMinus() == 0 ) continue;
	
	if ( nzMin == 0 ) nzMin = points[k];
	if ( points[k]->value() < nzMin->value() ) nzMin = points[k];
	k++;
      }
    }
  } // end if dim==2

  return nzMin;

}

Anaphe::IVectorFactory * AIDAVector::getFactory() const
{
  return myFact;
}

void AIDAVector::setFactory(IVectorFactory *f)
{
  myFact = f;
}
