 /**********************************************************************
  *                                                                    *
  * Copyright (c) 2002 Andreas Pfeiffer, CERN/IT                       *
  *                    <andreas.pfeiffer@cern.ch>                      *
  *                                                                    *
  * This library is free software; you can redistribute it and/or      *
  * modify it under the terms of the GNU Lesser General Public         *
  * License as published by the Free Software Foundation; either       *
  * version 2.1 of the License, or (at your option) any later version. *
  *                                                                    *
  * This library is distributed in the hope that it will be useful,    *
  * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
  * Lesser General Public License for more details.                    *
  *                                                                    *
  * You should have received a copy of the GNU Lesser General Public   *
  * License along with this library (see file COPYING); if not, write  *
  * to the Free Software Foundation, Inc., 59 Temple Place, Suite      *
  * 330, Boston, MA 02111-1307 USA, or contact the author              *
  **********************************************************************/


// 
// header file for class Vector 
// Created by: Andreas Pfeiffer  at Wed Sep  1 11:12:01 1999
// 
// Last update: 
//              
// 
#ifndef INCLUDED_VECTOR_H
#define INCLUDED_VECTOR_H

#include <vector>
#include <fstream>
#include <iostream>

#include "Interfaces/IVector.h"

#include "Interfaces/IAnnotation.h"

namespace Anaphe {

  class IVectorFactory;

  // ------------------------------------ Scalar --------------------------------------------

  class AIDAScalar : virtual public IScalar { 

  public: 
    AIDAScalar(double _v=0, double _ep=0, double _em=0);

    virtual ~AIDAScalar();

  public:
    // for assignment and copyConstructor rely on compiler's version

  public: 		// public methods
    void   scale(double s);
    void   shift(double s);

  public: 		// public methods
    double value()  const;
    double ePlus()  const;
    double eMinus() const;

  public: 		// public methods
    void setValue (double val);
    void setEPlus (double val);
    void setEMinus(double val);

  public:
#ifdef AIDA_DONT_USE_STD
    void fromAscii(     istream &is);
#else
    void fromAscii(std::istream &is);
#endif

  public:
    //
    // Arithmetic operations with other IScalars.
    // In order to avoid creating new objects (and problems with ownership)
    // the operations modify "this" like: "this" [+-*/]= "other" ...
    // The implementation has to correctly propagate the errors (assuming independence).
    // (this should be done using an external ErrorPropagationStrategy which can be
    // re-defined by the user).
    // Division by 0 sets this.value to 0, 
    // other.error == 0  sets  this.error = 0
    //
    void addScalar(const IScalar * other);
    void subScalar(const IScalar * other);
    void mulScalar(const IScalar * other);
    void divScalar(const IScalar * other);

    // ... and with doubles (implemented using scale/shift)
    void add(double val);
    void sub(double val);
    void mul(double val);
    void div(double val);

    // merging with another scalar. This sets the value to the
    // weighted average of the two and the error to the combined error
    void merge(const IScalar * other);

  private:		// private attributes
    double val;
    double eP;
    double eM;

  }; 

  // ------------------------------------ Point --------------------------------------------

  class AIDAPoint : virtual public IPoint { 

  public: 

#ifdef AIDA_DONT_USE_STD
    AIDAPoint(int _dim=0,      vector<AIDAScalar> _c=     vector<AIDAScalar>(), AIDAScalar _v=0);
#else
    AIDAPoint(int _dim=0, std::vector<AIDAScalar> _c=std::vector<AIDAScalar>(), AIDAScalar _v=0);
#endif
    virtual ~AIDAPoint() { coord.clear(); /* clearAndDestroy(); */ }

  public:
    // for assignment and copyConstructor rely on compiler's version

  public:
    virtual int    dimension() const;
    virtual void   setDimension(int _dim);

  public:
    virtual double value()     const;
    virtual double vPlus()     const;
    virtual double vMinus()    const;

  public:
    // non-const methods to manipulate the contents ...
    virtual void scaleValue(double scale);
    virtual void shiftValue(double scale);

    // non-const methods to manipulate the contents ...
    virtual IScalar * coordScalar(int index);
    virtual IScalar * vScalar();

    virtual void   scaleCoordinate(int index, double scale);
    virtual void   shiftCoordinate(int index, double scale);

    // const methods ...
    virtual double coordinate(int index) const;
    virtual double coordPlus (int index) const;
    virtual double coordMinus(int index) const;

    virtual const IScalar * coordScalar(int index) const;
    virtual const IScalar * vScalar() const;

  public:
#ifdef AIDA_DONT_USE_STD
    virtual void fromAscii(     istream &is);
#else
    virtual void fromAscii(std::istream &is);
#endif

  public:
    //
    // Arithmetic operations with other IPoints, acting on the "value" only,
    // the coordinates are not updated.
    // In order to avoid creating new objects (and problems with ownership)
    // the operations modify "this" like: "this" [+-*/]= "other" ...
    // The implementation has to correctly propagate the errors (assuming independence).
    // These operations are only defined if the dimension (number of coordinates) is
    // equal. 
    // 
    void addPoint(const IPoint * other);
    void subPoint(const IPoint * other);
    void mulPoint(const IPoint * other);
    void divPoint(const IPoint * other);
    // ... and with doubles (implemented using scale/shift) acting on the "value"
    void add(double val);
    void sub(double val);
    void mul(double val);
    void div(double val);

  protected:  // non-virtual methods
    inline bool checkDim() const {
      if ( static_cast<const size_t> (dim) != coord.size() ) 
#ifdef AIDA_DONT_USE_STD
	cerr << "AIDAPoint::checkDim()> dim != size : " << dim << " " << coord.size() <<      endl;
#else
      std::cerr << "AIDAPoint::checkDim()> dim != size : " << dim << " " << coord.size() << std::endl;
#endif
      return (static_cast<const size_t> (dim) == coord.size()); 
    }

  private:
    //-ap void clearAndDestroy();

  private:
    int dim;
#ifdef AIDA_DONT_USE_STD
    vector<AIDAScalar> coord;
#else
    std::vector<AIDAScalar> coord;
#endif
    AIDAScalar val;

  };


  // ------------------------------------ Vector --------------------------------------------

  class AIDAVector : virtual public IVector { 

  public: 
    AIDAVector();
//-ap 	  AIDAVector(IHistogram1D *h1, IHistogram2D *h2, IVectorFactory *f=0);
//-ap 	  AIDAVector(IProfileHistogram *h1, const char * options = 0, IVectorFactory *f=0);
#ifdef AIDA_DONT_USE_STD
    AIDAVector(int _dim, int pointsX, int pointsY ,      vector<AIDAPoint *> p, IVectorFactory *f=0);
#else
    AIDAVector(int _dim, int pointsX, int pointsY , std::vector<AIDAPoint *> p, IVectorFactory *f=0);
#endif

    virtual ~AIDAVector();

  public:
    // for assignment and copyConstructor rely on compiler's version
    IVector * deepClone();	// return a (deep) copy of self

  public: 		// public methods from the to-be-defined abstract interface
    int  dimension () const;
#ifdef AIDA_DONT_USE_STD
    void setPointsPerDim(     vector<int> v);
#else
    void setPointsPerDim(std::vector<int> v);
#endif

  public:
    IAnnotation * annotation();
    const IAnnotation * annotation() const;

  public: 		// public methods
    int      nPoints()          const;
    int      nPoints(int index) const;

    // remove point specified by index in linear access (even for higher dimensions)
    void removePoint(int index); 
    // add new point 
    void addPoint(IPoint * newPoint); 

    // modifiers
    IPoint * point(int i)               ;
    IPoint * point(int i, int j)        ;
    IPoint * point(int i, int j, int k );

    const IPoint * point(int i)                const;
    const IPoint * point(int i, int j)         const;
    const IPoint * point(int i, int j, int k ) const;
    //-ap IPoint * point(std::vector<int> ijk) const;

  public: 		// public methods
    IPoint * maxValue() const;
    IPoint * minValue() const;

  public: 		// public methods
    void scaleV(double scale);
    void shiftV(double scale);

    void scaleCoordinate(int index, double scale);
    void shiftCoordinate(int index, double shift);


  public: 		// public methods
    void toAscii(const char * file);
    void fromAscii(const char *file);

//    void fromHisto1D(IHistogram1D *);
//    void fromHisto2D(IHistogram2D *);
//    void fromProfile(IProfileHistogram *, const char * options = 0);

    void replaceFromFunction(int, double);

  public:
    //
    // Arithmetic operations with other IVectors, acting on the "value" only,
    // the coordinates are not updated.
    // In order to avoid creating new objects (and problems with ownership)
    // the operations modify "this" like: "this" [+-*/]= "other" ...
    // The implementation has to correctly propagate the errors (assuming independence).
    // These operations are only defined if the dimension (number of coordinates) is
    // equal. 
    // 
    void addVector(const IVector * other);
    void subVector(const IVector * other);
    void mulVector(const IVector * other);
    void divVector(const IVector * other);
    // ... and with doubles (implemented using scale/shift) acting on the "value"
    void add(double val);
    void sub(double val);
    void mul(double val);
    void div(double val);

  public:
    // non-Interface methods:
    IVectorFactory * getFactory() const;

    void setFactory(IVectorFactory * f);

  protected:

    bool chkIndex(int i)               const;
    bool chkIndex(int i, int j)        const;
    bool chkIndex(int i, int j, int k) const;

#ifdef AIDA_DONT_USE_STD
    void convertToAscii  (     ostream &);
    void convertFromAscii(     istream &);

    void convertFromAsciiV1(   istream &);
    void convertFromAsciiV2(   istream &);
			     
    void convertToAsciiV1  (   ostream &); // deprecated ! Don't use this !
    void convertToAsciiV2  (   ostream &);

    void mergeAnnotation(const IAnnotation * other,     string prefix);
#else
    void convertToAscii  (std::ostream &);
    void convertFromAscii(std::istream &);

    void convertFromAsciiV1(std::istream &);
    void convertFromAsciiV2(std::istream &);

    void convertToAsciiV1  (std::ostream &); // deprecated ! Don't use this !
    void convertToAsciiV2  (std::ostream &);

    void mergeAnnotation(const IAnnotation * other, std::string prefix);
#endif

    bool isCompatible(const IVector * other);


  private:		// private methods
    void deepCopy(const IVector&);

    // methods not in the Abstract Interface

  public: 		
    void clearAndDestroy();
    void setDimension(int dim);
    void computeMinMax();
    const IPoint * computeNonZeroMin();

    //  set from external source ...
    void setData(std::vector<AIDAPoint *> , std::vector<int> pointsPerDim);

  private:

#ifdef AIDA_DONT_USE_STD
    vector<AIDAPoint *> points;
    vector<int> pointsPerDim;     // number of points along each dimension
#else
    std::vector<AIDAPoint *> points;
    std::vector<int> pointsPerDim;     // number of points along each dimension
#endif

    int dim;

    AIDAPoint *vmax;
    AIDAPoint *vmin;

    IAnnotation *anno;

    IVectorFactory * myFact;
  }; 

} //end namespace Anaphe

#endif /* INCLUDED_VECTOR_H */ 
