# Helper targets/commands for package testing
#
# This file provides
#
#   awk filter to help implementing local test
#   (i.e. testing against the code in the checkout area).
#
#   a showlibs target to call ldd and print the list of libraries
#   used by the testprogram
#
#   a testrun target to run the testprogram with a modified LD_LIBRARY_PATH

# awk filter to create the library list for loader
# input is a list of shared library pathnames
# splits the basename from the directory
# if the directory is new, append a -L<dir> string
# extract library name comaptible with -l option 
# debug libraries (name containing dbg token) are ignored
# libraries with version numbers (>1 dot in the name) are ignored
# the leading "lib" string is dropped.
# Example:
# Input:
# ../../code/T_/HTL/Linux/libT_Histoxx.so
# ../../code/T_/HTL/Linux/libT_Histoxx-dbg.1.3.1.2.so
# ../../code/T_/HTL/Linux/libT_Histoxx-dbg.so
# ../../code/P_/HTL/Linux/libP_Histoxx.so
# ../../code/P_/HTL/Linux/libP_Histoxx.1.3.1.2.so
# ../../code/P_/HTL/Linux/libP_Histoxx-dbg.1.3.1.2.so
# ../../code/P_/HTL/Linux/libP_Histoxx-dbg.so
#
# Output:
# -L../../code/T_/HTL/Linux/ -lT_Histoxx -L../../code/P_/HTL/Linux/ -lP_Histoxx
#

AWK_LIBS_FILTER='BEGIN { lastdir= "" ;  outline="" }\
{\
  levels=split($$0,elems,"/"); \
  curdir = substr($$0,1,length-length(elems[levels])); \
  if (curdir != lastdir) {\
    outline = sprintf ("%s%s",outline," -L"curdir);\
    lastdir=curdir;\
  }\
  dotsplus1 = split(elems[levels],libnames,".");\
    if (dotsplus1 == 2)\
      if (index (libnames[1],"dbg") == 0) \
	outline = sprintf ("%s%s",outline," -l"substr(libnames[1],4));\
}\
END { print outline }'


# awk filter to create the include directory list for loader
# input is a list of header files
# splits the basename from the directory
# if the directory is new, append a -I<dir> string
# the trailing package directory is dropped.
# Example:
# Input:
# ../../code/Common/HTL/platform.h
# ../../code/T_/HTL/T_Lifetime.h
# Output:
# -I../../code/Common/ -I../../code/T_/ 
#
AWK_INCS_FILTER='BEGIN { lastdir= "" ;  outline="" }\
{\
  levels=split($$0,elems,"/"); \
  usefulPathLen=length-(length(elems[levels])+length(elems[levels-1]))-1; \
  curdir=substr($$0,1,usefulPathLen);\
  if (curdir != lastdir) {\
    outline = sprintf ("%s%s",outline," -I"curdir);\
    lastdir=curdir;\
  }\
}\
END { print outline }'

# awk filter to figure out the path to prepend to LD_LIBRARY_PATH
# input is a list of shared library pathnames
# splits the basename from the directory
# if the directory is new, append a <dir>: string
# Example:
# Input:
# ../../code/T_/HTL/Linux/libT_Histoxx.so
# ../../code/T_/HTL/Linux/libT_Histoxx-dbg.1.3.1.2.so
# ../../code/T_/HTL/Linux/libT_Histoxx-dbg.so
# ../../code/P_/HTL/Linux/libP_Histoxx.so
# ../../code/P_/HTL/Linux/libP_Histoxx.1.3.1.2.so
# ../../code/P_/HTL/Linux/libP_Histoxx-dbg.1.3.1.2.so
# ../../code/P_/HTL/Linux/libP_Histoxx-dbg.so
#
# Output:
# ../../code/T_/HTL/Linux/:../../code/P_/HTL/Linux/
#

AWK_LIBS_PATH='BEGIN { lastdir= "" ;  outline="" }\
{\
  levels=split($$0,elems,"/"); \
  curdir = substr($$0,1,length-length(elems[levels])); \
  if (curdir != lastdir) {\
    outline = sprintf ("%s%s:",outline, curdir);\
    lastdir=curdir;\
  }\
}\
END { print outline }'

# List libraries command
LDD=ldd

#VERBOSE=1
# target to show the set of libraries used by the test program
ifdef VERBOSE
showlibs:
	@(LD_LIBRARY_PATH=$(LIBPATH)$(LD_LIBRARY_PATH);export LD_LIBRARY_PATH;\
	$(LDD) $(OS)/$(APPL))

else
showlibs:
endif

testrun: showlibs
	@(LD_LIBRARY_PATH=$(LIBPATH)$(LD_LIBRARY_PATH);export LD_LIBRARY_PATH;\
	$(OS)/$(APPL))
